/*	Configuration manager for Hypertext Daemon		HTConfig.c
**	==========================================
**
**
** History:
**	 3 Jun 91	Written TBL
**	10 Aug 91	Authorisation added after Daniel Martin (pass, fail)
**			Rule order in file changed
**			Comments allowed with # on 1st char of rule line
**      17 Jun 92       Bug fix: pass and fail failed if didn't contain '*' TBL
**       1 Sep 93       Bug fix: no memory check - Nathan Torkington
**                      BYTE_ADDRESSING removed - Arthur Secret
**	11 Sep 93  MD	Changed %i into %d in debug printf. 
**			VMS does not recognize %i.
**			Bug Fix: in case of PASS, only one parameter to printf.
**	19 Sep 93  AL	Added Access Authorization stuff.
**	 1 Nov 93  AL	Added htbin.
**	30 Nov 93  AL	Added HTTranslateReq().
**	 4 Feb 94  AL	Moved to daemon distribution and renamed
**			(HTRules->HTConfig).  Library still has its
**			own HTRules.c, which this file overrides.
**	04 Mar 94  MD	Added case insensitivity option checking for rules.
**
*/

/* (c) CERN WorldWideWeb project 1990,91. See Copyright.html for details */
#include "HTRules.h"

#include "tcp.h"
#include "HTFile.h"
#include "HTParse.h"	/* HTParse() */
#include "HTAAUtil.h"
#include "HTAAProt.h"
#include "HTMulti.h"	/* HTAddWelcome() */
#include "HTDaemon.h"
#include "HTConfig.h"
#include "HTLog.h"


#ifndef VMS
#include <pwd.h>	/* Unix password file routine: getpwnam() */
#else /* VMS */
#include "HTVMSUtils.h"
#endif /* VMS */


#define LINE_LENGTH 256


typedef struct _rule {
	struct _rule *	next;
	HTRuleOp	op;
	char *		pattern;
	char *		equiv;
	char *		prot_file;
	char *		prot_ids;
	HTAAProt *	prot_setup;
} rule;


typedef struct _HTAAProtNode {
    char *	prot_name;
    HTAAProt *	prot;
} HTAAProtNode;

#define TIME_TRACE(m,t) \
    if (TRACE) fprintf(stderr,"%s %s\n", m, make_time_spec(&t))


/*	Global variables
**	----------------
*/
PUBLIC char * HTBinDir = NULL;		/* Physical /htbin directory	*/
                                        /* path (this is obsolite, and	*/
                                        /* only used if HTSearchScript	*/
                                        /* is given as non-absolute --	*/
                                        /* and this is only for rule	*/
                                        /* file backward-compatibility.	*/
PUBLIC char * HTSearchScript = NULL;	/* Search script name.		*/
PUBLIC char * HTPutScript = NULL;	/* Script to handle PUT		*/
PUBLIC char * HTPostScript = NULL;	/* Script to handle POST	*/
PUBLIC char * HTServerHostName = NULL;	/* Used for CGI scripts		*/
PUBLIC char * HTPortAddr = NULL;	/* Default port address string	*/
PUBLIC BOOL HTCachingExplicitlyOff = NO;/* Is caching disabled		*/
PUBLIC char * HTProxyCacheRoot =NULL;	/* Cache directory for gateway	*/
PUBLIC char * HTMetaDir = NULL;		/* Directory for metafiles	*/
PUBLIC char * HTMetaSuffix = NULL;	/* Suffix for metafiles		*/
PUBLIC HTList * HTNoLog = NULL;		/* Machines for which no logging*/
PUBLIC HTList * HTNoCaching = NULL;	/* No caching if matching these	*/
PUBLIC BOOL HTDisabled[(int)MAX_METHODS + 1];	/* Disabled methods	*/
PUBLIC BOOL HTDoSetUid = NO;		/* Do setuid() to auth'd uid	*/

PUBLIC char * HTCacheAccessLog = NULL;	/* Cache access log		*/
PUBLIC BOOL HTNewLogFileFormat = NO;	/* Use new common logfile format*/
PUBLIC BOOL HTUseGMT = NO;		/* Use GMT instead of localtime	*/
PUBLIC BOOL HTNoBackGround = NO;	/* Don't auto-go background	*/
PUBLIC BOOL HTStandAlone = NO;		/* Am I standalone?		*/
PUBLIC char * HTUserDir = NULL;		/* User supported directory name*/
PUBLIC char * HTUserId = NULL;		/* Default user id		*/
PUBLIC char * HTGroupId = NULL;		/* Default group id		*/

PUBLIC BOOL do_rfc931 = NO;		/* Should we check remote ident	*/

PUBLIC BOOL do_accept_hack = YES;
        /* Do we send the document even if there was no corresponding	*/
        /* Accept header line. This is because even currently browsers	*/
        /* which are NOT sending Accept's are widely used. I'll disable	*/
        /* this feature by default soon. For now it makes things easier.*/

PUBLIC char * server_root = NULL;	/* Server's "home directory"	*/
PUBLIC BOOL icons_inited = NO;		/* If not we'll use defaults	*/

PUBLIC int HTInputTimeOut  = DEFAULT_INPUT_TIMEOUT;
PUBLIC int HTOutputTimeOut = DEFAULT_OUTPUT_TIMEOUT;
PUBLIC int HTScriptTimeOut = DEFAULT_SCRIPT_TIMEOUT;

PUBLIC BOOL    always_welcome  = YES;	/* Redirect directory names to	*/
                                        /* welcome page on that dir.	*/

PUBLIC HTList* cache_clean	=NULL;		/* Remove always after	*/
PUBLIC time_t  cache_clean_def	=21*24*3600;	/* this time (3 weeks).	*/

PUBLIC HTList* cache_unused	=NULL;		/* Remove after unused	*/
PUBLIC time_t  cache_unused_def	=14*24*3600;	/* this long (2 weeks)	*/

PUBLIC HTList* cache_exp	=NULL;		/* Default expires when	*/
PUBLIC time_t  cache_exp_def	= 0;		/* not given.		*/
       /* 0 defaults to 1 day for ftp: and gopher:, but 0 for others	*/
       /* This fixes the problem with script-generated documents going	*/
       /* to the cache if they don't have Expires: field.		*/

PUBLIC long    cache_max_k	=5*1024;/* Default size in K (5MB small)*/
PUBLIC int     cache_max_f	=1024;	/* At most this many cache files*/
PUBLIC int     cache_limit_1	= 100;	/* Size no problem until 200K	*/
PUBLIC int     cache_limit_2	=4000;	/* No caching after 4MB		*/
PUBLIC float   cache_lm_factor  =0.1;	/* For approximating Expires:	*/
PUBLIC time_t  cache_time_margin=120;	/* Time accuracy (in seconds)	*/
PUBLIC BOOL    cache_exp_check	=YES;	/* Do expiry check		*/

PUBLIC BOOL    cache_no_connect =NO;
	/* Use only local cache and return error if the document was	*/
	/* not in the cache.						*/

PUBLIC BOOL    gc_disabled	= NO;	/* Never do garbage collection	*/
PUBLIC int     gc_req_interval	= 0;	/* gc every n reqs - Off	*/
PUBLIC time_t  gc_time_interval	= 0;	/* gc every n seconds - Off	*/
PUBLIC time_t  gc_daily_gc	= 3*3600;	/* gc daily at 3:00am	*/
PUBLIC int     gc_mem_usage_control=500;	/* How radical mem usage*/

PUBLIC time_t  cache_lock_timeout = DEFAULT_OUTPUT_TIMEOUT;

PUBLIC char *  pid_file = NULL;		/* File to write the pid number	*/

#ifdef VMS
PUBLIC BOOL HTRuleCaseSense = NO;	/* rules are checked case insensitive*/
#else /* not VMS */
PUBLIC BOOL HTRuleCaseSense = YES;	/* rules are checked case sensitive */
#endif /* not VMS */

extern BOOL HTSuffixCaseSense;		/* From HTFile.c		*/


PRIVATE FILE * hack_conf_fp = NULL;

PRIVATE char * rule_names[] =
{
    "Invalid.....", "Map.........", "Pass........", "Fail........",
    "DefProt.....", "Protect.....", "Exec........", "Redirect...."
};


/*	Module-wide variables
**	---------------------
*/

PRIVATE rule * rules = 0;	/* Pointer to first on list */
PRIVATE rule * rule_tail = 0;	/* Pointer to last on list */
PRIVATE BOOL reloading = NO;	/* Are we reloding rules */
PRIVATE HTList * rule_files = NULL;	/* All loaded files by name */
PRIVATE HTList * named_prots = NULL;	/* Named protections setups */



PRIVATE void HTAddNamedProtection ARGS2(char *,		name,
					HTAAProt *,	prot)
{
    HTAAProtNode * node = (HTAAProtNode*)calloc(1,sizeof(HTAAProtNode));
    if (!node) outofmem(__FILE__, "HTAddNamedProtection");

    StrAllocCopy(node->prot_name, name);
    node->prot = prot;

    if (!named_prots) named_prots = HTList_new();
    HTList_addObject(named_prots, (void*)node);
}


PRIVATE HTAAProt * HTGetNamedProtection ARGS1(char *, name)
{
    HTList * cur = named_prots;
    HTAAProtNode * node;

    if (!name) return NULL;

    CTRACE(stderr, "Looking up.. named protection \"%s\"\n", name);

    while ((node = (HTAAProtNode*)HTList_nextObject(cur))) {
	CTRACE(stderr, "Checking.... \"%s\"\n", node->prot_name);
	if (node->prot_name && !strcasecomp(node->prot_name, name)) {
	    CTRACE(stderr, "Found.......\n");
	    return node->prot;
	}
    }
    return NULL;
}


/*	Add rule to the list					HTAddRule()
**	--------------------
**
**  On entry,
**	pattern		points to 0-terminated string containing a single "*"
**	equiv		points to the equivalent string with * for the
**			place where the text matched by * goes.
**  On exit,
**	returns		0 if success, -1 if error.
*/

PUBLIC int HTAddRule ARGS3(HTRuleOp,		op,
			   CONST char *,	pattern,
			   CONST char *,	equiv)
{
    rule * r = (rule *)calloc(1, sizeof(rule));

    if (!r) outofmem(__FILE__, "HTAddRule"); 

    StrAllocCopy(r->pattern, pattern);

    if (equiv) {
	CONST char * from = equiv;
	CONST char * to = equiv + strlen(equiv) - 1;

	/* Ignore whitespace */
	while (*from && WHITE(*from)) from++;
	while (to>from && WHITE(*to)) to--;

	if (to>=from) {
	    char * p = (char*)malloc(to-from+2);

	    if (!p) outofmem(__FILE__, "HTAddRule");
	    strncpy(p, from, to-from+1);
	    p[to-from+1] = 0;

	    if (op == HT_Protect || op == HT_DefProt) {
		if (strchr(p,'/')) {	/* Abs.path of prot.setup file given */
		    char * killme = p;
		    char * pr_file = HTNextField(&p);
		    char * pr_ids = HTNextField(&p);
		    StrAllocCopy(r->prot_file, pr_file);
		    StrAllocCopy(r->prot_ids, pr_ids);
		    free(killme);
		}
		else {	/* Named protection setup used */
		    r->prot_setup = HTGetNamedProtection(p);
		    if (!r->prot_setup)
			HTLog_error2("Protection setup name undefined:",p);
		    free(p);
		}
	    }
	    else r->equiv = p;
	}
    }
    r->op = op;
    if (TRACE) {
	if (equiv)
	    fprintf(stderr, "%s \"%s\" --> \"%s\"\n",
		    rule_names[(int)op], r->pattern,
		    (r->equiv ? r->equiv :
		     (r->prot_file ? r->prot_file : "-null-")));
	else
	    fprintf(stderr, "%s \"%s\"\n",
		    rule_names[(int)op], r->pattern);
    }

    if (rule_tail) rule_tail->next = r;
    else rules = r;
    rule_tail = r;

    return 0;
}


PUBLIC void HTPrependPass ARGS2(CONST char *,	pattern,
				CONST char *,	equiv)
{
    rule * r = (rule *)calloc(1, sizeof(rule));

    if (!r) outofmem(__FILE__, "HTAddRule"); 

    r->op = HT_Pass;
    StrAllocCopy(r->pattern, pattern);
    if (equiv) {
	StrAllocCopy(r->equiv, equiv);
	CTRACE(stderr,"AutoRule.... Pass \"%s\" --> \"%s\"\n",pattern,equiv);
    }
    else {
	CTRACE(stderr,"AutoRule.... Pass \"%s\"\n",pattern);
    }

    r->next = rules;
    rules = r;
    if (!rule_tail) rule_tail = r;
}



PRIVATE int HTAddProtRule ARGS3(HTRuleOp,	op,
				CONST char *,	pattern,
				HTAAProt *,	prot)
{
    rule * r = (rule *)calloc(1, sizeof(rule));

    if (!r) outofmem(__FILE__, "HTAddProtRule"); 

    StrAllocCopy(r->pattern, pattern);
    r->op = op;
    r->prot_setup = prot;

    CTRACE(stderr, "%s \"%s\" (setup in config file, parsed above)\n",
	   rule_names[(int)op], r->pattern);

    if (rule_tail) rule_tail->next = r;
    else rules = r;
    rule_tail = r;

    return 0;
}



/*	Clear all rules						HTClearRules()
**	---------------
**
** On exit,
**	There are no rules
**	returns		0 if success, -1 if error.
**
** See also
**	HTAddRule()
*/
PUBLIC int HTClearRules NOARGS
{
    while (rules) {
    	rule * temp = rules;
	rules = temp->next;
	FREE(temp->pattern);
	FREE(temp->equiv);
	FREE(temp->prot_file);
	FREE(temp->prot_ids);
	free(temp);
    }
    rule_tail = 0;

    return 0;
}


#ifdef NOT_NEEDED
PRIVATE char * wrap ARGS1(char *, s)
{
    if (s && *s=='/') {
	char * n = (char*)malloc(strlen(s) + 6);
	if (!n) outofmem(__FILE__, "HTRules:wrap");
	sprintf(n, "file:%s", s);
	free(s);
	return n;
    }
    return s;
}
#endif


PRIVATE char * unwrap ARGS1(char *, s)
{
    if (s && !strncmp(s, "file:", 5)) {
	char * t;
	char * n = NULL;

	if (!strncmp(s, "file://", 7))
	    t = strchr(s+7, '/');
	else t = s+5;

	if (!t) {
	    free(s);
	    return NULL;
	}
	StrAllocCopy(n, t);
	free(s);
	return n;
    }
    return s;
}


PRIVATE char * map ARGS3(rule *,	r,
			 CONST char *,	q,
			 int,		m)
{
    char * ins;
    char * result = NULL;

    if (!r || !r->equiv || !q)
	return NULL;

    ins = strchr(r->equiv, '*');	/* Insertion point */
    if (ins) {	/* Consistent rule!!! */
	result = (char *)malloc(strlen(r->equiv)-1 + m + 1);
	if (!result) outofmem(__FILE__, "HTTranslate:map"); /* NT & AS */
	strncpy(result, r->equiv, ins-r->equiv); /* result may be unterm. */
	strncpy(result+(ins-r->equiv), q, m);	 /* Matched bit */
	strcpy (result+(ins-r->equiv)+m, ins+1);	 /* Last bit */
    } else {	/* No insertion point */
	result = (char *)malloc(strlen(r->equiv)+1);
	if (!result) outofmem(__FILE__, "HTTranslate:map"); /* NT & AS */
	strcpy(result, r->equiv);
    } /* If no insertion point exists */

    return result;
}


PRIVATE char * get_user_dir ARGS3(char *,	current,
				  char **,	out_r_home,
				  int *,	out_r_home_len)
{
    char * username = current+2;
    char * end = strchr(username, '/');
    char * r_home = NULL;
    int r_home_len = 0;

    if (end)  *end++ = 0;

    if (*username) {
	struct passwd * pw = getpwnam(username);

	if (pw && pw->pw_dir) {
	    int homelen = strlen(pw->pw_dir);
	    char * d = (char*)malloc(homelen +
				     strlen(HTUserDir) +
				     (end ? strlen(end) : 0) + 3);

	    r_home_len = homelen + 1;
	    r_home = (char*)malloc(r_home_len + 1);
	    if (!r_home) outofmem(__FILE__, "get_user_dir");
	    sprintf(r_home, "%s/", pw->pw_dir);

	    strcpy(d, pw->pw_dir);
	    if (pw->pw_dir[homelen-1] != '/' && HTUserDir[0] != '/')
		strcat(d, "/");
	    strcat(d, HTUserDir);
	    if (end) {
		if (HTUserDir[strlen(HTUserDir)-1] != '/')
		    strcat(d, "/");
		strcat(d, end);
	    }
	    CTRACE(stderr, "UserDir..... \"%s\" mapped to \"%s\"\n",
		   current, d);
	    free(current);
	    current = d;
	}
	else CTRACE(stderr,
		    "ERROR....... User dir for \"%s\" not found **\n",
		    current);
    }
    else CTRACE(stderr,
		"ERROR....... Invalid user dir request \"%s\"\n",
		current);

    if (*out_r_home) free(*out_r_home);
    *out_r_home = r_home;
    *out_r_home_len = r_home_len;

    return current;
}



/*	Translate by rules					HTTranslate()
**	------------------
**
** ATTENTION:
**	THIS FUNCTION HAS BEEN OBSOLITED BY HTTranslateReq()
**	ON SERVER SIDE -- ON BROWSER SIDE THIS IS STILL USED!
**	Don't add new server features to this, this already has
**	more than it can handle cleanly.
**
**	The most recently defined rules are applied last.
**
** On entry,
**	required	points to a string whose equivalent value is neeed
** On exit,
**	returns		the address of the equivalent string allocated from
**			the heap which the CALLER MUST FREE. If no translation
**			occured, then it is a copy of te original.
** NEW FEATURES:
**			When a "protect" or "defprot" rule is mathed,
**			a call to HTAA_setCurrentProtection() or
**			HTAA_setDefaultProtection() is made to notify
**			the Access Authorization module that the file is
**			protected, and so it knows how to handle it.
**								-- AL
*/
PUBLIC char * HTTranslate ARGS1(CONST char *, required)
{
    rule * r;
    char *current = NULL;
    static char *r_home = NULL;
    int r_home_len = 0;

    if (!required)  return NULL;

    FREE(r_home);	/* From previous call */
    StrAllocCopy(current, required);

#ifdef VMS 
   /* convert to lowercase */
   {
      char * q = current;
      for(; *q ; q++)
         *q = TOLOWER(*q);
      
      if (r_home)
         for(q = r_home; *q ; q++)
            *q = TOLOWER(*q);
   }
#endif

    for(r = rules; r; r = r->next) {
        char * p = r->pattern;
	int m=0;   /* Number of characters matched against wildcard */
	CONST char * q = NULL;

	if (HTUserDir && current[0] == '/' && current[1] == '~')
	    current = get_user_dir(current, &r_home, &r_home_len);

	q = current;
	for(;*p && *q; p++, q++) {   /* Find first mismatch */
            if (HTRuleCaseSense) {            
   	       if (*p!=*q) break;
            } else {
   	       if ((tolower(*p)) != (tolower(*q))) break;
            }
 	}

	if (*p == '*') {		/* Match up to wildcard */
	    m = strlen(q) - strlen(p+1); /* Amount to match to wildcard */
	    if(m<0) continue;           /* tail is too short to match */
            if (HTRuleCaseSense) {
      	       if (0!=strcmp(q+m, p+1)) continue;	/* Tail mismatch */
            } else {
   	       if (0!=strcasecomp(q+m, p+1)) continue;	/* Tail mismatch */
            }
	} else 				/* Not wildcard */
            if (HTRuleCaseSense) {
   	       if (*p != *q) continue;	/* plain mismatch: go to next rule */
            } else {
   	       if ((tolower(*p)) != (tolower(*q))) continue;	/* plain mismatch: go to next rule */
            }
	switch (r->op) {		/* Perform operation */

#ifdef ACCESS_AUTH
	  case HT_DefProt:
	  case HT_Protect:
	    CTRACE(stderr, "Match....... \"%s\" with %s rule, setup %s\n",
		   current,
		   rule_names[ (int)(r->op) ],
		   (r->equiv ? r->equiv :
		    (r->op==HT_Protect ? "-default-" :"-not-set-")));
	    /* continue translating rules */
	    break;
#endif /* ACCESS_AUTH */

	  case HT_Pass:				/* Authorised */
    		if (!r->equiv) {
		    CTRACE(stderr, "Pass........ \"%s\"\n", current);
		    return unwrap(current);
	        }
		/* Else fall through ...to map and pass */
		
	  case HT_Map:
	    if (*p == *q) { /* End of both strings, no wildcard */
		CTRACE(stderr,
		       "Mapping..... \"%s\" to \"%s\"\n", current, r->equiv);  
		StrAllocCopy(current, r->equiv); /* use entire translation */
	    } else {
		if (!r_home || !strncmp(r_home, r->pattern, r_home_len)) {
		    char * mapped = map(r, q, m);
		    if (mapped) {
			CTRACE(stderr, "Mapping..... \"%s\" to \"%s\"\n",
			       current, mapped);
			free(current);
			current = mapped;			/* Use this */
		    }
		}
		else {
		    CTRACE(stderr,
		   "NoMatch..... User dir \"%s\" does NOT match \"%s\"\n",
			   current, r->pattern);
		    CTRACE(stderr,
		   "Explanation. Entire home dir \"%s\" should be matched\n",
			   r_home);
		}

		if (r->op == HT_Pass) {
		    CTRACE(stderr, "Passing..... \"%s\"\n", current);
		    return unwrap(current);
		}
	    }
	    break;

	  case HT_Exec:
	  case HT_Invalid:
	  case HT_Fail:				/* Unauthorised */
	  default:
	    if (r_home && !strcmp(r->pattern,"*")) {
		CTRACE(stderr,
		       "No effect... User directories don't match Fail *\n");
	    }
	    else {
		CTRACE(stderr,"Failing..... \"%s\" **\n", current);
		free(current);	/* Leak fixed AL 6 Feb 1994 */
		return NULL;
	    }
	} /* if tail matches ... switch operation */

    } /* loop over rules */

    if (current) free(current);
    return NULL;	/* Default for server is to fail */
}



/*	Translate by rules				HTTranslateReq()
**	------------------
**
** On entry,
**	req		request structure.
**	HTReqArgPath	simplified pathname (no ..'s etc in it),
**			which will be translated.
**			If this starts with /htbin/ it is taken
**			to be a script call request.
**
** On exit,
**	returns		YES on success, NO on failure (Forbidden).
**	HTReqTranslated	contains the translated filename;
**			NULL if a script call.
**	HTReqScript	contains the executable script name;
**			NULL if not a script call.
*/
PUBLIC BOOL HTTranslateReq ARGS1(HTRequest *, req)
{
    rule * r;
    char *current = NULL;
    static char *r_home = NULL;
    int r_home_len = 0;

    if (!req  ||  !HTReqArgPath) {
	HTReason = HTAA_BY_RULE;
	return NO;
    }

    FREE(r_home);			/* From previous call */
    StrAllocCopy(current, HTReqArgPath);

#ifdef VMS 
   /* convert to lowercase */
   {
      char * q = current;
      for(; *q ; q++)
         *q = TOLOWER(*q);
      
      if (r_home)
         for(q = r_home; *q ; q++)
            *q = TOLOWER(*q);
   }
#endif

    for(r = rules; r; r = r->next) {
        char * p = r->pattern;
	int m=0;   /* Number of characters matched against wildcard */
	CONST char * q = NULL;

	if (HTUserDir && current[0] == '/' && current[1] == '~') {
	    if (!strchr(current+2, '/')) {  /* Need to add slash and redir */
		HTLocation = HTEscape(HTReqArgPath, URL_PATH);
		StrAllocCat(HTLocation, "/");
		if (HTReqArgKeywords) {
		    StrAllocCat(HTLocation, "?");
		    StrAllocCat(HTLocation, HTReqArgKeywords);
		}
		free(current);
		HTReason = HTAA_OK_REDIRECT;
		return YES;
	    }
	    else {
		current = get_user_dir(current, &r_home, &r_home_len);
	    }
	}

	q = current;
	for(;*p && *q; p++, q++) {   /* Find first mismatch */
            if (HTRuleCaseSense) {
   	       if (*p!=*q) break;
            } else {
   	       if ((tolower(*p)) != (tolower(*q))) break;
            }
	}

	if (*p == '*') {		/* Match up to wildcard */
	    m = strlen(q) - strlen(p+1); /* Amount to match to wildcard */
	    if(m<0) continue;           /* tail is too short to match */
            if (HTRuleCaseSense) {
   	       if (0!=strcmp(q+m, p+1)) continue;	/* Tail mismatch */
            } else {
   	       if (0!=strcasecomp(q+m, p+1)) continue;	/* Tail mismatch */
            }
	} else {			/* Not wildcard */
            if (HTRuleCaseSense) {
   	       if (*p != *q) continue;	/* plain mismatch: go to next rule */
            } else {
   	       if ((tolower(*p)) != (tolower(*q))) continue;	/* plain mismatch: go to next rule */
	   }
	}

	switch (r->op) {		/* Perform operation */

#ifdef ACCESS_AUTH
	  case HT_DefProt:
	  case HT_Protect:
	    CTRACE(stderr, "Match....... \"%s\" with %s rule, setup %s\n",
		   current,
		   rule_names[ (int)(r->op) ],
		   (r->prot_file ? r->prot_file :
		    (r->prot_setup ? "as defined in config file" :
		     (r->op==HT_Protect ? "-default-" : "-not-set-"))));

	    if (r->prot_setup) {
		if (r->op == HT_Protect)
		    HTProt = r->prot_setup;
		else HTDefProt = r->prot_setup;
	    }
	    else {
		if (r->op == HT_Protect)
		    HTAA_setCurrentProtection(req, r->prot_file, r->prot_ids);
		else HTAA_setDefaultProtection(req, r->prot_file, r->prot_ids);
	    }
	    /* continue translating rules */
	    break;
#endif  /* ACCESS_AUTH */

	case HT_Exec:
	    if (!r->equiv) {
		CTRACE(stderr, "DirectExec.. \"%s\" [no extra pathinfo]\n",
		       current);
		HTReqScript = current;
		HTScriptPathInfo = NULL;
		HTScriptPathTrans = NULL;
		return YES;
	    }
	    else if (*p == *q || !strchr(r->equiv, '*')) { /* No wildcards */
		CTRACE(stderr, "MapExec..... \"%s\" [no extra pathinfo]\n",
		       r->equiv);
		StrAllocCopy(HTReqScript, r->equiv);
		HTScriptPathInfo = NULL;
		HTScriptPathTrans = NULL;
		free(current);		/* Leak fixed AL 6 Feb 1994 */
		return YES;
	    }
	    else {
		char *ins = strchr(r->equiv, '*');
		char *pathinfo;

		HTReqScript = map(r, q, m);
		pathinfo = HTReqScript+(ins-r->equiv)+1;
		while (*pathinfo && *pathinfo != '/') pathinfo++;

		FREE(HTScriptPathInfo);
		FREE(HTScriptPathTrans);

		if (*pathinfo) {
		    StrAllocCopy(HTScriptPathInfo, pathinfo);
		    *pathinfo = 0;
		    HTScriptPathTrans =
			unwrap(HTTranslate(HTScriptPathInfo));
		}
		CTRACE(stderr,
		     "Exec........ \"%s\"\nPathinfo.... %s\nTranslated.. %s\n",
		     HTReqScript,
		     (HTScriptPathInfo ? HTScriptPathInfo : "-none-"),
		     (HTScriptPathTrans? HTScriptPathTrans: "-none-"));
		free(current);		/* Leak fixed AL 6 Feb 1994 */
		return YES;
	    }
	    break;

	  case HT_Redirect:
	    if (!r->equiv) {
		CTRACE(stderr, "ERROR....... No destination for redirect \"%s\"\n",
		       r->pattern);
		HTReason = HTAA_INVALID_REDIRECT;
		free(current);		/* Leak fixed AL 6 Feb 1994 */
		return NO;
	    }
	    else HTReason = HTAA_OK_REDIRECT;
	    /* And fall through to HT_Map... */

	case HT_Pass:				/* Authorised */
    		if (!r->equiv) {
		    CTRACE(stderr, "Passing..... \"%s\"\n", current);
		    HTReqTranslated = unwrap(current);
		    return YES;
	        }
		/* Else fall through ...to map and pass */

	case HT_Map:
	    if (*p == *q) { /* End of both strings, no wildcard */
		CTRACE(stderr,"Mapping..... \"%s\" to \"%s\"\n",
		       current, r->equiv);  
		StrAllocCopy(current, r->equiv); /* use entire translation */
	    } else {
		if (!r_home || !strncmp(r_home, r->pattern, r_home_len)) {
		    char * mapped = map(r, q, m);
		    if (mapped) {
			CTRACE(stderr, "Mapping..... \"%s\" to \"%s\"\n",
			       current, mapped);
			free(current);
			current = mapped;			/* Use this */
		    }
		}
		else {
		    CTRACE(stderr,
			   "NoMatch..... User dir \"%s\" does NOT match \"%s\"\n",
			   current, r->pattern);
		    CTRACE(stderr,
			   "Explanation. Entire home dir \"%s\" should be matched\n",
			   r_home);
		}
	    }

	    if (r->op == HT_Pass) {
		CTRACE(stderr, "Passing..... \"%s\"\n", current);
		HTReqTranslated = unwrap(current);
		return YES;
	    }
	    else if (r->op == HT_Redirect) {
		CTRACE(stderr, "Redirected.. to \"%s\"\n", current);
		HTLocation = current;
		return YES;
	    }
	    break;

	case HT_Invalid:
	case HT_Fail:				/* Unauthorised */
	    if (r_home && !strcmp(r->pattern,"*")) {
		CTRACE(stderr,
		       "No effect... User directories don't match Fail *\n");
	    }
	    else {
		CTRACE(stderr, "Fail........ \"%s\" **\n", current);
		free(current);	/* Leak fixed AL 6 Feb 1994 */
		HTReason = HTAA_BY_RULE;
		return NO;
	    }
	    break;
	} /* if tail matches ... switch operation */

    } /* loop over rules */

    if (current) {
	if (r_home) {	/* User supported directory => pass */
	    HTReqTranslated = current;
	    CTRACE(stderr, "AutoPass.... Passing user-supported directory \"%s\"\n",
		   current);
	    return YES;
	}
	else {
	    CTRACE(stderr, "DefaultFail. Failing \"%s\" by default\n",
		   current);
	    CTRACE(stderr, "Explanation. no Pass rule\n");
	    free(current);
	}
    }
    HTReqTranslated = NULL;
    HTReason = HTAA_BY_RULE;

    return NO;	/* Default for server is to fail (explicit Pass required) */
}


PRIVATE void standardize ARGS1(char *, word)
{
    char * cur = word;
    char * p;

    if (!word) return;

    while (*cur) {
	*cur = TOLOWER(*cur);
	if (*cur == '-' || *cur == '_') {
	    for (p=cur; *p; p++) *p = *(p+1);
	}
	else cur++;
    }
}


PRIVATE BOOL positive ARGS1(char *, word)
{
    standardize(word);
    if (word &&
	(!strcmp(word,"yes") || !strcmp(word,"on") || !strcmp(word,"ok") ||
	 !strncmp(word,"enable",6)))
	return YES;
    else
	return NO;
}


PRIVATE BOOL negative ARGS1(char *, word)
{
    standardize(word);
    if (word &&
	(!strcmp(word,"no") || !strcmp(word,"off") || !strcmp(word,"none") ||
	 !strncmp(word,"disable",7)))
	return YES;
    else
	return NO;
}


PRIVATE BOOL unlimited ARGS1(char *, word)
{
    standardize(word);
    if (word &&
	(!strncmp(word,"unlimited",7) || !strncmp(word,"nolimit",7)))
	return YES;
    else
	return NO;
}


typedef struct _HTTemplAndTime {
    char *	templ;
    time_t	time;
} HTTemplAndTime;

PRIVATE HTTemplAndTime * tt_new ARGS2(char *,	templ,
				      time_t,	t)
{
    HTTemplAndTime * tt = NULL;
    if (templ) {
	tt = (HTTemplAndTime*)calloc(1,sizeof(HTTemplAndTime));
	if (!tt) outofmem(__FILE__, "tt_new");
	StrAllocCopy(tt->templ,templ);
	tt->time = t;
    }
    return tt;
}

PRIVATE HTTemplAndTime * parse_tt ARGS1(char *, params)
{
    char * templ = HTNextField(&params);
    time_t t;

    if (!params || !templ || !parse_time(params, 24*3600, &t))
	return NULL;
    templ = HTStrip(templ);
    CTRACE(stderr, "template \"%s\", time %s\n", templ, make_time_spec(&t));
    return tt_new(templ,t);
}

PRIVATE time_t get_tt ARGS3(HTList *,	list,
			    char *,	url,
			    time_t,	deflt)
{
    HTList * cur = list;
    HTTemplAndTime * tt;

    while ((tt = (HTTemplAndTime*)HTList_nextObject(cur))) {
	if (tt->templ  &&  HTAA_templateMatch(tt->templ, url))
	    return tt->time;
    }

    if (!deflt && (!strncmp(url,"ftp:",4) || !strncmp(url,"gopher:",7)))
	return (time_t)(24*3600);  /* One day by default for ftp and gopher */
    else
	return deflt;	/* Zero (or what specified) for http and others */
}

PUBLIC time_t get_default_expiry ARGS1(char *, url)
{
    return get_tt(cache_exp, url, cache_exp_def);
}

PUBLIC time_t get_unused ARGS1(char *, url)
{
    return get_tt(cache_unused, url, cache_unused_def);
}

PUBLIC time_t get_clean ARGS1(char *, url)
{
    return get_tt(cache_clean, url, cache_clean_def);
}




/*	Load one line of configuration
**	------------------------------
**
**	Call this, for example, to load a X resource with config info.
**
** returns	0 OK, < 0 syntax error.
*/
PUBLIC int HTSetConfiguration ARGS1(CONST char *, config)
{
    HTRuleOp op;
    char * line = NULL;
    char * pointer = line;
    char *word1, *word2, *word3;
    float quality, secs, secs_per_byte;
    int status;
    static char * params = NULL;
    
    StrAllocCopy(line, config);
    {
	char * p = strchr(line, '#');	/* Chop off comments */
	if (p) *p = 0;
    }
    pointer = line;
    word1 = HTNextField(&pointer);
    if (!word1) {
    	free(line);
	return 0;
    }	/* Comment only or blank */

    standardize(word1);

    StrAllocCopy(params,pointer);
    word2 = HTNextField(&pointer);

    if (!strncmp(word1, "defprot", 7) ||
	!strncmp(word1, "protect", 7))
	word3 = pointer;  /* The rest of the line to be parsed by AA module */
    else
	word3 = HTNextField(&pointer);	/* Just the next word */

    if (!word2) {
	fprintf(stderr, "HTRule: Insufficient operands: %s\n", line);
	free(line);
	return -2;	/*syntax error */
    }

    if (!strcmp(word1, "suffix") ||
	!strcmp(word1, "addtype")) {
        char * encoding = HTNextField(&pointer);
	if (pointer) status = sscanf(pointer, "%f", &quality);
	else status = 0;
	HTAddType(word2, word3,
		  encoding ? encoding : "binary",
		  status >= 1? quality : 1.0);

    } else if (!strcmp(word1, "addencoding") ||
	       !strcmp(word1, "encoding")) {
	if (pointer)
	    status = sscanf(pointer, "%f", &quality);
	else status = 0;
	HTAddEncoding(word2, word3,
		      status >= 1 ? quality : 1.0);

    } else if (!strncmp(word1, "addlang", 7) ||
	       !strncmp(word1, "lang", 4)) {
	if (pointer)
	    status = sscanf(pointer, "%f", &quality);
	else status = 0;
	HTAddLanguage(word2, word3,
		      status >= 1 ? quality : 1.0);

    } else if (!strcmp(word1, "presentation")) {
        if (pointer) status = sscanf(pointer, "%f%f%f",
				     &quality, &secs, &secs_per_byte);
        else status = 0;
	if (!HTConversions) HTConversions = HTList_new();
	HTSetPresentation(HTConversions, word2, word3,
			  status >= 1 ? quality 	: 1.0,
			  status >= 2 ? secs 		: 0.0,
			  status >= 3 ? secs_per_byte 	: 0.0 );

    } else if (!strncmp(word1, "metadir", 7)) {
	StrAllocCopy(HTMetaDir, word2);

    } else if (!strncmp(word1, "metasuff", 8)) {
	StrAllocCopy(HTMetaSuffix, word2);

    } else if (!strncmp(word1, "nolog", 5)) {
	char * mask = NULL;

	if (!HTNoLog) HTNoLog = HTList_new();

	StrAllocCopy(mask, word2);
	HTList_addObject(HTNoLog, (void*)mask);
	CTRACE(stderr, "NoLogging... for host \"%s\"\n", mask);

	while (word3) {
	    mask = NULL;
	    StrAllocCopy(mask, word3);
	    CTRACE(stderr, "NoLogging... for host \"%s\"\n", mask);
	    HTList_addObject(HTNoLog, (void*)mask);
	    word3 = HTNextField(&pointer);
	}

    } else if (strstr(word1, "dirshow")) {	/* All DirShow* directives */
	BOOL flag = positive(word2);

	if (!strncmp(word1,"dirshowicons",10)) {
	    if (flag)	HTDirShowMask |= HT_DIR_SHOW_ICON;
	    else	HTDirShowMask &= ~HT_DIR_SHOW_ICON;
	    CTRACE(stderr, "DirIcons.... %s\n", flag ? "On" : "Off");
	}
	else if (!strncmp(word1,"dirshowbrackets",10)) {
	    HTDirShowBrackets = flag;
	    CTRACE(stderr, "ALT tags.... %ssurrounded by square brackets\n",
		   flag ? "" : "not ");
	}
	else if (!strncmp(word1,"dirshowmaxdescriptionlength",12)) {
	    int n = atoi(word2);
	    if (n > 0) {
		HTDirMaxDescrLength = n;
		CTRACE(stderr,"Maximum..... description length %d chars\n",n);
	    }
	    else {
		HTLog_error2("Invalid DirShowMaxDescriptionLength parameter:",
			     word2);
	    }
	}
	else if (!strncmp(word1,"dirshowminlength",12)) {
	    int n = atoi(word2);
	    if (n > 2) {
		HTDirMinFileLength = n;
		CTRACE(stderr,"Minimum..... chars in filename %d\n",
		       HTDirMinFileLength);
		if (n > HTDirMaxFileLength) {
		    HTDirMaxFileLength = n;
		    CTRACE(stderr,"Maximum..... also set to the same value\n");
		}
	    }
	    else {
		HTLog_error2("Invalid DirShowMinLength parameter:",word2);
	    }
	}
	else if (!strncmp(word1,"dirshowmaxlength",12)) {
	    int n = atoi(word2);
	    if (n > 2) {
		HTDirMaxFileLength = n;
		CTRACE(stderr,"Maximum..... chars in filename %d\n",
		       HTDirMaxFileLength);
		if (n < HTDirMinFileLength) {
		    HTDirMinFileLength = n;
		    CTRACE(stderr,"Minimum..... also set to the same value\n");
		}
	    }
	    else {
		HTLog_error2("Invalid DirShowMaxLength parameter:",word2);
	    }
	}
	else if (!strncmp(word1,"dirshowdate",10)) {
	    if (flag)	HTDirShowMask |= HT_DIR_SHOW_DATE;
	    else	HTDirShowMask &= ~HT_DIR_SHOW_DATE;
	    CTRACE(stderr, "DirDate..... %s\n", flag ? "On" : "Off");
	}
	else if (!strncmp(word1,"dirshowsize",10)) {
	    if (flag)	HTDirShowMask |= HT_DIR_SHOW_SIZE;
	    else	HTDirShowMask &= ~HT_DIR_SHOW_SIZE;
	    CTRACE(stderr, "DirDate..... %s\n", flag ? "On" : "Off");
	}
	else if (!strncmp(word1,"dirshowbytes",10)) {
	    HTDirShowBytes = flag;
	    CTRACE(stderr,"DirBytes.... shown as %s\n", flag ? "bytes" : "1K");
	}
	else if (!strncmp(word1,"dirshowhidden",10)) {
	    if (flag)	HTDirShowMask |= HT_DIR_SHOW_HID;
	    else	HTDirShowMask &= ~HT_DIR_SHOW_HID;
	    CTRACE(stderr, "DirHidden... %s\n", flag ? "On" : "Off");
	}
	else if (!strncmp(word1,"dirshowowner",10)) {
	    if (flag)	HTDirShowMask |= HT_DIR_SHOW_OWNER;
	    else	HTDirShowMask &= ~HT_DIR_SHOW_OWNER;
	    CTRACE(stderr, "DirOwner.... %s\n", flag ? "On" : "Off");
	}
	else if (!strncmp(word1,"dirshowgroup",10)) {
	    if (flag)	HTDirShowMask |= HT_DIR_SHOW_GROUP;
	    else	HTDirShowMask &= ~HT_DIR_SHOW_GROUP;
	    CTRACE(stderr, "DirGroup.... %s\n", flag ? "On" : "Off");
	}
	else if (!strncmp(word1,"dirshowmode",10)) {
	    if (flag)	HTDirShowMask |= HT_DIR_SHOW_MODE;
	    else	HTDirShowMask &= ~HT_DIR_SHOW_MODE;
	    CTRACE(stderr, "DirMode..... %s\n", flag ? "On" : "Off");
	}
	else if (!strncmp(word1,"dirshowdescription",10)) {
	    HTDirDescriptions = flag;
	    CTRACE(stderr, "DirDescribe. %s\n", flag ? "On" : "Off");
	}
	else if (!strncmp(word1,"dirshowcase",10)) {
	    if (flag)	HTDirShowMask |= HT_DIR_SHOW_CASE;
	    else	HTDirShowMask &= ~HT_DIR_SHOW_CASE;
	    CTRACE(stderr, "DirCaseSense %s\n", flag ? "On" : "Off");
	}

    } else if (!strncmp(word1,"dircasesense",7)) {
	BOOL flag = positive(word2);
	if (flag) HTDirShowMask |= HT_DIR_SHOW_CASE;
	else	  HTDirShowMask &= ~HT_DIR_SHOW_CASE;
	CTRACE(stderr, "DirCaseSense %s\n", flag ? "On" : "Off");

    } else if (strstr(word1,"icon")) {		/* All *Icon* directives */
	if (!strcmp(word1,"addicon")) {
	    char * t;
	    while ((t = HTNextField(&pointer)))
		HTAddIcon(word2,word3,t);
	}
	else if (!strncmp(word1,"addunknownicon",6))
	    HTAddUnknownIcon(word2,word3);
	else if (!strncmp(word1,"addparenticon",6))
	    HTAddParentIcon(word2,word3);
	else if (!strncmp(word1,"adddiricon",6))
	    HTAddDirIcon(word2,word3);
	else if (!strncmp(word1,"addblankicon",6))
	    HTAddBlankIcon(word2,word3);
	else
	    HTLog_error2("Unknown Icon directive:",word1);

    } else if (!strncmp(word1, "serverroot", 10)) {
	StrAllocCopy(server_root, word2);
	CTRACE(stderr, "ServerRoot.. %s\n", server_root);

    } else if (!strncmp(word1, "disable", 7)) {
	HTMethod m = HTMethod_enum(word2);
	if (m != METHOD_INVALID) {
	    HTDisabled[(int)m] = YES;
	    CTRACE(stderr,"Disabled.... method %s\n", HTMethod_name(m));
	}
	else CTRACE(stderr, "ERROR....... Invalid method to disable \"%s\"\n",
		    word2);

    } else if (!strncmp(word1, "enable", 6)) {
	HTMethod m = HTMethod_enum(word2);
	if (m != METHOD_INVALID) {
	    HTDisabled[(int)m] = NO;
	    CTRACE(stderr,"Enabled..... method %s\n", HTMethod_name(m));
	}
	else CTRACE(stderr, "ERROR....... Invalid method to enable \"%s\"\n",
		    word2);

    } else if (!strncmp(word1, "identitycheck", 5)) {
	do_rfc931 = positive(word2);
	CTRACE(stderr, "IdentCheck.. %s\n", do_rfc931 ? "On" : "Off");

    } else if (!strncmp(word1, "accepthack", 10)) {
	/*
	 * No desire to document this... ;-)
	 */
	do_accept_hack = positive(word2);
	CTRACE(stderr,"AcceptHack.. %sabled\n", do_accept_hack ? "en" : "dis");

    } else if (!strncmp(word1, "cacheroot", 9)) {
	if (HTCachingExplicitlyOff) {
	    CTRACE(stderr, "CacheRoot... ignored (caching is off)\n");
	}
	else {
	    CTRACE(stderr, "CacheRoot... %s\n", word2);
	    StrAllocCopy(HTProxyCacheRoot, word2);
	}

    } else if (!strcmp(word1, "caching")) {
	HTCachingExplicitlyOff = negative(word2);
	if (HTCachingExplicitlyOff) {
	    CTRACE(stderr, "Caching..... off%s\n",
		   HTProxyCacheRoot ? " (CacheRoot directive ignored)" : "");
	    FREE(HTProxyCacheRoot);
	}
	else {
	    CTRACE(stderr, "Caching..... on\n");
	}

    } else if (!strncmp(word1, "nocaching", 6)) {
	char * mask = NULL;

	if (!HTNoCaching) HTNoCaching = HTList_new();

	StrAllocCopy(mask, word2);
	HTList_addObject(HTNoCaching, (void*)mask);
	CTRACE(stderr,"No caching.. if matching mask %s\n", mask);

    } else if (!strncmp(word1, "gctimeinterval", 9)) {
	if (negative(word2)) {
	    gc_time_interval = 0;
	    CTRACE(stderr,"gc.......... fixed interval disabled\n");
	}
	else {
	    if (parse_time(params, 3600, &gc_time_interval)) {
		TIME_TRACE("gc.......... interval", gc_time_interval);
	    } else HTLog_error2("Invalid gc_time_interval spec:",params);
	}

    } else if (!strncmp(word1, "gcreqinterval", 8)) {
	if (negative(word2)) {
	    gc_req_interval = 0;
	    CTRACE(stderr,"gc.......... request counter disabled\n");
	}
	else {
	    gc_req_interval = atoi(word2);
	    CTRACE(stderr,"gc.......... every %d requests\n",gc_req_interval);
	}

    } else if (!strncmp(word1, "gcdailygc", 7)) {
	if (negative(word2)) {
	    gc_daily_gc = 0;
	    CTRACE(stderr, "Daily gc.... disabled\n");
	}
	else {
	    if (parse_time(params, 3600, &gc_daily_gc)) {
		TIME_TRACE("Daily gc at. at", gc_daily_gc);
	    } else HTLog_error2("Invalid daily gc time spec:",params);
	}

    } else if (!strncmp(word1,"garbagecollection",10) || !strcmp(word1,"gc")) {
	gc_disabled = negative(word2);
	CTRACE(stderr, "Garbage collection %s\n", gc_disabled ? "Off" : "On");

    } else if (!strncmp(word1, "gcmemusage", 5)) {
	gc_mem_usage_control = atoi(word2);
	if (gc_mem_usage_control < 20) {
	    HTLog_errorN("gc-mem-usage must be >20; using 100; small values prevent gc from being smart; you suggested:",
			 gc_mem_usage_control);
	    gc_mem_usage_control = 100;
	}
	else CTRACE(stderr, "gc.......... Memory usage value set to %d\n",
		    gc_mem_usage_control);

    } else if (!strcmp(word1, "cachesize")) {
	if (unlimited(word2)) {
	    cache_max_k = 0;
	    CTRACE(stderr, "Cache....... no size limit\n");
	}
	else {
	    cache_max_k = atoi(word2) * 1024;
	    CTRACE(stderr,"Cache....... size %ldK\n", cache_max_k);
	}

    } else if (!strncmp(word1, "cachefiles", 9)) {
	/*
	 * Not documented nor implemented.
	 */
	if (unlimited(word2)) {
	    cache_max_f = 0;
	    CTRACE(stderr, "Cache....... no limit on the number of files\n");
	}
	else {
	    cache_max_f = atoi(word2);
	    CTRACE(stderr,"Cache....... max number of files %d\n",cache_max_f);
	}

    } else if (!strcmp(word1, "cachelimit1")) {
	cache_limit_1 = atoi(word2);
	CTRACE(stderr,"Cache....... files smaller than %dK equally valuable\n",
	       cache_limit_1);

    } else if (!strcmp(word1, "cachelimit2")) {
	cache_limit_2 = atoi(word2);
	CTRACE(stderr,"Cache....... files bigger than %dK will be discarded\n",
	       cache_limit_2);

    } else if (!strcmp(word1, "cacheclean")) {
	if (!word3 || !*word3) {	/* OLD STYLE */
	    CTRACE(stderr,
		   "OLD-STYLE... CacheClean directive, use new format\n");
	    if (parse_time(params, 24*3600, &cache_clean_def)) {
		TIME_TRACE("Cache....... remove files older than",
			   cache_clean_def);
	    } else HTLog_error2("Invalid cache_clean_def time spec:",params);
	}
	else {			/* NEW ENHANCED FORMAT WITH TEMPLATES */
	    CTRACE(stderr, "CacheClean.. ");
	    if (!cache_clean) cache_clean = HTList_new();
	    HTList_addObject(cache_clean, parse_tt(params));
	}

    } else if (!strncmp(word1, "cacheunused", 7)) {
	if (!word3 || !*word3) {	/* OLD STYLE */
	    CTRACE(stderr,
		   "OLD-STYLE... CacheUnused directive, use new format\n");
	    if (parse_time(params, 24*3600, &cache_unused_def)) {
		TIME_TRACE("Cache....... remove if unused", cache_unused_def);
	    } else HTLog_error2("Invalid cache_unused_def time spec:",params);
	}
	else {			/* NEW ENHANCED FORMAT WITH TEMPLATES */
	    CTRACE(stderr, "CacheUnUsed. ");
	    if (!cache_unused) cache_unused = HTList_new();
	    HTList_addObject(cache_unused, parse_tt(params));
	}

    } else if (!strncmp(word1, "cachedefaultexp", 15)) {
	if (!word3 || !*word3) {
	    CTRACE(stderr,
	   "OLD-STYLE... CacheDefaultExpiry directive, use new format\n");
	    if (parse_time(params, 24*3600, &cache_exp_def)) {
		TIME_TRACE("Cache....... default expiry after", cache_exp_def);
	    } else HTLog_error2("Invalid default expiry time spec:", params);
	}
	else {			/* NEW ENHANCED FORMAT WITH TEMPLATES */
	    CTRACE(stderr, "CacheDefaultExpiry ");
	    if (!cache_exp) cache_exp = HTList_new();
	    HTList_addObject(cache_exp, parse_tt(params));
	}

    } else if (!strncmp(word1, "cachelocktimeout", 10)) {
	if (parse_time(params, 1, &cache_lock_timeout)) {
	    TIME_TRACE("Cache....... lock timeout", cache_lock_timeout);
	} else HTLog_error2("Invalid cache lock timeout spec:", params);

    } else if (!strncmp(word1, "cachelastmodifiedfactor", 10)) {
	if (negative(word2)) {
	    cache_lm_factor = -1;
	    CTRACE(stderr, "Cache....... Last-Modified hack disabled\n");
	}
	else {
	    float f;
	    if (sscanf(word2, "%f", &f)) {
		cache_lm_factor = f;
		CTRACE(stderr,
		       "Cache....... Last-Modified hack factor set to %f\n",
		       cache_lm_factor);
	    }
	    else HTLog_error2("Invalid CacheLastModifiedFactor arg:",word2);
	}

    } else if (!strncmp(word1, "cachetimemargin", 12)) {
	if (parse_time(params, 1, &cache_time_margin)) {
	    TIME_TRACE("Cache....... time margin set to", cache_time_margin);
	} else HTLog_error2("Invalid cache time margin spec:", params);

    } else if (!strncmp(word1, "cacheexpirycheck", 14)) {
	cache_exp_check = positive(word2);
	CTRACE(stderr, "ExpiryCheck. %s\n", cache_exp_check ? "On" : "Off");

    } else if (!strncmp(word1, "cachenoconnect", 10)) {
	cache_no_connect = positive(word2);
	CTRACE(stderr, "%s\n",
	   cache_no_connect
	   ? "Standalone.. cache mode [no external document retrieves]"
	   : "Normal...... cache mode [docs not in cache will be retrieved]");

    } else if (strstr(word1, "proxy") || strstr(word1, "gateway")) {
	char * access = NULL;

	if      (!strcmp(word1,"httpproxy")   || !strncmp(word1,"wwwhttp",7))
	    access = "http";
	else if (!strcmp(word1,"gopherproxy") || !strncmp(word1,"wwwgopher",9))
	    access = "gopher";
	else if (!strcmp(word1,"ftpproxy")    || !strncmp(word1,"wwwftp",6))
	    access = "ftp";
	else if (!strcmp(word1,"waisproxy")   || !strncmp(word1,"wwwwais",7))
	    access = "wais";
	else if (!strcmp(word1,"newsproxy")   || !strncmp(word1,"wwwnews",7))
	    access = "news";

	if (access) {
	    char * envstr = (char *)malloc(strlen(word2) + 20);
	    char * envname = (char *)malloc(20);
	    char * old;

	    CTRACE(stderr,
		   "Proxy....... using another proxy %s for %s access\n",
		   word2,access);

	    sprintf(envname, "%s_proxy", access);
	    sprintf(envstr, "%s_proxy=%s", access, word2);

	    old = getenv(envname);
#ifndef NeXT
	    putenv(envstr);
#else
	    HTLog_error("xxx_proxy directive not supported on the NeXT");
	    HTLog_error("Use xxx_proxy environment variables instead");
	    fprintf(stderr,"xxx_proxy directive not supported on the NeXT");
	    fprintf(stderr,"Use xxx_proxy environment variables instead");
	    exit(0);
#endif
	    if (old)
		HTLog_error2("Environment variable overridden:", old);
	}
	else HTLog_error2("Unknown proxy directive:", word1);

    } else if (!strncmp(word1, "suffixcasesense", 10)) {
	HTSuffixCaseSense = positive(word2);

    } else if (!strncmp(word1, "rulecasesense", 10)) {
	/*
	 * Not documented, at least on Unix side.
	 * I dunno about VMS...
	 */
	HTRuleCaseSense = positive(word2);

    } else if (!strncmp(word1, "inputtimeout", 10) ||
	       !strncmp(word1, "outputtimeout", 11) ||
	       !strncmp(word1, "bodytimeout", 9)) {
	time_t n;

	if (!parse_time(params, 1, &n))
	    HTLog_error2("Invalid timeout specifier:", params);
	else {
	    if (word1[0] == 'i')
		HTInputTimeOut = n;
	    else if (word1[0] == 'o')
		HTOutputTimeOut = n;
	    else
		HTScriptTimeOut = n;
	    CTRACE(stderr, "TimeOut..... for %s set to %ld secs\n",
		   (word1[0]=='i' ? "input" :
		    (word1[0]=='o' ? "output" : "scripts")), n);
	}

    } else if (!strncmp(word1, "dir", 3)) {
	standardize(word2);
	if (!strncmp(word1, "diraccess", 6)) {
	    if (!strncmp(word2, "selective", 3))
	      	HTDirAccess = HT_DIR_SELECTIVE;
	    else if (positive(word2))
	      	HTDirAccess = HT_DIR_OK;
	    else if (negative(word2))
	      	HTDirAccess = HT_DIR_FORBID;
	    else {
		CTRACE(stderr, "Unknown..... DirAccess mode \"%s\"\n", word2);
		HTLog_error2("Unknown DirAccess mode:", word2);
	    }
	}
	else if (!strncmp(word1, "dirreadme", 7)) {
	    if (!strncmp(word2, "bottom", 3))
	      	HTDirReadme = HT_DIR_README_BOTTOM;
	    else if (!strncmp(word2, "top", 3))
	      	HTDirReadme = HT_DIR_README_TOP;
	    else if (negative(word2))
	      	HTDirReadme = HT_DIR_README_NONE;
	    else
		HTLog_error2("Unknown DirReadme mode:", word2);
	}
	else
	    HTLog_error2("Unknown Dir directive:", word1);

    } else if (!strcmp(word1, "setuid")) {
	/*
	 * Not documented because semantics is still vague...
	 */
	HTDoSetUid = positive(word2);
	CTRACE(stderr, "SetUid...... turned %s\n", HTDoSetUid ? "on" : "off");

    } else if (!strcmp(word1, "userid")) {
	StrAllocCopy(HTUserId, word2);
	CTRACE(stderr, "Default..... user id for httpd is %s\n", word2);

    } else if (!strcmp(word1, "groupid")) {
	StrAllocCopy(HTGroupId, word2);
	CTRACE(stderr, "Default..... group id for httpd is %s\n", word2);

    } else if (!strncmp(word1,"accesslog",9) || !strcmp(word1,"logfile")) {
	HTLog_accessOpen(word2);

    } else if (!strncmp(word1,"cacheaccesslog",12)) {
	StrAllocCopy(HTCacheAccessLog,word2);
	if (TRACE) {
	    if (strchr(word2, '/'))
		fprintf(stderr,
			"Cache log... all accesses are logged to %s\n",word2);
	    else
		fprintf(stderr,
		"Cache log... CacheRoot/{http,ftp,gopher}/machine.name/%s\n",
			word2);
	}

    } else if (!strcmp(word1, "errorlog")) {
	HTLog_errorOpen(word2);

    } else if (!strncmp(word1, "logformat", 7) ||
	       !strncmp(word1, "logfileformat", 8)) {
	standardize(word2);
	if (!strcmp(word2,"new") || !strcmp(word2,"common"))
	    HTNewLogFileFormat = YES;
	else if (!strcmp(word2,"old"))
	    HTNewLogFileFormat = NO;
	else
	    HTLog_error2("Unknown logfile format [use Old or Common]:", word2);

    } else if (!strncmp(word1, "logtime", 4)) {
	standardize(word2);
	if (!strcmp(word2,"gmt"))
	    HTUseGMT = YES;
	else if (!strncmp(word2, "localtime", 5))
	    HTUseGMT = NO;
	else
	    HTLog_error2("Unknown logtime [use GMT or LocalTime]:", word2);

    } else if (!strncmp(word1, "putscript", 9)) {
	StrAllocCopy(HTPutScript, word2);
	CTRACE(stderr,"PUT......... handled by script \"%s\"\n", word2);

    } else if (!strncmp(word1, "postscript", 10)) {
	StrAllocCopy(HTPostScript, word2);
	CTRACE(stderr,"POST........ handled by script \"%s\"\n", word2);

    } else if (!strncmp(word1, "userdir", 7)) {
	CTRACE(stderr,"UserDirs.... enabled [\"%s\" under each user's home]\n",
	       word2);
	StrAllocCopy(HTUserDir, word2);

    } else if (!strncmp(word1, "alwayswelcome", 9)) {
	always_welcome = positive(word2);
	CTRACE(stderr,
	       "Directory... names %sget redirected to welcome pages\n",
	       always_welcome ? "" : "don't ");

    } else if (!strncmp(word1, "welcome", 7)) {
	HTAddWelcome(word2);
	CTRACE(stderr,"Welcome pages are called \"%s\"\n", word2);

    } else if (!strncmp(word1, "htbin", 5) ||
	       !strncmp(word1, "bindir", 6)) {
	char *bindir = (char*)malloc(strlen(word2) + 3);
	if (!bindir) outofmem(__FILE__, "HTSetConfiguration");
	strcpy(bindir, word2);
	strcat(bindir, "/*");
	HTAddRule(HT_Exec, "/htbin/*", bindir);
	CTRACE(stderr,
	   "NOTE......... HTBin rule is obsolete -- rather use Exec rule\n");
	free(bindir);
	/*
	** Physical /htbin location -- this is almost obsolite
	** (only search may need it).
	*/
	StrAllocCopy(HTBinDir, word2);

    } else if (!strncmp(word1, "search", 6)) {
	if (strchr(word2, '/')) {
	    StrAllocCopy(HTSearchScript, word2); /* Full search script path */
	}
	else if (HTBinDir) {
	    if (!(HTSearchScript =
		  (char*)malloc(strlen(HTBinDir) + strlen(word2) + 2)))
		outofmem(__FILE__, "HTSetConfiguration");
	    strcpy(HTSearchScript, HTBinDir);
	    strcat(HTSearchScript, "/");
	    strcat(HTSearchScript, word2);
	}
	else {
	    CTRACE(stderr,
	   "WARNING..... Search directive without HTBin before ignored\n");
	    CTRACE(stderr,
	   "NOTE........ Search script should be specified as absolute path\n");
	}
	if (HTSearchScript)
	    CTRACE(stderr, "Search...... handled by \"%s\"\n", HTSearchScript);
	else
	    CTRACE(stderr, "Search...... not turned on\n");

    } else if (!strcmp(word1, "hostname") || !strcmp(word1, "servername")) {
	CTRACE(stderr, "HostName.... \"%s\" propagated to scripts\n", word2);
	StrAllocCopy(HTServerHostName, word2);

    } else if (!strcmp(word1, "port") ||
	       !strcmp(word1, "serverport") ||
	       !strcmp(word1, "defaultport")) {
	if (!HTPortAddr) {
	    HTPortAddr = (char*)malloc(strlen(word2) + 10);
	    if (!HTPortAddr) outofmem(__FILE__, "HTPortAddr");
	    sprintf(HTPortAddr, "*:%s", word2);
	    CTRACE(stderr, "Port........ %s\n", word2);
	}
	else CTRACE(stderr,
	"Warning..... Command line option overrides config port setting %s\n", 
		    word2);
    } else if (!strncmp(word1,"pidfile",7)) {
	StrAllocCopy(pid_file, word2);

    } else if (!strcmp(word1,"protection")) {
	CTRACE(stderr, "Protection.. definition for name \"%s\"\n", word2);
	if (word3 && strchr(word3,'{'))
	    HTAddNamedProtection(word2, HTAAProt_parseInlined(hack_conf_fp));
	else {
	    HTLog_error2("Syntax error in config file, expecting '{' to start protection definition:",word2);
	}

    } else {
	op =	!strcmp(word1, "map")		? HT_Map
	    :	!strcmp(word1, "pass")		? HT_Pass
	    :	!strcmp(word1, "fail")		? HT_Fail
	    :   !strcmp(word1, "exec")		? HT_Exec
	    :   !strncmp(word1, "redirect", 5)	? HT_Redirect
	    :   !strncmp(word1, "defprot", 7)	? HT_DefProt
	    :	!strncmp(word1, "protect", 7)	? HT_Protect
	    :					  HT_Invalid;
	if (op==HT_Invalid) {
	    HTLog_error2("Bad configuration directive:", config);
	} else if ((op==HT_Protect || op==HT_DefProt) && word3 &&
		   strchr(word3, '{')) {
	    HTAddProtRule(op, word2, HTAAProt_parseInlined(hack_conf_fp));
	} else {  
	    HTAddRule(op, word2, word3);
	} 
    }
    free(line);
    return 0;
}


/*	Load the rules from a file				HTLoadRules()
**	--------------------------
**
** On entry,
**	Rules can be in any state
** On exit,
**	Any existing rules will have been kept.
**	Any new rules will have been loaded.
**	Returns		0 if no error, 0 if error!
**
** Bugs:
**	The strings may not contain spaces.
*/

int HTLoadRules ARGS1(CONST char *, filename)
{
    FILE * fp = fopen(filename, "r");
    char line[LINE_LENGTH+1];

    if (!reloading) {
	if (!rule_files) rule_files = HTList_new();
	HTList_addObject(rule_files, (char*)filename);
    }
    if (!fp) {
        CTRACE(stderr, "ERROR....... Can't open config file \"%s\"\n", filename);
	return -1; /* File open error */
    }
    hack_conf_fp = fp;
    for(;;) {
	if (!fgets(line, LINE_LENGTH+1, fp)) break;	/* EOF or error */
	(void) HTSetConfiguration(line);
    }
    fclose(fp);
    hack_conf_fp = NULL;
    return 0;		/* No error or syntax errors ignored */
}


PUBLIC void HTReLoadRules NOARGS
{
    HTList * cur = rule_files;
    char * filename;

    CTRACE(stderr, "CLEARING.... rules\n");
    HTClearRules();

    reloading = YES;

    CTRACE(stderr, "RELOADING... rules\n");
    while ((filename = (char*)HTList_nextObject(cur)))
	HTLoadRules(filename);
}

