/*		TCP/IP based server for HyperText		HTDaemon.c
**		---------------------------------
**
**
** Compilation options:
**	RULES		If defined, use rule file and translation table
**	DIR_OPTIONS	If defined, -d options control directory access
**	ACCESS_AUTH	If defined, include access authorization code
**      POSIXWAIT       If defined, uses waitpid instead of wait3
**
**  Authors:
**	TBL	Tim Berners-Lee, CERN
**	JFG	Jean-Francois Groff, CERN
**	JS	Jonathan Streets, FNAL
**	AL	Ari Luotonen, CERN
**	MD	Mark Donszelmann, CERN
**
**  History:
**	   Sep 91  TBL 	Made from earlier daemon files. (TBL)
**	26 Feb 92  JFG	Bug fixes for Multinet.
**       8 Jun 92  TBL  Bug fix: Perform multiple reads in case we don't get
**                      the whole command first read.
**	25 Jun 92  JFG  Added DECNET option through TCP socket emulation.
**	 6 Jan 93  TBL  Plusses turned to spaces between keywords
**	 7 Jan 93  JS   Bug fix: addrlen had not been set for accept() call
**			Logging in GMT to file-YYMM in name
**	   Sep 93  AL   Added Access Authorization, and
**			some minor fixes to make server more tolerant of
**			missing stuff (NULLs here and there). It can be
**			expected that new browsers might screw up with
**			Access Authorization at first.
**	20 Oct 93  AL	Now makes sure that it is really an HTTP1
**			request and not just an HTTP0 request with
**			spaces in the URL.
**	 6 Nov 93  MD	Bug fix: gmtime replace by localtime on VMS.
**			Bug fix: changed vms into VMS (ifdef is case sensitive)
**			calls to strcasecmp were changed into strcasecomp 
**			(same for strncasecmp).
**			Added file sharing of the logfile (VMS).
**			Added switching off of SYSPRV after bind (VMS).
**			disabled Forking (VMS).
**	12 Nov 93  AL	Understands wildcards in Content-Type.
**			Handles SIGPIPE.
**	23 Nov 93  MD	Bug fix: Corrected timestamp if running in stand-alone
**			mode. Used to give same time after start.
**			Flushing of buffers now also flushes RMS buffer.
**	 2 Dec 93  MD 	Fixed reuse of ports with open connections (VMS)
**	11 Dec 93  AL	Added HTRequest, HTHandle() heavily rewritten.
**			Added also a new module, HTRequest.c, which
**			now needs to be linked together with daemon.
**	14 Dec 93  AL	Forking moved to the beginning of HTHandle().
**	22 Feb 94  MD	Changed for VMS.
**			Added user init routines
**	27 Mar 94  AL	Added SOCKS stuff by Ian Dunkin <imd1707@ggr.co.uk>
*/

/* (c) CERN WorldWideWeb project 1990-1992. See Copyright.html for details */


/*	Module parameters:
**	-----------------
**
**  These may be undefined and redefined by syspec.h
*/

#define FORKING

#define LISTEN_BACKLOG 2	/* Number of pending connect requests (TCP)*/
#define MAX_CHANNELS 20		/* Number of channels we will keep open */
#define WILDCARD '*'	    	/* Wildcard used in addressing */
#define FIRST_TCP_PORT	5000	/* When using dynamic allocation */
#define LAST_TCP_PORT 5999

#define MAX_LINE 512		/* HTTP request field line */

#ifndef RULE_FILE
#define RULE_FILE		"/etc/httpd.conf"
#endif

#ifndef DEFAULT_EXPORT
#define DEFAULT_EXPORT		"/Public"
#endif

#ifdef __osf__
#define _OSF_SOURCE
#define _BSD
#endif

#include "HTUtils.h"
#include "tcp.h"		/* The whole mess of include files */
#include "HTTCP.h"		/* Some utilities for TCP */
#include "HTFormat.h"
#include "HTInit.h"
#include "HTSUtils.h"

#ifdef VMS
#include "HTVMSUtils.h"
#undef FORKING
#endif /* VMS */

#ifdef RULES			/* Use rules? */
#include "HTRules.h"
#endif

#ifdef ACCESS_AUTH		/* Use Access Authorization? */
#include "HTWriter.h"
#include "HTAAServ.h"
#endif

#include "HTFile.h"
#include "HTParse.h"
#include "HTRequest.h"
#include "HTAuth.h"		/* HTUser */
#include "HTConfig.h"
#include "HTDaemon.h"
#include "HTLog.h"
#include "HTCache.h"
#include "HTUserInit.h"
#include "HTFWriter.h"

/* Forking */
#ifdef FORKING
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <sys/param.h>
#include <errno.h>

#ifndef SIGCLD
#ifdef SIGCHLD
#define SIGCLD SIGCHLD
#endif
#endif

#ifdef SIGTSTP	/* BSD */
#include <sys/file.h>
#include <sys/ioctl.h>
#endif /* BSD */

#endif /* FORKING */


#ifndef VMS
#ifdef SUN4
#include <sys/timeb.h>
#endif

#include <pwd.h>	/* Unix password file routine: getpwnam() */
#include <grp.h>	/* Unix group file routine: getgrnam()	  */
#endif /* not VMS */

#ifdef __osf__
#include <sys/wait.h>
#include <sys/resource.h>
#endif



#ifdef VMS
#define gc()
/* dummy declaration to make sure that HTML.OBJ is not included
   and other references (used in clients only) are resolved */
#ifdef __DECC
#include "HText.h"
PUBLIC CONST HTStructuredClass HTMLPresentation;
PUBLIC int HTDiag;
PUBLIC HText * HTMainText;
PUBLIC BOOL interactive;
#endif /* DECC */
#endif /* VMS */

/*	Module-Global Variables
**	-----------------------
*/
PRIVATE enum role_enum {master, slave, transient, passive} role;
PRIVATE SockA	server_soc_addr;
PRIVATE SockA	client_soc_addr;
PRIVATE int	master_soc;	/* inet socket number to listen on */
PRIVATE int	com_soc;	/* inet socket number to read on */
#ifdef SELECT
PRIVATE fd_set	open_sockets;	/* Mask of channels which are active */
PRIVATE int	num_sockets;	/* Number of sockets to scan */
#endif
PRIVATE BOOLEAN	dynamic_allocation;	/* Search for port? */

#ifdef FORKING
PRIVATE BOOL	HTForkEnabled = NO; /* Do we fork() when serving.	*/
                                    /* This is enabled when server is	*/
                                    /* standalone (-a or -p option).	*/
#endif /* FORKING */

/*	Program-Global Variables
**	------------------------
*/
PUBLIC char * HTAppName = "CERN-HTTPD";	/* Application name		*/
PUBLIC char * HTAppVersion = VD; 	/* Application version		*/
PUBLIC char * log_file_name = NULL;	/* Log file name (WAIS code)	*/
PUBLIC char * HTClientProtocol = 0;	/* Protocol and version number	*/
PUBLIC char * HTServerProtocol = "HTTP/1.0";
extern char * HTClientHost;		/* Clients IP address		*/
PUBLIC char * HTClientHostName = NULL;	/* Clients host name		*/
PUBLIC int HTServerPort = 80;		/* The portnumber of this server*/
PUBLIC int child_serial = 0;		/* Serial number of child	*/

PUBLIC BOOL ignore_sigpipes = NO;	/* Should we ignore SIGPIPE	*/
PUBLIC BOOL sigpipe_caught = NO;	/* So have we caught a SIGPIPE	*/

char * rfc931 PARAMS((struct sockaddr_in *	rmt_sin,
		      struct sockaddr_in *	our_sin));


/*	Server environment when handling requests
**	-----------------------------------------
*/
PUBLIC int		HTStatusCode		= 0;
PUBLIC char *		HTReasonLine		= NULL;
PUBLIC int		HTSoc			= 0;
PUBLIC char *		HTReqLine		= NULL;
PUBLIC char *		HTReqArg		= NULL;
PUBLIC char *		HTReqArgPath		= NULL;
PUBLIC char *		HTReqArgKeywords	= NULL;
PUBLIC char *		HTReqTranslated		= NULL;
PUBLIC char *		HTReqScript		= NULL;
PUBLIC char *		HTScriptPathInfo	= NULL;
PUBLIC char *		HTScriptPathTrans	= NULL;
PUBLIC char *		HTLocation		= NULL;
PUBLIC char *		HTLastModified		= NULL;
PUBLIC char *		HTExpires		= NULL;
PUBLIC char *		HTUri			= NULL;
PUBLIC char *		HTMessageId		= NULL;
PUBLIC HTList *		HTAllowed		= NULL;
PUBLIC HTList *		HTPublic		= NULL;
PUBLIC char *		HTMetaFile		= NULL;
PUBLIC char *		HTAuthString		= NULL;
PUBLIC char *		HTWWWAuthenticate	= NULL;
PUBLIC HTAAFailReason	HTReason		= HTAA_OK;
PUBLIC char *		HTUserAgent		= NULL;
PUBLIC char *		HTReferer		= NULL;
PUBLIC time_t		HTIfModifiedSince	= 0;

PUBLIC char *		remote_ident		= NULL;

PUBLIC long		HTCacheUsage		= 0;
PUBLIC int		HTChildExitStatus	= 0;
PUBLIC long		HTCachedBytes		= 0;

PUBLIC char *		gc_info_file		= NULL;
PUBLIC int		gc_pid			= 0;
PUBLIC time_t		gc_last_time		= 0;
PUBLIC int		gc_reqs_since_last	= 0;

PUBLIC int		bytes_from_cache	= 0;

PUBLIC time_t		cur_time		= 0;

PUBLIC BOOL		WWW_VeryVerboseFlag	= NO;


PUBLIC void reset_server_env NOARGS
{
    FREE(HTReqLine);
    FREE(HTReqArg);
    FREE(HTReqArgPath);
    FREE(HTReqTranslated);
    FREE(HTReqScript);
    FREE(HTScriptPathTrans);
    FREE(HTLocation);
    FREE(HTLastModified);
    FREE(HTExpires);
    FREE(HTUri);
    FREE(HTMessageId);
    FREE(HTMetaFile);
    FREE(HTAuthString);
    FREE(HTWWWAuthenticate);
    FREE(HTProxyHeaders);
    FREE(HTUserAgent);
    FREE(HTReferer);
    HTIfModifiedSince = 0;
    if (HTAllowed) {
	HTList_delete(HTAllowed);
	HTAllowed = NULL;
    }
    if (HTPublic) {
	HTList_delete(HTPublic);
	HTPublic = NULL;
    }

    HTStatusCode = 0;
    HTReasonLine = NULL;
    HTSoc = 0;
    HTReqArgKeywords = NULL;
    HTScriptPathInfo = NULL;
    HTReason = HTAA_OK;
    HTProxyHeaders = NULL;

    HTUser = NULL;
    HTProt = NULL;
    HTDefProt = NULL;

    HTChildExitStatus = 0;
    HTCachedBytes = 0;
}


PUBLIC void compute_server_env NOARGS
{
    char *keywords;

    FREE(HTReqArgPath);
    FREE(HTReqArgKeywords);

    FREE(HTReqTranslated);
    FREE(HTReqScript);
    FREE(HTScriptPathInfo);
    FREE(HTScriptPathTrans);
    FREE(HTLocation);

    if (HTReqArg) {
	int len;

	if ((keywords = strchr(HTReqArg, '?'))) {
	    *keywords = 0;		/* Temporarily truncate */
	    StrAllocCopy(HTReqArgPath, HTReqArg);
	    StrAllocCopy(HTReqArgKeywords, keywords+1);
	    *keywords = '?';		/* Reconstruct */
	}
	else StrAllocCopy(HTReqArgPath, HTReqArg);

	HTUnEscape(HTReqArgPath);
	HTSimplify(HTReqArgPath);	/* Remove ".." etc. */

	/*
	** HTSimplify will leave in a "/../" at the top, which can
	** be a security hole.
	*/
	len = strlen(HTReqArgPath);
	if (strstr(HTReqArgPath, "/../") ||
	    !strncmp(HTReqArgPath,"/..",3) ||
	    !strncmp(HTReqArgPath,"..",2) ||
	    (len >= 3  &&  !strcmp(&HTReqArgPath[len-3], "/.."))) {
	    CTRACE(stderr, "Illegal..... attempt to use /../ (\"%s\")\n",
		   HTReqArg);
	    HTReason = HTAA_DOTDOT;
	}
    }
}






#if defined(__hpux) || defined(POSIXWAIT)
PRIVATE pid_t	pgrp;
#else
PRIVATE int	pgrp;
#endif


extern HTAAProt * HTProt;
extern HTAAProt * HTDefProt;
extern char *	HTProxyCacheRoot;
extern HTList *	HTNoLog;	/* Machines for which log entry not made */
extern BOOL	HTDisabled[(int)MAX_METHODS + 1];	/* Disabled methods */
extern BOOL	HTImProxy;


PRIVATE void read_gc_results NOARGS
{
    FILE * fp;
    char buf[MAX_LINE];

    if (!HTProxyCacheRoot) {
	HTLog_error("BUG: gc returned yet we're not caching -- weird");
	return;
    }
    if (!gc_info_file) {
	HTLog_error2("GC info file not set although caching to",
		     HTProxyCacheRoot);
	return;
    }
    
    fp = fopen(gc_info_file, "r");
    if (!fp) {
	HTLog_error2("Can't open gc report file:", gc_info_file);
	return;
    }
    if (fgets(buf, MAX_LINE, fp)) {
	int pid;
	long bytes_used = 0;
	long bytes_freed = 0;
	int files_used = 0;
	int files_freed = 0;
	long corrected = 0;

	sscanf(buf, "%d %ld %ld %d %d",
	       &pid, &bytes_used, &bytes_freed, &files_used, &files_freed);

	if (gc_pid  &&  pid != gc_pid) {
	    HTLog_error(
	"WARNING: gc report has wrong process id - real gc may have crashed");
	}

	corrected = HTCacheUsage - bytes_freed/1024;
	if (TRACE) {
	    fprintf(stderr, "GC REPORT... from pid %d\n", pid);
	    fprintf(stderr, "GC REPORT... %ldK USED (%d files)\n",
		    bytes_used/1024, files_used);
	    fprintf(stderr, "GC REPORT... %ldK COLLECTED (%d files)\n",
		    bytes_freed/1024, files_freed);
	    fprintf(stderr, "PARENT FIX.. %ldK USED => %ldK USED\n",
		    HTCacheUsage, corrected);
	}
	if (corrected < 0 ||
	    corrected < bytes_used/1140 ||
	    corrected > bytes_used/930) {
	    CTRACE(stderr, "PARENT...... Correcting %ldK => %ldK USED\n",
		   corrected, bytes_used/1024);
	    corrected = bytes_used/1024;
	    if (corrected<0) corrected = 0;
	}
	HTCacheUsage = corrected;
	CTRACE(stderr,"PARENT...... %ldK of cache in use\n",HTCacheUsage);
    }
    else HTLog_error2("gc report file is empty:", gc_info_file);
    fclose(fp);
}


#ifdef FORKING
/* PRIVATE							sig_child()
**
** This function is taken from:
**	W.Richard Stevens: UNIX Network Programming,
**	Prentice Hall 1990, page 82.
**
** This is a 4.3BSD SIGCLD signal handler that can be used by a
** server that's not interested in its child's exit status, but needs
** to wait for them , to avoid clogging up the system with zombies.
**
** Beware that the calling process may get an interrupted system
** call when we return, so they had better handle that.
*/
PRIVATE int sig_child NOARGS
{
#ifdef SIGTSTP	/* BSD */
#if defined(POSIXWAIT) || defined(__hpux)
    pid_t pid;
    int status;

    while ((pid = waitpid((pid_t)-1, &status, WNOHANG)) > 0) {
	if (pid == gc_pid) {
	    CTRACE(stderr, "Parent...... gc has completed\n");
	    read_gc_results();
	    gc_pid = 0;
	}
	if (HTProxyCacheRoot  &&  WIFEXITED(status)) {
	    int child_usage = HTExitStatusToKilos(WEXITSTATUS(status));
	    HTCacheUsage += child_usage;
	    CTRACE(stderr,
		   "Parent...... %dK taken up by child, total %dK used\n",
		   child_usage, HTCacheUsage);
	}
    }
#else
    int		pid;
    union wait	status;

    while ((pid = wait3(&status, WNOHANG, (struct rusage*)NULL)) > 0) {
	if (pid == gc_pid) {
	    CTRACE(stderr, "Parent...... gc has completed\n");
	    read_gc_results();
	    gc_pid = 0;
	}
	if (HTProxyCacheRoot  &&  WIFEXITED(status)) {
	    int child_usage = HTExitStatusToKilos(WEXITSTATUS(status));
	    HTCacheUsage += child_usage;
	    CTRACE(stderr,
		   "Parent...... %dK taken up by child, total %ldK used\n",
		   child_usage, HTCacheUsage);
	}
    }
#endif
#endif

    return 0;	/* Just to avoid a compiler warning */
}
#endif /* FORKING */


#ifndef VMS
PRIVATE void sig_term NOARGS
{
    if (HTStandAlone) {
#if defined(__svr4__) || defined(_POSIX_SOURCE) || defined(__hpux)
	kill(-pgrp, SIGKILL);
#else
	killpg(pgrp, SIGKILL);
#endif
	shutdown(master_soc, 2);
	close(master_soc);
    }
}


PRIVATE void sig_segv NOARGS
{
    timeout_off();	/* So timeouts don't go to parent */
    HTLog_error("CRASHED (segmentation fault, core dumped)");
    if (HTReqLine)
	HTLog_error2("Crashing request was:", HTReqLine);
    if (server_root)
	chdir(server_root);
    abort();
}

PRIVATE void sig_bus NOARGS
{
    timeout_off();	/* So timeouts don't go to parent */
    HTLog_error("CRASHED (bus error, core dumped)");
    if (HTReqLine)
	HTLog_error2("Crashing request was:", HTReqLine);
    if (server_root)
	chdir(server_root);
    abort();
}


PRIVATE void set_signals NOPARAMS;

PRIVATE void sig_hup NOARGS
{
    HTLog_error("SIGHUP caught -- reloading configuration files\n");
    HTReLoadRules();
    set_signals();
}

PRIVATE void sig_pipe NOARGS
{
    if (!sigpipe_caught) {
	if (HTReqLine)
	    HTLog_error2("Connection interrupted [SIGPIPE], req:",HTReqLine);
	else
	    HTLog_error("Connection interrupted [SIGPIPE]");
    }

    if (ignore_sigpipes) {
	CTRACE(stderr, "SIGPIPE..... ignored (writing to cache)\n");
	sigpipe_caught = YES;
	set_signals();
    }
    else {
	timeout_off();	/* So timeouts don't go to parent */
	CTRACE(stderr, "SIGPIPE..... caught, exiting...\n");
	exit(0);
    }
}

#endif /* not VMS */


PRIVATE void set_signals NOARGS
{
#ifndef VMS
    signal(SIGTERM, (void (*)())sig_term);
    signal(SIGHUP, (void (*)())sig_hup);
    signal(SIGSEGV, (void (*)())sig_segv);
    signal(SIGBUS, (void (*)())sig_bus);

    /*
    ** Ignore the terminal stop signals (BSD)
    */
#ifdef SIGTTOU
    signal(SIGTTOU, SIG_IGN);
#endif
#ifdef SIGTTIN
    signal(SIGTTIN, SIG_IGN);
#endif
#ifdef SIGTSTP
    signal(SIGTSTP, SIG_IGN);
#endif

    /*
    ** Parent process is not interested in child exit statuses.
    ** We have to prevent the children from becoming zombies.
    ** In System V we just ignore SIGCLD signal.
    ** In BSD we should handle SIGCLD signal and call wait3().
    ** However, because we don't want to handle interrupted
    ** system calls on BSD, we just call sig_child() after every
    ** fork() to check if any previous childs have finished.
    */
#if 0
#ifndef SIGTSTP	/* SysV */
    signal(SIGCLD, SIG_IGN);
#endif
#endif

    signal(SIGCLD, (void (*)())sig_child);

    /*
    ** Ignore SIGPIPE signals so we don't die if client
    ** suddenly closes connection and we keep on writing
    ** to the socket.
    */
#if 0
    signal(SIGPIPE, SIG_IGN);
#else
    signal(SIGPIPE, (void (*)())sig_pipe);
#endif
#endif /* not VMS */
}


/*
 *	Return time in 100ths of seconds to the next daily gc.
 */
PRIVATE int get_next_timeout NOARGS
{
    struct tm * t;
    int secs;
    int timeout;

    time(&cur_time);
    t = localtime(&cur_time);
    secs = t->tm_hour*3600 + t->tm_min*60 + t->tm_sec;

    if (!gc_disabled) {
	if (!HTProxyCacheRoot) {
	    CTRACE(stderr, "Disabling... gc [caching not enabled either]\n");
	    gc_disabled = YES;
	}
	else if (cache_no_connect) {
	    CTRACE(stderr, "Disabling... gc [standalone cache mode]\n");
	}
    }

    if (gc_disabled) {
	CTRACE(stderr, "Gc.......... disabled altogether\n");
	return -1;
    }
    if (!gc_daily_gc) {
	CTRACE(stderr, "Daily gc.... not turned on\n");
	return -1;
    }
    if (gc_pid) {
	CTRACE(stderr, "Daily gc.... Another gc has not yet finished\n");
	return -1;
    }

    timeout = gc_daily_gc - secs;
    if (timeout < 0) {
	if (cur_time - gc_last_time < 3600) {
	    CTRACE(stderr,
		   "Daily gc.... Another gc done less than an hour ago\n");
	    return 100 * (timeout + 24*3600);
	}
	else return 0;
    }
    else return 100 * timeout;
}



/*
 *	Find out if it's time to do garbage collection.
 */
PRIVATE BOOL time_to_do_gc NOARGS
{
    BOOL flag = NO;

    if (!gc_disabled) {
	if (!HTProxyCacheRoot) {
	    CTRACE(stderr, "Disabling... gc [caching not enabled either]\n");
	    gc_disabled = YES;
	}
	else if (cache_no_connect) {
	    CTRACE(stderr, "Disabling... gc [standalone cache mode]\n");
	}
    }

    gc_reqs_since_last++;

    if (gc_disabled) {
	CTRACE(stderr, "Cache....... I never do gc\n");
	return NO;
    }

    if (cache_max_k > 0  &&  HTCacheUsage >= cache_max_k) {
	CTRACE(stderr, "Cache....... limit reached: %ldK used (max %ldK)\n",
	       HTCacheUsage, cache_max_k);
	flag = YES;
    }

    if (gc_req_interval && gc_reqs_since_last >= gc_req_interval) {
	CTRACE(stderr,
	       "Cache....... Child number %d will do gc (every %d does)\n",
	       child_serial, gc_req_interval);
	flag = YES;
    }

    if (gc_time_interval && cur_time-gc_last_time >= gc_time_interval) {
	CTRACE(stderr, "Cache....... Time for gc, last gc done at %s",
	       ctime(&gc_last_time));
	flag = YES;
    }

    if (cur_time - gc_last_time >= 600 &&
	get_next_timeout() >= 0 && get_next_timeout() < 120) {
	CTRACE(stderr, "Cache....... Time for daily gc\n");
	flag = YES;
    }

    if (flag) {
	gc_reqs_since_last = 0;
	gc_last_time = cur_time;
    }
    return flag;
}




/*	Send a string down a socket
**	---------------------------
**
**	The trailing zero is not sent.
**	There is a maximum string length.
*/
PUBLIC int HTWriteASCII ARGS2(int, soc, char *, s)
{
#ifdef NOT_ASCII
    char * p;
    char ascii[255];
    char *q = ascii;
    for (p=s; *p; p++) {
        *q++ = TOASCII(*p);
    }
    return NETWRITE(soc, ascii, p-s);
#else
    return NETWRITE(soc, s, strlen(s));
#endif
}



/*				MIME OUTPUT
**				===========
*/

struct _HTStream {
	HTStreamClass *		isa;
	HTStreamClass		class;		/* special per object */
};


/*			Catch error messages
**			--------------------
**
**	This shouldn't be necessary most of the time as HTLoadError
**	will be called.
**
**	These entry points suppress the loading of HTAlert from the WWW library.
**	These could be cleaned up and made very useful, esp
**	remote progress reporting...
*/

PUBLIC void HTAlert ARGS1(CONST char *, Msg)
{
    fprintf(stderr, "500   Server reports:  %s\r\n", Msg);
}


PUBLIC void HTProgress ARGS1(CONST char *, Msg)
{
    /* fprintf(stderr, "   %s ...\n", Msg); */
}


PUBLIC BOOL HTConfirm ARGS1(CONST char *, Msg)
{
    return(NO);
}

/*	Prompt for answer and get text back
*/
PUBLIC char * HTPrompt ARGS2(CONST char *, Msg, CONST char *, deflt)
{
    char * rep = 0;
    StrAllocCopy(rep, deflt);
    return rep;
}

PUBLIC char * HTPromptPassword ARGS1(CONST char *, Msg)
{
    char * rep = NULL;
    StrAllocCopy(rep, "");
    return rep;
}

PUBLIC void HTPromptUsernameAndPassword ARGS3(CONST char *,	Msg,
					      char **,		username,
					      char **,		password)
{
    *username = HTPrompt("Username: ", *username);
    *password = HTPromptPassword("Password: ");
}


/*	Write Error Message
**	-------------------
**
**
*/
PUBLIC int HTLoadError ARGS3(HTRequest *,	req,
			     int,		num,
			     CONST char *,	msg)
{
    CTRACE(stderr, "Returning... ERROR %d:\n ** %s\n", num,
	   msg ? msg : "-no-message-");

    if (num >= 500 || num==400 || num <= 0) {  /* To make error logs smaller */
	char * buf = (char*)malloc((HTReqLine ? strlen(HTReqLine) : 0) + 20);
	if (!buf) outofmem(__FILE__, "HTLoadError");

	if (HTReqLine)
	    sprintf(buf, "(%d) \"%s\"", num, HTReqLine);
	else
	    sprintf(buf, "(%d)", num);
	HTLog_error2(msg, buf);
	free(buf);
    }
    else HTLog_error2(HTReqArg,NULL); /* Other info goes there automatically */

    if (req && req->output_stream) {
        HTStream * sink = req->output_stream;
	char * buffer = (char*)malloc(strlen(msg?msg:"") + 400);

	if (!buffer) outofmem(__FILE__, "HTLoadError");

	req->content_type = HTAtom_for("text/html");
	req->content_length = msg ? strlen(msg) : 0;
	HTStatusCode = num;
	HTReasonLine = (char*)msg;
	req->content_encoding = NULL;
	req->content_language = NULL;

	if (HTClientProtocol) {
	    char * headers;

	    if (num == 401)
		HTWWWAuthenticate = HTAA_composeAuthHeaders(req);

	    headers = HTReplyHeaders(req);
	    if (headers) {
		(*sink->isa->put_string)(sink, headers);
		free(headers);
	    }
	}
	sprintf(buffer,
		"<BODY><H1>Error %d</H1>\r\n\r\n%s\n%s%s%s\r\n</BODY>\r\n",
		num, (msg ? msg : ""),
		"<P><HR><ADDRESS><A HREF=\"http://info.cern.ch/hypertext/WWW/Daemon/User/Guide.html\">CERN httpd ", HTAppVersion, "</A></ADDRESS>");
	(*sink->isa->put_string)(sink, buffer);
	(*sink->isa->free)(sink);
	free(buffer);
    }
    return HT_LOADED;
}


PUBLIC int HTLoadStrToStream ARGS2(HTStream *,	  sink,
				   CONST char *,  str)
{
    if (!sink || !sink->isa || !str) return -1;
    (*sink->isa->put_string)(sink, str);
    return 0;
}


PUBLIC int HTCloseStream ARGS1(HTStream *, sink)
{
    if (sink && sink->isa) {
	(*sink->isa->free)(sink);
	return 0;
    }
    return -1;
}



/*____________________________________________________________________
**
**			Networking code
*/

/*		Bind to a TCP port
**		------------------
**
** On entry,
**	tsap	is a string explaining where to take data from.
**		"" 	means data is taken from stdin.
**		"*:1729" means "listen to anyone on port 1729"
**
** On exit,
**	returns		Negative value if error.
*/
int do_bind ARGS1(CONST char *, tsap)
{
#ifdef SELECT
    FD_ZERO(&open_sockets);	/* Clear our record of open sockets */
    num_sockets = 0;
#endif

/*  Deal with PASSIVE socket:
**
**	A passive TSAP is one which has been created by the inet daemon.
**	It is indicated by a void TSAP name.  In this case, the inet
**	daemon has started this process and given it, as stdin, the connection
**	which it is to use.
*/
    if (*tsap == 0) {			/* void tsap => passive */

	dynamic_allocation = FALSE;		/* not dynamically allocated */
	role = passive;	/* Passive: started by daemon */

#ifdef VMS

	{   unsigned short channel;	    /* VMS I/O channel */
	    struct string_descriptor {	    /* This is NOT a proper descriptor*/
		    int size;		    /*  but it will work.	      */
		    char *ptr;		    /* Should be word,byte,byte,long  */
	    } sys_input = {10, "SYS$INPUT:"};
	    int	status;		    /* Returned status of assign */
	    extern int sys$assign();

	    status = sys$assign(&sys_input, &channel, 0, 0);
	    com_soc = channel;	/* The channel is stdin */
	    CTRACE(tfp, "IP.......... Opened PASSIVE socket %d\n", channel);
	    return 1 - (status&1);
	}	
#else
	com_soc = 0;	    /* The channel is stdin */
	CTRACE(tfp, "IP.......... PASSIVE socket 0 assumed from inet daemon\n");
	return 0;		/* Good */
#endif

/*  Parse the name (if not PASSIVE)
*/
    } else {				/* Non-void TSAP */
	char *p;		/* pointer to string */
	char *q;
	struct hostent  *phost;	    /* Pointer to host - See netdb.h */
	char buffer[256];		/* One we can play with */
	register SockA * sin = &server_soc_addr;

	strcpy(buffer, tsap);
	p = buffer;

/*  Set up defaults:
*/
#ifdef DECNET
	sin->sdn_family = AF_DECnet;	    /* Family = DECnet, host order  */
	sin->sdn_objnum = 0;                /* Default: new object number, */
#else  /* Internet */
	sin->sin_family = AF_INET;	    /* Family = internet, host order  */
	sin->sin_port = 0;		    /* Default: new port,    */
#endif
	dynamic_allocation = TRUE;	    /*  dynamically allocated */
	role = passive; 		    /*  by default */

	/*  Check for special characters: */
	if (*p == WILDCARD) {		/* Any node */
	    role = master;
	    p++;
	}

	/*  Strip off trailing port number if any: */
	for(q=p; *q; q++)
	    if (*q==':') {
	        int status = 0;
		*q++ = 0;		/* Terminate node string */
#ifdef DECNET
		sin->sdn_objnum =
		    (unsigned char)HTCardinal(&status, &q, (unsigned int)65535);
#else
		sin->sin_port =
		    htons((unsigned short)HTCardinal(&status, &q,
						     (unsigned int)65535));
		if (status<0) return status;
#endif
		if (*q) return -2;  /* Junk follows port number */
		dynamic_allocation = FALSE;
		break;	    /* Exit from loop before we skip the zero */
	    } /*if*/

	/* Get node name: */
#ifdef DECNET  /* Empty address (don't care about the command) */
	sin->sdn_add.a_addr[0] = 0;
	sin->sdn_add.a_addr[1] = 0;
	CTRACE(tfp, "Daemon...... Parsed address as port %d, DECnet %d.%d\n",
	       (int) sin->sdn_objnum,
	       (int) sin->sdn_add.a_addr[0],
	       (int) sin->sdn_add.a_addr[1] ) ;
#else
	if (*p == 0) {
	    sin->sin_addr.s_addr = INADDR_ANY; /* Default: any address */

	} else if (*p>='0' && *p<='9') {   /* Numeric node address: */
	    sin->sin_addr.s_addr = inet_addr(p); /* See arpa/inet.h */

	} else {		    /* Alphanumeric node name: */
	    phost=gethostbyname(p);	/* See netdb.h */
	    if (!phost) {
		CTRACE(tfp,
		       "IP.......... Can't find internet node name \"%s\"\n",p);
		return HTInetStatus("gethostbyname");  /* Fail? */
	    }
	    memcpy(&sin->sin_addr, phost->h_addr, phost->h_length);
	}
	CTRACE(tfp, 
	       "Daemon...... Parsed address as port %d, inet %d.%d.%d.%d\n",
	       (int)ntohs(sin->sin_port),
	       (int)*((unsigned char *)(&sin->sin_addr)+0),
	       (int)*((unsigned char *)(&sin->sin_addr)+1),
	       (int)*((unsigned char *)(&sin->sin_addr)+2),
	       (int)*((unsigned char *)(&sin->sin_addr)+3));
	HTServerPort = (int)ntohs(sin->sin_port);
#endif
    } /* scope of p */


    /*  Master socket for server: */
    if (role == master) {
	int one=1;
#ifdef STDC
	struct linger Linger = {0,0};
#else
	struct linger Linger;
	Linger.l_onoff = 0;
	Linger.l_linger = 0;
#endif

	/*  Create internet socket */
#ifdef DECNET
	master_soc = socket(AF_DECnet, SOCK_STREAM, 0);
#else
	master_soc = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
#endif
	if (master_soc<0) {
	    fprintf(stderr, "Couldn't create socket\n");
	    HTLog_error("Couldn't create socket");
	    return HTInetStatus("socket");
	}
	CTRACE(tfp, "IP.......... Opened socket number %d\n", master_soc);


/* #ifndef __svr4__ */
	if ((setsockopt(master_soc,SOL_SOCKET,SO_REUSEADDR,
			(char*)&one,sizeof(one)))
	    == -1) {
	    fprintf(stderr, "Couldn't set socket option\n");
	    HTLog_error("Couldn't set socket option");
	}
/* #endif */

/*
 * Trying again without these on Unix...
 */
#ifdef VMS
	if ((setsockopt(master_soc,SOL_SOCKET,SO_KEEPALIVE,
			(char*)&one,sizeof(one)))
	    == -1) {
	    fprintf(stderr, "Couldn't set socket option\n");
	    HTLog_error("Couldn't set socket option");
	}
	if ((setsockopt(master_soc,SOL_SOCKET,SO_LINGER,
			(char*)&Linger,sizeof(Linger)))
	    == -1) {
	    fprintf(stderr, "Couldn't set socket option\n");
	    HTLog_error("Couldn't set socket option");
	}
#endif /* VMS */


	/*  If the port number was not specified look for a free one */
#ifndef DECNET  /* irrelevant: no inetd */
	if (dynamic_allocation) {
	    unsigned short try;
	    for (try=FIRST_TCP_PORT; try<=LAST_TCP_PORT; try++) { 
		server_soc_addr.sin_port = htons(try);
		if (bind(master_soc,
			 (struct sockaddr*)&server_soc_addr,
			 sizeof(server_soc_addr)) == 0)
		    break;
		if (try == LAST_TCP_PORT)
		    return HTInetStatus("bind");
	    }
	    CTRACE(tfp, "IP..........  Bound to port %d.\n",
		   ntohs(server_soc_addr.sin_port));
	} else
#endif
	{					/* Port was specified */
	    if (bind(master_soc,
		     (struct sockaddr*)&server_soc_addr,
		     sizeof(server_soc_addr))<0)
		return HTInetStatus("bind");
	}
	if (listen(master_soc, LISTEN_BACKLOG)<0)
	    return HTInetStatus("listen");

	CTRACE(tfp, "Daemon...... Master socket(), bind() and listen() all OK\n");
#ifdef SELECT
	FD_SET(master_soc, &open_sockets);
	if ((master_soc+1) > num_sockets) num_sockets=master_soc+1;
#endif
	return master_soc;
    } /* if master */
    
    return -1;		/* unimplemented role */

} /* do_bind */



PRIVATE void input_timed_out NOARGS
{
    HTLog_error("Timed out when reading request");
    exit(0);
}

PRIVATE void output_timed_out NOARGS
{
    HTLog_error("Timed out while sending response");
    exit(0);
}

PRIVATE void script_timed_out NOARGS
{
    HTLog_error("Script timed out");
    exit(0);
}

PRIVATE void cleanup_timed_out NOARGS
{
    CTRACE(stderr,"TIMED OUT... after serving [probably on log file write]\n");
    /* No error log entry since that would probably block, too */
    exit(0);
}

PRIVATE void input_timeout_on NOARGS
{
#ifdef VMS
    HTVMS_start_timer(0, input_timed_out, HTInputTimeOut);
#else /* not VMS */
    signal(SIGALRM, (void (*)())input_timed_out);
    alarm(HTInputTimeOut);
#endif /* not VMS */
}

PUBLIC void output_timeout_on NOARGS
{
#ifdef VMS
    HTVMS_start_timer(0, output_timed_out, HTOutputTimeOut);
#else /* not VMS */
    signal(SIGALRM, (void (*)())output_timed_out);
    alarm(HTOutputTimeOut);
#endif /* not VMS */
}

PUBLIC void script_timeout_on NOARGS
{
#ifdef VMS
    HTVMS_start_timer(0, script_timed_out, HTScriptTimeOut);
#else /* not VMS */
    signal(SIGALRM, (void (*)())script_timed_out);
    alarm(HTScriptTimeOut);
#endif /* not VMS */
}

PUBLIC void cleanup_timeout_on NOARGS
{
#ifdef VMS
    HTVMS_start_timer(0, cleanup_timed_out, DEFAULT_CLEANUP_TIMEOUT);
#else /* not VMS */
    signal(SIGALRM, (void (*)())cleanup_timed_out);
    alarm(DEFAULT_CLEANUP_TIMEOUT);
#endif /* not VMS */
}

PUBLIC void timeout_off NOARGS
{
#ifdef VMS
    HTVMS_cancel_timer(0);
#else /* not VMS */
    alarm(0);
    signal(SIGALRM, SIG_IGN);
#endif /* not VMS */
}



/*	Handle one message
**	------------------
**
** On entry,
**	soc		A file descriptor for input and output.
** On exit,
**	returns		>0	Channel is still open.
**			0	End of file was found, please close file
**			<0	Error found, please close file.
*/
PUBLIC int HTHandle ARGS1(int, soc)
{
    HTRequest *req;
    int status;
    int redir_hops = 0;
    HTInputSocket * isoc;
    BOOL already_logged = NO;

    reset_server_env();

    isoc = HTInputSocket_new(soc);

    input_timeout_on();
    req = HTParseRequest(isoc);
    timeout_off();

    if (!req) {
	HTLog_error("Request parsing failed fatally");
	goto clean_up;
    }

    HTSoc = soc;
    req->isoc = isoc;
    req->output_stream = HTWriter_newNoClose(soc);

    if (req->method == METHOD_INVALID) {
	if (HTReqLine) {
	    char * msg = (char*)malloc(strlen(HTReqLine) + 60);
	    sprintf(msg, "Invalid request \"%s\" (unknown method)", HTReqLine);
	    HTStatusCode = 400;
	    HTReasonLine = "Invalid method";
	    HTLoadError(req, 400, msg);
	    free(msg);
	}
	else HTLoadError(req, 400,
			 "Invalid request -- completely unable to parse it");
	goto clean_up;
    }
    if (HTDisabled[(int)req->method]) {
	char msg[80];
	sprintf(msg, "Method %s is disabled on this server",
		HTMethod_name(req->method));
	HTStatusCode = 403;
	HTReasonLine = "Disabled method";
	HTLoadError(req, 403, msg);
	goto clean_up;
    }
    if (!HTReqArg) {
	char * msg = (char*)malloc((HTReqLine ?strlen(HTReqLine) :0)+30);
	if (!msg) outofmem(__FILE__, "HTHandle");
	sprintf(msg, "Invalid request \"%s\"",
		(HTReqLine ? HTReqLine : "-null-"));
	HTStatusCode = 400;
	HTReasonLine = "Invalid request";
	HTLoadError(req, 400, msg);
	free(msg);
	goto clean_up;
    }


do_redirection_on_fly:

/*
**	Check Access Authorization (and translate by rules at the same time).
*/
#ifdef ACCESS_AUTH

#ifdef VMS
    /* enable sysprv to get to security files */
    HTVMS_enableSysPrv();
#endif /* VMS */

    HTStatusCode = HTAA_checkAuthorization(req);
    CTRACE(stderr, "AA.......... check returned %d\n", HTStatusCode);
    CTRACE(stderr, "Translated.. \"%s\"\n",
	   HTReqTranslated ? HTReqTranslated : "-null-");

#ifdef VMS
    HTVMS_disableSysPrv();
#endif /* VMS */

    if (HTStatusCode != 200  &&  HTStatusCode != 302) {
	HTLoadError(req, HTStatusCode, HTAA_statusMessage(req));
	goto clean_up;
    }
    /* otherwise access is authorized, continue */

#ifdef VMS
  if (HTVMS_authSysPrv() == YES) {
      char *filename = HTReqTranslated;
      char *uid_name = NULL;

      if (!filename)
         filename = HTReqScript;

      /* if SETUID is specified we will try to access the file for that UID */
      if (HTDoSetUid  &&  HTUser  &&  HTUser->username)
         uid_name = HTUser->username;

      /* if UID still not specified we take the one from the protect rule */
      if (!uid_name)
         uid_name = HTAA_getUidName();

      /* check if we can access the file from UID; GID is ignored */
      if ((filename) && (0 != strcmp(uid_name,""))) {
      	 CTRACE(stderr, "HTHandle.... checking access to file '%s' for user '%s'\n",
	       	     "server", filename, uid_name);

         /* enable SYSPRV if indeed we have access; actual access is done via system field in the protection */
         HTVMS_enableSysPrv();

	CTRACE(stderr,
	       "WARNING..... serving document when running with SYSPRV\n");

         if (HTVMS_checkAccess(filename,uid_name,req->method) != YES) {
            HTVMS_disableSysPrv();
            HTLoadError(req, 403, "Forbidden -- by rule");
            goto clean_up;
         }

         /* access under VMS authorized, go on with SYSPRV enabled... */
      } /* No Protection defined, or no uidname given */
  }
  else {
      CTRACE(stderr, "HTHandle.... Not running with SYSPRV\n");
  }

#else /* not VMS */
    if (getuid() == 0) {
	struct passwd *pw = NULL;
	int uid = 0;
	int gid = 0;

	if (HTDoSetUid  &&  HTUser  &&  HTUser->username) {
	    pw = getpwnam(HTUser->username);
	    if (pw) {
		uid = pw->pw_uid;
		gid = pw->pw_gid;
		CTRACE(stderr, "SETUID...... %s [uid %d] [gid %d]\n",
		       HTUser->username, uid, gid);
	    }
	    else {
		HTLog_error2("SETUID ERROR no such user:", HTUser->username);
	    }
	}
	if (!pw) { 
	    uid = HTAA_getUid();
	    gid = HTAA_getGid();
	}

	if (!pw && !(pw = getpwuid(uid)))
	    HTLog_errorN("Failed to get passwd entry for uid",uid);
	else if (initgroups(pw->pw_name, gid) == -1)
	    HTLog_error2("Failed to init groups for user",pw->pw_name);

	CTRACE(stderr, "Doing... setgid(%d) and setuid(%d)\n", gid, uid);
	if (setgid(gid) == -1)
	    HTLog_errorN("Failed to set group id to", gid);
	if (setuid(uid) == -1) {
	    HTLog_errorN("Failed to set user id to", uid);
	    CTRACE(stderr, "WARNING..... serving doc when running as root\n");
	}
    }
    else {
	CTRACE(stderr, "HTHandle.... Not running as root (uid=%d)\n",
	       getuid());
    }
#endif /* not VMS */
#endif	/* ACCESS_AUTH */



/*
**	Handle command
*/
    if (HTReason == HTAA_OK_GATEWAY) {  /* NOT a local file -- gatewaying */

	output_timeout_on();
	HTImProxy = YES;

	if (HTProxyCacheRoot  &&  req->method==METHOD_GET &&
	    !req->authorization  &&  !HTReqArgKeywords) {
	    char * cfn = NULL;
	    FILE * cf = NULL;
	    time_t if_ms = 0;
	    time_t expires = 0;
	    HTCacheTask task;

	    /*
	     * We don't want to die when we are doing cache operations,
	     * even if we get a SIGPIPE signal.
	     */
	    ignore_sigpipes = YES;

	    task = cache_lookup(HTReqTranslated,&cfn,&cf,&if_ms,&expires);

	    if (task == CACHE_FOUND) {		/* Cache file exists */
		ignore_sigpipes = NO;
		CTRACE(stderr,"Cache hit... Reading from \"%s\"\n", cfn);
		cf = fopen(cfn, "r");
		if (cf) {
		    bytes_from_cache =
			HTLoadCacheToStream(cf, req->output_stream, expires);
		    fclose(cf);
		    (*req->output_stream->isa->free)(req->output_stream);
		    status = HT_LOADED;
		    HTLog_cacheAccess(cfn,
				      remote_ident,
				      HTUser ? HTUser->username : NULL,
				      req);
		    already_logged = YES;
		    free(cfn);
		}
		else {
		    HTLog_error2("Couldn't open cache file for reading:",cfn);
		    free(cfn);
		    /* Load from remote host */
		    if (cache_no_connect)
			status = HTLoadError(req, 400, "Internal proxy cache standalone mode error; couldn't open cache file for reading.");
		    else
			status = HTLoadToStream(HTReqTranslated, NO, req);
		}
	    }
	    else if (cache_no_connect) {	/* Standalone cache mode */
		if (task == CACHE_NO)
		    HTLoadError(req, 404, "Proxy is in standalone cache mode [no external connections enabled] and the document was <B>not found in the cache</B>.");
		else /* task == CACHE_IF_MODIFIED */
		    HTLoadError(req, 404, "Proxy is in standalone cache mode with expiry checking enabled; document <B>was found in the cache but it had expired</B>. If it is desirable to see all cached documents, even if they have expired, expiry checking should be turned off.");
	    }
	    else {
		if (task == CACHE_CREATE && cfn && cf) {
		    /* Write to cache file */
		    req->output_stream =
			HTProxyCache(req->output_stream,
				     cf, cfn, HTReqTranslated, (time_t)0);
#if 0
			HTTee(req->output_stream,
			      HTProxyCache(cf, cfn, HTReqTranslated));
#endif
		    CTRACE(stderr,
			   "Cache....... Writing to lockfile \"%s%s\"\n",
			   cfn, LOCK_SUFFIX);
		    free(cfn);
		}
		else if (task == CACHE_IF_MODIFIED && cfn && cf && if_ms) {
		    CTRACE(stderr,
			   "Cache....... conditional GET if modif.since %s",
			   ctime(&if_ms));
		    req->output_stream =
			HTProxyCache(req->output_stream,
				     cf, cfn, HTReqTranslated, if_ms);
		    if (HTProxyHeaders) {
			char * tmp = (char*)malloc(strlen(HTProxyHeaders)+100);
			sprintf(tmp, "If-Modified-Since: %s\r\n%s",
				http_time(&if_ms), HTProxyHeaders);
			free(HTProxyHeaders);
			HTProxyHeaders = tmp;
		    }
		    else {
			HTProxyHeaders = (char*)malloc(100);
			sprintf(HTProxyHeaders, "If-Modified-Since: %s\r\n",
				http_time(&if_ms));
		    }
		}
		else {		/* task == CACHE_NO */
		    CTRACE(stderr, "Cache....... Not caching\n");
		    ignore_sigpipes = NO;
		}

		status = HTLoadToStream(HTReqTranslated, NO, req);
	    }
	}
	else {
	    /* Normal retrieve with no caching */
	    CTRACE(stderr, "No caching.. %s\n",
		   !HTProxyCacheRoot		? "enabled" :
		   req->method!=METHOD_GET	? "for other methods than GET":
		   req->authorization		? "for protected documents" :
		   HTReqArgKeywords		? "queries" :
		   "but dunno why not...??");
	    status = HTLoadToStream(HTReqTranslated, NO, req);
	}

	HTImProxy = NO;
	timeout_off();

    }
    else if (HTReason == HTAA_OK_REDIRECT) {	/* Do a redirection */
	CTRACE(stderr, "Redirection.\n");
	output_timeout_on();
	status = HTLoadRedirection(req);
	timeout_off();
    }
    else if (HTReqScript) {	/* Call a script to generate a document */
	CTRACE(stderr, "Script call.\n");
	status = HTCallScript(req);
    }
    else if (HTReqArgKeywords) {
	CTRACE(stderr, "Search......\n");
	if (HTSearchScript) {	/* Search enabled */
	    if (HTReqTranslated) {
		StrAllocCopy(HTScriptPathTrans, HTReqTranslated);
		StrAllocCopy(HTScriptPathInfo, HTReqArgPath);
		StrAllocCopy(HTReqScript, HTSearchScript);
		req->content_length = 0;
		status = HTCallScript(req);
	    }
	    else {
		HTLog_error("Translated NULL when should do search");
		HTLoadError(req, 403, "Sorry, forbidden by rule");
	    }
	}
	else {
	    CTRACE(stderr, "HTHandle.... can't perform search %s\n", HTReqArg);
	    status =
		HTLoadError(req, 501,
			    "Sorry, this server does not perform searches.");
	}
    } /* If keywords given */
    else {  /* Load the document normally into the client */

	CTRACE(stderr, "HTHandle.... method %s\n", HTMethod_name(req->method));

	output_timeout_on();

        if (req->method == METHOD_GET  ||
	    req->method == METHOD_CHECKOUT) {

	    if (HTIfModifiedSince  &&  HTLastModified  &&
		parse_http_time(HTLastModified) <= HTIfModifiedSince) {
		CTRACE(stderr,
		       "Not modified since %s", ctime(&HTIfModifiedSince));
		HTStatusCode = 304;
		HTReasonLine = "Not modified";
		status = HTLoadHead(req);
	    }
	    else
		status = HTRetrieve(req);

	} else if (req->method == METHOD_HEAD) {

	    status = HTLoadHead(req);

	} else if (req->method == METHOD_PUT  ||
		   req->method == METHOD_CHECKIN) {

	    BOOL OkToPut = NO;

	    if (OkToPut) {
		HTPut(req);
	    }
	    else if (HTPutScript) {
		CTRACE(stderr,
		       "HTHandle.... handling PUT with script \"%s\"\n",
		       HTPutScript);

		StrAllocCopy(HTReqScript, HTPutScript);
		StrAllocCopy(HTScriptPathInfo, HTReqArgPath);
		HTScriptPathTrans = HTParse(HTReqTranslated, "",
					    PARSE_PATH | PARSE_PUNCTUATION);
		status = HTCallScript(req);
	    }
	    else {
		HTLoadError(req, 500,
			    "This server is not configured to handle PUT");
	    }

	} else if (req->method == METHOD_POST) {

	    BOOL OkToAnnotate = NO;

	    if (OkToAnnotate) {
		HTAnnotate(req);
	    }
	    else if (HTPostScript) {
		CTRACE(stderr,
		       "HTHandle.... handling POST with script \"%s\"\n",
		       HTPostScript);

		StrAllocCopy(HTReqScript, HTPostScript);
		StrAllocCopy(HTScriptPathInfo, HTReqArgPath);
		HTScriptPathTrans=HTParse(HTReqTranslated, "",
					  PARSE_PATH | PARSE_PUNCTUATION);
		status = HTCallScript(req);
	    }
	    else {
		HTLoadError(req, 500,
			    "This server is not configured to handle POST");
	    }

	} else {
	    char *msg;

	    msg = (char*)malloc(strlen(HTMethod_name(req->method))+50);
	    if (!msg) outofmem(__FILE__, "HTHandle");

	    sprintf(msg, "Method '%s' invalid or not implemented",
		    HTMethod_name(req->method));
	    HTLoadError(req, 501, msg);
	    CTRACE(stderr, "Daemon...... %s\n", msg);
	    free(msg);
	}

	timeout_off();

    } /* Handle command */

    if (status == HT_REDIRECTION_ON_FLY) {
	CTRACE(stderr, "Redirection. on the fly for \"%s\"\n", HTLocation);
	if (++redir_hops > 10)
	    HTLoadError(req, 501,
	    "Too many redirection-on-the-fly hops, max 10 (probably looping)");
	else {
	    if (req->method != METHOD_GET  &&  req->method != METHOD_HEAD)
		req->method = METHOD_GET;
	    FREE(HTReqArg);	/* Leak fixed AL 6 Feb 1994 */
	    HTReqArg = HTLocation;
	    HTLocation = NULL;
	    compute_server_env();

	    goto do_redirection_on_fly;
	}
    }

#ifdef VMS
#ifdef ACCESS_AUTH
    /* disable SYSPRV again */
    HTVMS_disableSysPrv();
#endif /* ACCESS_AUTH */
#endif /* VMS */


 clean_up:

    cleanup_timeout_on();

    /*
    **	Log the call
    */
    if (!already_logged)
	HTLog_access(remote_ident,  HTUser ? HTUser->username : NULL,  req);
    if (req) {
	if (req->isoc) HTInputSocket_free(req->isoc);
	HTRequest_delete(req);
    }

    timeout_off();

    return 0;		/* End of file - please close socket */

} /* HTHandle */




/*      Handle incoming messages				server_loop()
**	-------------------------
**
** On entry:
**
**      timeout         -1 for infinite, 0 for poll, else in units of 10ms
**
** On exit,
**	returns		The status of the operation, <0 if failed.
**			0	means end of file
**
*/
PRIVATE int server_loop NOARGS
{
    int tcp_status;		/* <0 if error, in general */

    for(;;) {

	int soc_addr_len = sizeof(client_soc_addr);

	/*  If it's a master socket, then find a slave: */
    	if (role == master) {
#ifdef SELECT
	    fd_set		read_chans;
	    fd_set		write_chans;
	    fd_set		except_chans;
	    int			nfound;	    /* Number of ready channels */
	    struct timeval	max_wait;   /* timeout in form for select() */

	    FD_ZERO(&write_chans);	    /* Clear the write mask */
	    FD_ZERO(&except_chans);	    /* Clear the exception mask */

	    for (com_soc=(-1); com_soc<0;) {	/* Loop while connections keep coming */
		int timeout;

	      retry_select:

		/*
		 * The read mask expresses interest in the master
		 * channel for incoming connections) or any slave
		 * channel (for incoming messages).
		 */

		/*  Wait for incoming connection or message */
	        read_chans = open_sockets;   /* Read on all active channels */
		CTRACE(stderr, "Daemon...... %s (Mask=%x hex, max=%x hex).\n",
		       "Waiting for connection or message (select())",
		       *(unsigned int *)(&read_chans),
		       (unsigned int)num_sockets);

		timeout = get_next_timeout();
		if (timeout < 0) {
		    CTRACE(stderr, "No timeout.. for select()\n");
		}
		else {
		    CTRACE(stderr,
			   "Next timeout after %d hours %d mins %d secs\n",
			   timeout/360000,
			   (timeout/6000)%60,
			   (timeout/100)%60);
		}

		if (timeout>=0) {
		    max_wait.tv_sec = timeout/100;
		    max_wait.tv_usec = (timeout%100)*10000;
		}

		nfound=select(num_sockets, &read_chans,
			      &write_chans, &except_chans,
			      timeout >= 0 ? &max_wait : NULL);

		if (nfound < 0) {	/* Interrupted */
		    (void)HTInetStatus("select");
		    CTRACE(stderr, "Interrupted. select() -- retrying\n");
		    goto retry_select;
		}

		if (nfound==0) {	/* Timeout */
		    CTRACE(stderr,
			   "I think..... it might be time to do daily gc\n");
		    if (time_to_do_gc()) {
			CTRACE(stderr, "Yep......... Let's do gc\n");
			gc_pid = fork();
			if (!gc_pid) {	/* Child */
			    CTRACE(stderr, "Child....... DOING GC\n");
			    gc();
			    CTRACE(stderr, "Child....... GC COMPLETE\n");
			    exit(0);		
			}
			else {
			    if (gc_pid < 0) {
				HTLog_error("FORK FAILED when launching gc\n");
				gc_pid = 0;
			    }
			}
		    }
		    else {
			CTRACE(stderr, "No.......... Not yet\n");
		    }

		    goto retry_select;
		}


		/*
		 * We give priority to existing connected customers. When
		 * there are no outstanding commands from them, we look for
		 * new customers.
		 *
		 * If a message has arrived on one of the channels, take
		 * that channel:
		 */
		{
		    int i;
		    for(i=0; i<num_sockets; i++)
			if (i != master_soc)
			    if (FD_ISSET(i, &read_chans)) {
				CTRACE(stderr,
				       "Message..... waiting on sock %d\n",i);
				com_soc = i;		/* Got one! */
				break;
			    }
		    if (com_soc>=0) break; /* Found input socket */
		    
		} /* block */
		
		/*
		 * If an incoming connection has arrived, accept the new
		 * socket:
		 */
		if (FD_ISSET(master_soc, &read_chans)) {
#ifdef OLD_CODE
		    soc_addrlen = sizeof(soc_address); /* JS 930107 */
		    CTRACE(tfp, "Daemon...... New incoming connection:\n");
  retry_accept_1:
		    tcp_status = accept(master_soc,
					(struct sockaddr *)&soc_address,
					&soc_addrlen);
#endif
		    CTRACE(tfp, "Daemon...... New incoming connection:\n");
  retry_accept_1:
		    tcp_status = accept(master_soc,
					(struct sockaddr *)&client_soc_addr,
					&soc_addr_len);

		    if (tcp_status<0) {
			CTRACE(stderr, "Interrupted. accept() -- retrying\n");
			(void)HTInetStatus("accept");
			goto retry_accept_1;
		    }

		    CTRACE(tfp,"Accepted.... new socket %d\n",tcp_status);
		    FD_SET(tcp_status, &open_sockets);
		    if ((tcp_status+1) > num_sockets)
			num_sockets=tcp_status+1;

		} /* end if new connection */

	    } /* loop on event */

#else	/* SELECT not supported */

#ifdef OLD_CODE
	    SockA peer_soc;
	    int peer_len = sizeof (peer_soc);
#endif
	    CTRACE(tfp,"Daemon...... %s\n",
		   "Waiting for incoming connection (accept())");
	  retry_accept_2:
#ifdef DECNET
	    tcp_status = accept(master_soc, &client_soc_addr, &soc_addr_len);
#else  /* For which machine is this ??? rsoc is undeclared, what's mdp ? */
	    tcp_status = accept(master_soc,
				&rsoc->mdp.soc_tcp.soc_address,
				&rsoc->mdp.soc_tcp.soc_addrlen);
#endif
	    if (tcp_status<0) {
		CTRACE(stderr, "Interrupted. accept() -- retrying\n");
		(void)HTInetStatus("accept");
		goto retry_accept_2;
	    }

	    com_soc = tcp_status;	/* socket number */
	    CTRACE(tfp, "Accepted.... socket %d\n", tcp_status);
    
#endif

	}  /* end if master */


/* com_soc is now valid for read */
/* Do fork() to handle it*/

	{
	    int fork_status;
	    BOOL child = NO;
	    BOOL do_gc = NO;

	    time(&cur_time);

	    if (do_rfc931) {
		CTRACE(stderr, "RFC 931..... Getting remote ident\n");
		remote_ident = rfc931(&client_soc_addr,&server_soc_addr);
		CTRACE(stderr, "Ident....... %s\n",
		       remote_ident ? remote_ident : "-none-");
	    }
	    else remote_ident = NULL;

#ifdef FORKING
	    if (HTForkEnabled) {	/* fork -- child handles the request */
		CTRACE(stderr, "server_loop. Doing fork()\n");
		child_serial++;
		if (HTProxyCacheRoot  &&  !gc_pid  &&  time_to_do_gc())
		    do_gc = YES;
		fork_status = fork();
	    }
	    else fork_status = 0;
#else
	    fork_status = 0;
#endif /* FORKING */

	    if (fork_status < 0) {	/* fork() failed */
		HTLog_error("fork() FAILED");
	    }
	    else if (fork_status > 0) {	/* Parent */
		CTRACE(stderr, "HTHandle.... fork() succeeded\n");
		if (do_gc) {
		    CTRACE(stderr, "Parent...... That child will do gc\n");
		    gc_pid = fork_status;
		    do_gc = NO;
		}
	    }
	    if (fork_status != 0) {	/* fork() failed or parent */
		NETCLOSE(com_soc); /* We can close this -- child handles it */
#ifdef SELECT
		FD_CLR(com_soc, &open_sockets);
#endif

#ifdef SIGTSTP	/* BSD */
		/* If previous children have finished wait the zombie
		** processes away.
		*/
		sig_child();
#endif
	    }
	    else {	/* Child */
		/* This no-linger stuff is borrowed from NCSA httpd code */
		struct linger sl;

#ifdef FORKING
		if (HTForkEnabled) {
		    CTRACE(stderr, "Child....... I'm alive\n");
		    child = YES;
		    if (do_gc) {
			CTRACE(stderr,"Child....... I will do gc upon exit\n");
		    }
		}
#endif
		if (TRACE && !child)
		    fprintf(stderr, "server_loop. Parent serving\n");

		sl.l_onoff = 1;
		sl.l_linger = 600; /* currently ignored anyway */
		/* this should check error status, but it's not crucial */

/* #ifndef __svr4__  How would I do this for Solaris??? */
		setsockopt(com_soc, SOL_SOCKET, SO_LINGER,
			   (char*)&sl, sizeof(sl));
/* #endif */

		{
		    SockA addr;
		    int namelen = sizeof(addr);
		    char ip_address[16];
		    FREE(HTClientHostName);

#ifdef DECNET
		    StrAllocCopy(HTClientHost<, "DecnetClient");
		    /* TBD */
		    HTClientHostName = NULL;
#else
		    getpeername(com_soc, (struct sockaddr*)&addr, &namelen);

		    strncpy(ip_address,
			    (char*)inet_ntoa(addr.sin_addr),
			    sizeof(ip_address));
		    ip_address[15] = 0;
		    StrAllocCopy(HTClientHost, ip_address);
		    HTClientHostName = HTGetHostName(com_soc);
#endif
		}

		/*  Read the message now on whatever channel there is */
		CTRACE(tfp,"Reading..... socket %d from host %s\n",
		       com_soc, HTClientHost);

		tcp_status = HTHandle(com_soc);

		if(tcp_status <= 0) {		/* EOF or error */
		    if (tcp_status < 0) {		/* error */
			CTRACE(tfp,
			       "ERROR....... %d handling msg (errno=%d).\n",
			       tcp_status, errno);
			/* DONT return HTInetStatus("netread");	 error */
		    } else {
			CTRACE(tfp,
			       "Disconnected Socket %d disconnected by peer\n",
			       com_soc);
		    }
		    if (role==master) {
			shutdown(com_soc, 2);
			NETCLOSE(com_soc);
#ifdef SELECT
			FD_CLR(com_soc, &open_sockets);
#endif
		    } else {  /* Not multiclient mode */
#ifdef VM
			return -69;
#else
			return -ECONNRESET;
#endif
		    }
		} else {/* end if handler left socket open */
		    shutdown(com_soc, 2);
		    NETCLOSE(com_soc);
#ifdef SELECT
		    FD_CLR(com_soc, &open_sockets);
#endif
		}
		if (child) {
		    if (do_gc) gc();
		    CTRACE(stderr,"Child....... exiting with status %d\n",
			   HTChildExitStatus);
		    exit(HTChildExitStatus);
		}
	    }
	}

    } /* for loop */
    
    /* NOT REACHED */

} /* end server_loop */



#ifdef FORKING
/*
** This function (modified) is taken from:
**	W.Richard Stevens: UNIX Network Programming,
**	Prentice Hall 1990, pp. 82-85
*/
PRIVATE int daemon_start NOARGS
{
    register int	childpid, fd;

#ifndef NO_BACKGROUNDING	/* If this should prove to be non-portable */

    HTStandAlone = YES;

    /*
    ** If started by init there's no need to detach
    */
    if (getppid() == 1)
	goto out;

    /*
    ** Fork and let the parent exit to go backgroud
    */
    if ((childpid = fork()) < 0)
	HTLog_error("Can't fork to go background");
    else if (childpid > 0)	/* Parent */
	exit(0);

    /* I'm first child now */

    /*
    ** Disassociate from controlling terminal and process group.
    ** Ensure the process can't reacquire a new controlling terminal.
    */
#ifdef SIGTSTP	/* BSD */

#if defined(__svr4__) || defined(_POSIX_SOURCE) || defined(__hpux)
    pgrp = setsid();
#else
    pgrp = setpgrp(0, getpid());

    if ((fd = open("/dev/tty", O_RDWR)) >= 0) {
	ioctl(fd, TIOCNOTTY, (char*)NULL);	/* Lose controlling tty */
	close(fd);
    }
#endif
    if (pgrp == -1)
	HTLog_error("Can't change process group");

#else	/* SysV */

    pgrp = setpgrp();
    if (pgrp == -1)
	HTLog_error("Can't change process group");

    if ((childpid = fork()) < 0)
	HTLog_error("Can't fork second child");
    else if (childpid > 0)	/* First child */
	exit(0);

    /* I'm second child now */

#endif

  out:

#ifdef OLD_CODE	/* For some reason this prevents NPH-scripts from working */
    /*
    ** Close any open file descriptors except the socket we are listening
    */
    for (fd=0; fd<NOFILE; fd++) {
	if (fd != master_soc  &&
	    (access_log_fileno == -1 || fd != access_log_fileno) &&
	    (error_log_fileno == -1  || fd != error_log_fileno))
	    close(fd);
    }
    errno = 0;	/* Probably got set to EBADF a few times by close() */
#endif

    /*
    ** Move the current working directory to root, to make sure we
    ** aren't on a mounted filesystem.
    */
    chdir("/");

    /*
    ** Clear any inherited file mode creation mask
    */
    umask(0);

#endif /* NO_BACKGROUNDING */

    return 0;
}
#endif /* FORKING */



#ifdef OLD_CODE		/* CLIENT_CONTROL is phased out by native httpd AA */
#if defined(SOCKS) && defined(CLIENT_CONTROL)
typedef struct client_entry {
    char *address;
    struct client_entry *next;
} client_entry;

client_entry *client_list;

PUBLIC int init_allowed_hosts NOARGS
{

#define ALLOWED_HOSTS_FILE  "/etc/httpd.yacf"
#define MAXLINE             132

    char inbuf[MAXLINE+1];
    FILE *yacf;
    client_entry *this_client;
    client_entry *previous_client;

    if ((yacf = fopen(ALLOWED_HOSTS_FILE, "r")) == NULL) {
        fprintf(stderr, "Unable to open allowed hosts file.\n");
        exit(1);
    }

    previous_client = NULL;

    while (fscanf(yacf, "%s ", inbuf) == 1) {
        this_client = (client_entry *) malloc(sizeof(client_entry));
        if (previous_client == NULL)
            client_list = this_client;
        else
            previous_client->next = this_client;
        this_client->address = (char *) malloc(strlen(inbuf) + 1);
        strcpy(this_client->address, inbuf);
        previous_client = this_client;
        this_client->next = NULL;
    }        
    fclose(yacf);
    
}    

PUBLIC BOOL is_host_allowed ARGS1(char *, h)
{
    client_entry *this_client;

    if (!h) return NO;

    this_client = client_list;
    while (1) {
        if (strcmp(this_client->address, h) == 0)
            return YES;
        if ((this_client = this_client->next) == NULL)
            break;
    }
    HTLog_error2("REFUSED connection from ", h);
    return NO;
}
#endif
#endif /* OLD_CODE */



/*		Main program
**		------------
**
**	Options:
**	-v		verify: turn trace output on to stdout 
**	-a addr		Use different tcp port number and style
**	-p port		Prefered
**	-l file		Log requests in ths file
**	-r file		Take rules from this file
**	-R file		Clear rules and take rules from file.
**	-setuid		Set uid to authenticated uid
**	-cache dir	Set gateway http cache root directory
**	-nolog templ	Don't log requests coming from hosts matching template
**
**	Parameters:
**		directory	directory to export
*/
int main ARGS2 (
	int,	argc,
	char**,	argv)
{
    int status;
#ifdef RULES
    int rulefiles = 0;		/* Count number loaded */
#endif
    char *directory = NULL;
    BOOL gc_only = NO;

#ifdef SOCKS
    SOCKSinit(argv[0]);

#ifdef OLD_CODE		/* CLIENT_CONTROL is phased out by native httpd AA */
#ifdef CLIENT_CONTROL
    init_allowed_hosts();
#endif    
#endif
#endif

    WWW_TraceFlag = 0;		/* diagnostics off by default */
    HTImServer = YES;		/* To prevent HTAccess from translating */
    HTImProxy = YES;		/* Always be able to run as proxy */

    time(&cur_time);

    HTDisabled[ (int)METHOD_PUT ] = YES;
    HTDisabled[ (int)METHOD_DELETE ] = YES;
    HTDisabled[ (int)METHOD_CHECKIN ] = YES;
    HTDisabled[ (int)METHOD_CHECKOUT ] = YES;
    HTDisabled[ (int)METHOD_SHOWMETHOD ] = YES;
    HTDisabled[ (int)METHOD_LINK ] = YES;
    HTDisabled[ (int)METHOD_UNLINK ] = YES;

#ifdef VMS
   /* enable sysprv to get to rule file and allocate port */
   HTVMS_enableSysPrv();
#endif /* VMS */

#ifdef RULES
    HTClearRules();
#endif
#ifndef NO_INIT
    HTFileInit();	/* Initialize filename suffix table */
#endif

    {
    	int a;
	    
	for (a=1; a<argc; a++) {
	    
	    if (0==strcmp(argv[a], "-v") || 0==strcmp(argv[a], "-vv")) {
	        WWW_TraceFlag = 1;
		CTRACE(stderr,
	       "............ This is %s, version %s, using libwww version %s\n",
		       HTAppName, HTAppVersion, HTLibraryVersion);
		if (!strcmp(argv[a], "-vv"))
		    WWW_VeryVerboseFlag = YES;

	    } else if (0==strcmp(argv[a], "-version")) {
	        printf("CERN World-Wide Web Daemon %s, libwww %s\n",
			VD, HTLibraryVersion);
		exit(0);

	    } else if (0==strcmp(argv[a], "-setuid")) {
		HTDoSetUid = YES;
		CTRACE(stderr, "Option...... SETUID feature on\n");

	    } else if (0==strncmp(argv[a], "-cacheroot", 5)) {
		if (++a < argc) {
		    StrAllocCopy(HTProxyCacheRoot, argv[a]);
		    CTRACE(stderr, "Cache root.. \"%s\"\n", argv[a]);
		}

	    } else if (0==strncmp(argv[a], "-gc_only", 3)) {
		CTRACE(stderr, "gc-only mode\n");
		gc_only = YES;

	    } else if (0==strcmp(argv[a], "-nobg")) {
		HTNoBackGround = YES;

	    } else if (0==strcmp(argv[a], "-nolog")) {
		if (++a < argc) {
		    if (!HTNoLog) HTNoLog = HTList_new();
		    HTList_addObject(HTNoLog, (void*)argv[a]);
		}

	    } else if (0==strcmp(argv[a], "-disable")) {
		if (++a < argc) {
		    HTMethod m = HTMethod_enum(argv[a]);
		    if (m != METHOD_INVALID) {
			HTDisabled[(int)m] = YES;
			CTRACE(stderr, "Disabled.... %s\n", HTMethod_name(m));
		    }
		    else {
			fprintf(stderr, "Invalid method to disable: \"%s\"\n",
				argv[a]);
			HTLog_error2("Invalid method to disable:", argv[a]);
		    }
		}

	    } else if (0==strcmp(argv[a], "-enable")) {
		if (++a < argc) {
		    HTMethod m = HTMethod_enum(argv[a]);
		    if (m != METHOD_INVALID) {
			HTDisabled[(int)m] = NO;
			CTRACE(stderr,"Enabled..... %s\n",HTMethod_name(m));
		    }
		    else {
			fprintf(stderr, "Invalid method to enable: \"%s\"\n",
				argv[a]);
			HTLog_error2("Invalid method to enable:", argv[a]);
		    }
		}

	    } else if (0==strcmp(argv[a], "-a")) {
	        if (++a<argc) {
		    StrAllocCopy(HTPortAddr, argv[a]);
#ifdef FORKING
		    HTForkEnabled = YES;
#endif /* FORKING */
		}
	    } else if (0==strcmp(argv[a], "-p")) {
	        if (++a<argc) {
		    HTPortAddr = (char*)malloc(strlen(argv[a])+10);
		    sprintf(HTPortAddr, "*:%s", argv[a]);
#ifdef FORKING
		    HTForkEnabled = YES;
#endif /* FORKING */
		}
#ifdef FORKING
	    } else if (0==strcmp(argv[a], "-fork")) {	/* fork when serving */
		HTForkEnabled = YES;
		/*
		** This is here only for debugging anymore.
		*/
#endif /* FORKING */

#ifdef RULES
	    } else if (0==strcmp(argv[a], "-r")) {
	        if (++a<argc) { 
		    if (HTLoadRules(argv[a]) < 0) exit(-1);
		    rulefiles++;
		}
	    } else if (0==strcmp(argv[a], "-R")) {
		rulefiles++;		/* Inhibit rule file load */
#endif
#ifdef DIR_OPTIONS
	    } else if (0==strncmp(argv[a], "-d", 2)) {
	    	char *p = argv[a]+2;
		for(;*p;p++) {
		    switch (*p) {
		      case 'b':	HTDirReadme = HT_DIR_README_BOTTOM;	break;
		      case 'n':	HTDirAccess = HT_DIR_FORBID;		break;
		      case 'r':	HTDirReadme = HT_DIR_README_NONE;	break;
		      case 's':	HTDirAccess = HT_DIR_SELECTIVE;		break;
		      case 't':	HTDirReadme = HT_DIR_README_TOP;	break;
		      case 'y':	HTDirAccess = HT_DIR_OK;		break;
		      default:
			fprintf(stderr, 
				"ERROR....... bad -d option %s\n", argv[a]);
			HTLog_error2("Bad -d option", argv[a]);
			exit(-4);
		    }
		} /* loop over characters */
#endif
	    } else if (0==strcmp(argv[a], "-gmt")) {
		CTRACE(stderr, "Logging..... in GMT\n");
		HTUseGMT = YES;

	    } else if (0==strcmp(argv[a], "-localtime")) {
		CTRACE(stderr, "Logging..... in local time\n");
		HTUseGMT = NO;

	    } else if (0==strcmp(argv[a], "-l") ||
		       0==strcmp(argv[a], "-newlog")) { /* template */
		if (!strcmp(argv[a], "-newlog"))
		    HTNewLogFileFormat = YES;
	        if (++a<argc)
		    HTLog_accessOpen(argv[a]);
		else {
		    fprintf(stderr,
			    "ERROR....... No parameter for -l option\n");
		    HTLog_error("No parameter for -l/-newlog option");
		}

	    } else if (0==strcmp(argv[a], "-errlog")) { /* template */
	        if (++a<argc)
		    HTLog_errorOpen(argv[a]);
		else {
		    fprintf(stderr,
			    "ERROR....... No parameter for -errlog option\n");
		    HTLog_error("No parameter for -errlog option");
		}

	    } else if (argv[a][0] != '-') {	/* Parameter */
	    	if (!directory) directory = argv[a];
		
	    } /*ifs */
	} /* for each arg */
    } /* scope of a */

#ifdef RULES

    if (rulefiles==0) {		/* No mention */
    	if (!directory) {
	    if (HTLoadRules(RULE_FILE) < 0) {	/* Default rule file? */
	        directory = DEFAULT_EXPORT;
	    };
	}
	if (directory) {
	    CONST char * myname = HTHostName();
	    char * mapto = malloc(strlen(myname) + strlen(directory) + 10);
	    sprintf(mapto, "file://%s%s/*", myname, directory);
	    HTAddRule(HT_Pass, "/*", mapto);
	    HTAddRule(HT_Fail, "*", NULL);
	    free(mapto);
	}
    } else {
        if (directory) {
	    fprintf(stderr,
	    	"Warning: -r or -R specified so %s directory param ignored\n",
	    	directory);
	}
    }
#endif

    HTLog_errorOpen(NULL);	/* Open if not already opened */

    /*
     * Initialize default icons.
     */
    if (server_root && !icons_inited  &&  (HTDirShowMask & HT_DIR_SHOW_ICON)) {
	CONST char * host = HTServerHostName ? HTServerHostName : HTHostName();
	char portstr[10];
	char * icon_url_prefix = NULL;
	char * icon_dir = NULL;

	CTRACE(stderr, "Automatic... icon initialization\n");

	StrAllocCopy(icon_dir, server_root);
	if (server_root[strlen(server_root)-1] != '/') {
	    StrAllocCat(icon_dir, "/icons/*");
	} else {
	    StrAllocCat(icon_dir, "icons/*");
	}
	HTPrependPass("/httpd-internal-icons/*", icon_dir);

	if (host) {
	    char * icon_url_templ = NULL;
	    if (HTPortAddr && (int)strlen(HTPortAddr) > 2 &&
		strcmp(HTPortAddr,"*:80"))
		strcpy(portstr, HTPortAddr+1);
	    else
		*portstr = 0;

	    icon_url_prefix = (char*)malloc(strlen(host)+100);
	    if (!icon_url_prefix) outofmem(__FILE__, "server-root-icon-init");
	    sprintf(icon_url_prefix,
		    "http://%s%s/httpd-internal-icons/", host, portstr);

	    /*
	     * Pass also absolute URL to proxy server itself
	     */
	    StrAllocCopy(icon_url_templ, icon_url_prefix);
	    StrAllocCat(icon_url_templ, "*");
	    HTPrependPass(icon_url_templ, icon_dir);
	    FREE(icon_url_templ);
	}
	else {
	    HTLog_error("Can't get hostname - icon URLs won't have host name");
	    StrAllocCopy(icon_url_prefix, "/httpd-internal-icons");
	}

	CTRACE(stderr,"Icon URL.... prefix is %s\n", icon_url_prefix);
	CTRACE(stderr,"Icons....... should be in %s\n", icon_dir);
	HTStdIconInit(icon_url_prefix);
	FREE(icon_dir);
	FREE(icon_url_prefix);
    }

    if (!HTPortAddr) StrAllocCopy(HTPortAddr, "");

#ifdef FORKING
    if (HTPortAddr && *HTPortAddr) {
	HTForkEnabled = YES;
	CTRACE(stderr, "ServerType.. standalone (turning on forking)\n");
    }
    else
#endif
      {
	CTRACE(stderr, "ServerType.. inetd or standalone nonforking\n");
      }

    status = do_bind(HTPortAddr);
    if (status<0) {
	HTLog_error(
	  "Can't bind and listen on port (maybe one already running)");
	HTLog_error(
	  "If running from inetd DO NOT use -p option or Port directive");
    	fprintf(stderr,
	  "HTTPD ERROR: Bad setup: Can't bind and listen on port.\n");
    	fprintf(stderr,
	  "Explanation: Possibly server already running, or if running");
	fprintf(stderr,
	  "from inetd make sure you're not using -p flag or Port directive\n");
	exit(status);
    }

#ifdef VMS
    HTVMS_disableSysPrv();
#endif /* VMS */
    
#ifdef FORKING
    /*
     * Go to background and disconnect from the terminal
     * when run standalone.
     */
    if (HTPortAddr && *HTPortAddr && !WWW_TraceFlag && !HTNoBackGround)
	daemon_start();

    set_signals();

    /*
     * Write the process id to pid_file.
     */
    if (HTPortAddr) {   /* Write pid to file */
	FILE * fp = NULL;
	char * fn = NULL;
	if (pid_file && pid_file[0] == '/')
	    StrAllocCopy(fn, pid_file);
	else {
	    StrAllocCopy(fn, server_root ? server_root : "/tmp");
	    StrAllocCat(fn, "/");
	    StrAllocCat(fn, pid_file ? pid_file : "httpd-pid");
	}
	fp = fopen(fn, "w");
	if (!fp) HTLog_error2("Can't open pid file for writing:",fn);
	else {
	    CTRACE(stderr, "PidFile..... %s\n", fn);
	    fprintf(fp,"%d\n",(int)getpid());
	    fclose(fp);
	}
	free(fn);
    }

#endif /* FORKING */

#ifdef VMS
   HTVMS_disableSysPrv();
#endif /* VMS */

    if (HTProxyCacheRoot) {
	gc_info_file = (char*)malloc(strlen(HTProxyCacheRoot) + 20);
	sprintf(gc_info_file, "%s/.gc_info", HTProxyCacheRoot);
	CTRACE(stderr, "Recovering.. cache to last gc state...\n");
	read_gc_results();
	if (gc_only) {
	    CTRACE(stderr, "FIRST ROUND...\n");
	    gc();
	    CTRACE(stderr, "FIRST ROUND COMPLETE.\n");
	    read_gc_results();
	    CTRACE(stderr, "SECOND ROUND...\n");
	    gc();
	    CTRACE(stderr, "SECOND ROUND COMPLETE.\n");
	    read_gc_results();
	    CTRACE(stderr, "BATCH-MODE GARBAGE COLLECTION COMPLETE.\n");
	    exit(0);
	}
    }
    else if (gc_only) {
	fprintf(stderr,
	"ERROR: Cache root not specified when started in gc-only mode\n");
	HTLog_error("Cache root not specified when started in gc-only mode\n");
	exit(-1);
    }

    status = HTUserInit();
    if (status<0) {
       exit(status);
    }

    status = server_loop();

    if (status<0) {
	exit(status);
    }
    
    exit(0);
    return 0;	/* NOTREACHED -- For gcc */
}

