/*		TCP/IP based server for HyperText		HTDaemon.c
**		---------------------------------
**
**
** Compilation options:
**	RULES		If defined, use rule file and translation table
**	DIR_OPTIONS	If defined, -d options control directory access
**	ACCESS_AUTH	If defined, include access authorization code
**      POSIXWAIT       If defined, uses waitpid instead of wait3
**
**  Authors:
**	TBL	Tim Berners-Lee, CERN
**	JFG	Jean-Francois Groff, CERN
**	JS	Jonathan Streets, FNAL
**	AL	Ari Luotonen, CERN
**	MD	Mark Donszelmann, CERN
**
**  History:
**	   Sep 91  TBL 	Made from earlier daemon files. (TBL)
**	26 Feb 92  JFG	Bug fixes for Multinet.
**       8 Jun 92  TBL  Bug fix: Perform multiple reads in case we don't get
**                      the whole command first read.
**	25 Jun 92  JFG  Added DECNET option through TCP socket emulation.
**	 6 Jan 93  TBL  Plusses turned to spaces between keywords
**	 7 Jan 93  JS   Bug fix: addrlen had not been set for accept() call
**			Logging in GMT to file-YYMM in name
**	   Sep 93  AL   Added Access Authorization, and
**			some minor fixes to make server more tolerant of
**			missing stuff (NULLs here and there). It can be
**			expected that new browsers might screw up with
**			Access Authorization at first.
**	20 Oct 93  AL	Now makes sure that it is really an HTTP1
**			request and not just an HTTP0 request with
**			spaces in the URL.
**	 6 Nov 93  MD	Bug fix: gmtime replace by localtime on VMS.
**			Bug fix: changed vms into VMS (ifdef is case sensitive)
**			calls to strcasecmp were changed into strcasecomp 
**			(same for strncasecmp).
**			Added file sharing of the logfile (VMS).
**			Added switching off of SYSPRV after bind (VMS).
**			disabled Forking (VMS).
**	12 Nov 93  AL	Understands wildcards in Content-Type.
**			Handles SIGPIPE.
**	23 Nov 93  MD	Bug fix: Corrected timestamp if running in stand-alone
**			mode. Used to give same time after start.
**			Flushing of buffers now also flushes RMS buffer.
**	 2 Dec 93  MD 	Fixed reuse of ports with open connections (VMS)
**	11 Dec 93  AL	Added HTRequest, HTHandle() heavily rewritten.
**			Added also a new module, HTRequest.c, which
**			now needs to be linked together with daemon.
**	14 Dec 93  AL	Forking moved to the beginning of HTHandle().
**	22 Feb 94  MD	Changed for VMS.
**			Added user init routines
*/
/* (c) CERN WorldWideWeb project 1990-1992. See Copyright.html for details */


/*	Module parameters:
**	-----------------
**
**  These may be undefined and redefined by syspec.h
*/

#define FORKING

#define LISTEN_BACKLOG 2	/* Number of pending connect requests (TCP)*/
#define MAX_CHANNELS 20		/* Number of channels we will keep open */
#define WILDCARD '*'	    	/* Wildcard used in addressing */
#define FIRST_TCP_PORT	5000	/* When using dynamic allocation */
#define LAST_TCP_PORT 5999

#define MAX_LINE 512		/* HTTP request field line */

#ifndef RULE_FILE
#define RULE_FILE		"/etc/httpd.conf"
#endif

#ifndef DEFAULT_EXPORT
#define DEFAULT_EXPORT		"/Public"
#endif

#ifdef __osf__
#define _OSF_SOURCE
#define _BSD
#endif

#include "HTUtils.h"
#include "tcp.h"		/* The whole mess of include files */
#include "HTTCP.h"		/* Some utilities for TCP */
#include "HTFormat.h"
#include "HTInit.h"
#include "HTSUtils.h"

#ifdef VMS
#include "HTVMSUtils.h"
#undef FORKING
#endif /* VMS */

#ifdef RULES			/* Use rules? */
#include "HTRules.h"
#endif

#ifdef ACCESS_AUTH		/* Use Access Authorization? */
#include "HTWriter.h"
#include "HTAAServ.h"
#endif

#include "HTFile.h"
#include "HTParse.h"
#include "HTRequest.h"
#include "HTAuth.h"		/* HTUser */
#include "HTConfig.h"
#include "HTDaemon.h"
#include "HTLog.h"
#include "HTCache.h"
#include "HTTee.h"
#include "HTFWriter.h"

/* Forking */
#ifdef FORKING
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <sys/param.h>
#include <errno.h>

#ifdef SIGTSTP	/* BSD */
#include <sys/file.h>
#include <sys/ioctl.h>
#endif /* BSD */

#endif /* FORKING */


#ifndef VMS
#include <sys/timeb.h>

#include <pwd.h>	/* Unix password file routine: getpwnam() */
#include <grp.h>	/* Unix group file routine: getgrnam()	  */
#endif /* not VMS */

#ifdef __osf__
#include <sys/wait.h>
#include <sys/resource.h>
#endif



#ifdef VMS
#define gc()
/* dummy declaration to make sure that HTML.OBJ is not included
   and other references (used in clients only) are resolved */
#ifdef __DECC
#include "HText.h"
PUBLIC CONST HTStructuredClass HTMLPresentation;
PUBLIC int HTDiag;
PUBLIC HText * HTMainText;
PUBLIC BOOL interactive;
#endif /* DECC */
#endif /* VMS */

/*	Module-Global Variables
**	-----------------------
*/
PRIVATE enum role_enum {master, slave, transient, passive} role;
PRIVATE SockA	soc_address;
PRIVATE int	soc_addrlen;
PRIVATE int	master_soc;	/* inet socket number to listen on */
PRIVATE int	com_soc;	/* inet socket number to read on */
#ifdef SELECT
PRIVATE fd_set	open_sockets;	/* Mask of channels which are active */
PRIVATE int	num_sockets;	/* Number of sockets to scan */
#endif
PRIVATE BOOLEAN	dynamic_allocation;	/* Search for port? */

#ifdef FORKING
PRIVATE BOOL	HTForkEnabled = NO; /* Do we fork() when serving.	*/
                                    /* This is enabled when server is	*/
                                    /* standalone (-a or -p option).	*/
#endif /* FORKING */

/*	Program-Global Variables
**	------------------------
*/
PUBLIC char * HTAppName = "CERN-HTTPD";	/* Application name		*/
PUBLIC char * HTAppVersion = VD; 	/* Application version		*/
PUBLIC char * log_file_name = NULL;	/* Log file name (WAIS code)	*/
PUBLIC char * HTClientProtocol = 0;	/* Protocol and version number	*/
PUBLIC char * HTServerProtocol = "HTTP/1.0";
extern char * HTClientHost;		/* Clients IP address		*/
PUBLIC char * HTClientHostName = NULL;	/* Clients host name		*/
PUBLIC int HTServerPort = 80;		/* The portnumber of this server*/
PUBLIC int child_serial = 0;		/* Serial number of child	*/

/*	Server environment when handling requests
**	-----------------------------------------
*/
PUBLIC int		HTStatusCode		= 0;
PUBLIC char *		HTReasonLine		= NULL;
PUBLIC int		HTSoc			= 0;
PUBLIC char *		HTReqLine		= NULL;
PUBLIC char *		HTReqArg		= NULL;
PUBLIC char *		HTReqArgPath		= NULL;
PUBLIC char *		HTReqArgKeywords	= NULL;
PUBLIC char *		HTReqTranslated		= NULL;
PUBLIC char *		HTReqScript		= NULL;
PUBLIC char *		HTScriptPathInfo	= NULL;
PUBLIC char *		HTScriptPathTrans	= NULL;
PUBLIC char *		HTLocation		= NULL;
PUBLIC char *		HTLastModified		= NULL;
PUBLIC char *		HTExpires		= NULL;
PUBLIC char *		HTUri			= NULL;
PUBLIC char *		HTMessageId		= NULL;
PUBLIC HTList *		HTAllowed		= NULL;
PUBLIC HTList *		HTPublic		= NULL;
PUBLIC char *		HTMetaFile		= NULL;
PUBLIC char *		HTAuthString		= NULL;
PUBLIC char *		HTWWWAuthenticate	= NULL;
PUBLIC HTAAFailReason	HTReason		= HTAA_OK;
PUBLIC char *		HTUserAgent		= NULL;
PUBLIC char *		HTReferer		= NULL;
PUBLIC time_t		HTIfModifiedSince	= 0;

PUBLIC long		HTCacheUsage		= 0;
PUBLIC int		HTChildExitStatus	= 0;
PUBLIC long		HTCachedBytes		= 0;

PUBLIC char *		gc_info_file		= NULL;
PUBLIC int		gc_pid			= 0;
PUBLIC time_t		gc_last_time		= 0;
PUBLIC int		gc_reqs_since_last	= 0;

PUBLIC time_t		cur_time		= 0;

PUBLIC BOOL		WWW_VeryVerboseFlag	= NO;


PUBLIC void reset_server_env NOARGS
{
    FREE(HTReqLine);
    FREE(HTReqArg);
    FREE(HTReqArgPath);
    FREE(HTReqTranslated);
    FREE(HTReqScript);
    FREE(HTScriptPathTrans);
    FREE(HTLocation);
    FREE(HTLastModified);
    FREE(HTExpires);
    FREE(HTUri);
    FREE(HTMessageId);
    FREE(HTMetaFile);
    FREE(HTAuthString);
    FREE(HTWWWAuthenticate);
    FREE(HTProxyHeaders);
    FREE(HTUserAgent);
    FREE(HTReferer);
    HTIfModifiedSince = 0;
    if (HTAllowed) {
	HTList_delete(HTAllowed);
	HTAllowed = NULL;
    }
    if (HTPublic) {
	HTList_delete(HTPublic);
	HTPublic = NULL;
    }

    HTStatusCode = 0;
    HTReasonLine = NULL;
    HTSoc = 0;
    HTReqArgKeywords = NULL;
    HTScriptPathInfo = NULL;
    HTReason = HTAA_OK;
    HTProxyHeaders = NULL;

    HTUser = NULL;
    HTProt = NULL;
    HTDefProt = NULL;

    HTChildExitStatus = 0;
    HTCachedBytes = 0;
}


PUBLIC void compute_server_env NOARGS
{
    char *keywords;

    FREE(HTReqArgPath);
    FREE(HTReqArgKeywords);

    FREE(HTReqTranslated);
    FREE(HTReqScript);
    FREE(HTScriptPathInfo);
    FREE(HTScriptPathTrans);
    FREE(HTLocation);

    if (HTReqArg) {
	int len;

	if ((keywords = strchr(HTReqArg, '?'))) {
	    *keywords = 0;		/* Temporarily truncate */
	    StrAllocCopy(HTReqArgPath, HTReqArg);
	    StrAllocCopy(HTReqArgKeywords, keywords+1);
	    *keywords = '?';		/* Reconstruct */
	}
	else StrAllocCopy(HTReqArgPath, HTReqArg);

	HTUnEscape(HTReqArgPath);
	HTSimplify(HTReqArgPath);	/* Remove ".." etc. */

	/*
	** HTSimplify will leave in a "/../" at the top, which can
	** be a security hole.
	*/
	len = strlen(HTReqArgPath);
	if (strstr(HTReqArgPath, "/../") ||
	    !strncmp(HTReqArgPath,"/..",3) ||
	    !strncmp(HTReqArgPath,"..",2) ||
	    (len >= 3  &&  !strcmp(&HTReqArgPath[len-3], "/.."))) {
	    CTRACE(stderr, "Illegal..... attempt to use /../ (\"%s\")\n",
		   HTReqArg);
	    HTReason = HTAA_DOTDOT;
	}
    }
}






#ifdef __hpux
PRIVATE pid_t	pgrp;
#else
#ifdef POSIXWAIT  /* if you have waitpid() instead of wait3() */
PRIVATE pid_t	pgrp;
#else
PRIVATE int	pgrp;
#endif
#endif


extern HTAAProt * HTProt;
extern HTAAProt * HTDefProt;
extern char *	HTProxyCacheRoot;
extern HTList *	HTNoLog;	/* Machines for which log entry not made */
extern BOOL	HTDisabled[(int)MAX_METHODS + 1];	/* Disabled methods	*/
extern BOOL	HTImProxy;


PRIVATE void read_gc_results NOARGS
{
    FILE * fp;
    char buf[MAX_LINE];

    if (!HTProxyCacheRoot) {
	HTLog_error("BUG: gc returned yet we're not caching -- weird");
	return;
    }
    if (!gc_info_file) {
	HTLog_error2("GC info file not set although caching to",
		     HTProxyCacheRoot);
	return;
    }
    
    fp = fopen(gc_info_file, "r");
    if (!fp) {
	HTLog_error2("Can't open gc report file:", gc_info_file);
	return;
    }
    if (fgets(buf, MAX_LINE, fp)) {
	int pid;
	long bytes_used = 0;
	long bytes_freed = 0;
	int files_used = 0;
	int files_freed = 0;
	long corrected = 0;

	sscanf(buf, "%d %ld %ld %d %d",
	       &pid, &bytes_used, &bytes_freed, &files_used, &files_freed);

	if (gc_pid  &&  pid != gc_pid) {
	    HTLog_error(
	"WARNING: gc report has wrong process id - real gc may have crashed");
	}

	corrected = HTCacheUsage - bytes_freed/1024;
	if (TRACE) {
	    fprintf(stderr, "GC REPORT... from pid %d\n", pid);
	    fprintf(stderr, "GC REPORT... %ldK USED (%d files)\n",
		    bytes_used/1024, files_used);
	    fprintf(stderr, "GC REPORT... %ldK COLLECTED (%d files)\n",
		    bytes_freed/1024, files_freed);
	    fprintf(stderr, "PARENT FIX.. %ldK USED => %ldK USED\n",
		    HTCacheUsage, corrected);
	}
	if (corrected < 0 ||
	    corrected < bytes_used/1140 ||
	    corrected > bytes_used/930) {
	    CTRACE(stderr, "PARENT...... Correcting %ldK => %ldK USED\n",
		   corrected, bytes_used/1024);
	    corrected = bytes_used/1024;
	    if (corrected<0) corrected = 0;
	}
	HTCacheUsage = corrected;
	CTRACE(stderr,"PARENT...... %ldK of cache in use\n",HTCacheUsage);
    }
    else HTLog_error2("gc report file is empty:", gc_info_file);
    fclose(fp);
}


#ifdef FORKING
/* PRIVATE							sig_child()
**
** This function is taken from:
**	W.Richard Stevens: UNIX Network Programming,
**	Prentice Hall 1990, page 82.
**
** This is a 4.3BSD SIGCLD signal handler that can be used by a
** server that's not interested in its child's exit status, but needs
** to wait for them , to avoid clogging up the system with zombies.
**
** Beware that the calling process may get an interrupted system
** call when we return, so they had better handle that.
*/
PRIVATE int sig_child NOARGS
{
#ifdef SIGTSTP	/* BSD */
#if defined(POSIXWAIT) || defined(__hpux)
    pid_t pid;
    int status;

    while ((pid = waitpid((pid_t)-1, &status, WNOHANG)) > 0) {
	if (pid == gc_pid) {
	    CTRACE(stderr, "Parent...... gc has completed\n");
	    read_gc_results();
	    gc_pid = 0;
	}
	if (HTProxyCacheRoot  &&  WIFEXITED(status)) {
	    int child_usage = HTExitStatusToKilos(WEXITSTATUS(status));
	    HTCacheUsage += child_usage;
	    CTRACE(stderr,
		   "Parent...... %dK taken up by child, total %dK used\n",
		   child_usage, HTCacheUsage);
	}
    }
#else
    int		pid;
    union wait	status;

    while ((pid = wait3(&status, WNOHANG, (struct rusage*)NULL)) > 0) {
	if (pid == gc_pid) {
	    CTRACE(stderr, "Parent...... gc has completed\n");
	    read_gc_results();
	    gc_pid = 0;
	}
	if (HTProxyCacheRoot  &&  WIFEXITED(status)) {
	    int child_usage = HTExitStatusToKilos(WEXITSTATUS(status));
	    HTCacheUsage += child_usage;
	    CTRACE(stderr,
		   "Parent...... %dK taken up by child, total %ldK used\n",
		   child_usage, HTCacheUsage);
	}
    }
#endif
#endif

    return 0;	/* Just to avoid a compiler warning */
}
#endif /* FORKING */


#ifndef VMS
PRIVATE void sig_term NOARGS
{
    if (HTStandAlone) {
#if defined(__svr4__) || defined(_POSIX_SOURCE) || defined(__hpux)
	kill(-pgrp, SIGKILL);
#else
	killpg(pgrp, SIGKILL);
#endif
	shutdown(master_soc, 2);
	close(master_soc);
    }
}

PRIVATE void set_signals NOPARAMS;

PRIVATE void sig_hup NOARGS
{
    HTReLoadRules();
    set_signals();
}
#endif /* not VMS */


PRIVATE void set_signals NOARGS
{
#ifndef VMS
    signal(SIGTERM, (void (*)())sig_term);
    signal(SIGHUP, (void (*)())sig_hup);

#if 0
    signal(SIGHUP, SIG_IGN);	/* Immune to pgrp leader death */
#endif

    /*
    ** Ignore the terminal stop signals (BSD)
    */
#ifdef SIGTTOU
    signal(SIGTTOU, SIG_IGN);
#endif
#ifdef SIGTTIN
    signal(SIGTTIN, SIG_IGN);
#endif
#ifdef SIGTSTP
    signal(SIGTSTP, SIG_IGN);
#endif

    /*
    ** Parent process is not interested in child exit statuses.
    ** We have to prevent the children from becoming zombies.
    ** In System V we just ignore SIGCLD signal.
    ** In BSD we should handle SIGCLD signal and call wait3().
    ** However, because we don't want to handle interrupted
    ** system calls on BSD, we just call sig_child() after every
    ** fork() to check if any previous childs have finished.
    */
#ifndef SIGTSTP	/* SysV */
    signal(SIGCLD, SIG_IGN);
#endif

    /*
    ** Ignore SIGPIPE signals so we don't die if client
    ** suddenly closes connection and we keep on writing
    ** to the socket.
    */
    signal(SIGPIPE, SIG_IGN);
#endif /* not VMS */
}


PRIVATE BOOL time_to_do_gc NOARGS
{
    BOOL flag = NO;

    gc_reqs_since_last++;

    if (!gc_mem_usage_control) {
	CTRACE(stderr, "Cache....... I never do gc\n");
	return NO;
    }

    if (cache_max_k > 0  &&  HTCacheUsage >= cache_max_k) {
	CTRACE(stderr, "Cache....... limit reached: %ldK used (max %ldK)\n",
	       HTCacheUsage, cache_max_k);
	flag = YES;
    }

    if (gc_req_interval && gc_reqs_since_last >= gc_req_interval) {
	CTRACE(stderr,
	       "Cache....... Child number %d will do gc (every %d does)\n",
	       child_serial, gc_req_interval);
	flag = YES;
    }

    if (gc_time_interval && cur_time-gc_last_time >= gc_time_interval) {
	CTRACE(stderr, "Cache....... Time for gc, last gc done at %s",
	       ctime(&gc_last_time));
	flag = YES;
    }

    if (flag) {
	gc_reqs_since_last = 0;
	gc_last_time = cur_time;
    }
    return flag;
}



/*	Send a string down a socket
**	---------------------------
**
**	The trailing zero is not sent.
**	There is a maximum string length.
*/
PUBLIC int HTWriteASCII ARGS2(int, soc, char *, s)
{
#ifdef NOT_ASCII
    char * p;
    char ascii[255];
    char *q = ascii;
    for (p=s; *p; p++) {
        *q++ = TOASCII(*p);
    }
    return NETWRITE(soc, ascii, p-s);
#else
    return NETWRITE(soc, s, strlen(s));
#endif
}



/*				MIME OUTPUT
**				===========
*/

struct _HTStream {
	HTStreamClass *		isa;
	HTStreamClass		class;		/* special per object */
};


/*			Catch error messages
**			--------------------
**
**	This shouldn't be necessary most of the time as HTLoadError
**	will be called.
**
**	These entry points suppress the loading of HTAlert from the WWW library.
**	These could be cleaned up and made very useful, esp
**	remote progress reporting...
*/

PUBLIC void HTAlert ARGS1(CONST char *, Msg)
{
    fprintf(stderr, "500   Server reports:  %s\r\n", Msg);
}


PUBLIC void HTProgress ARGS1(CONST char *, Msg)
{
    /* fprintf(stderr, "   %s ...\n", Msg); */
}


PUBLIC BOOL HTConfirm ARGS1(CONST char *, Msg)
{
    return(NO);
}

/*	Prompt for answer and get text back
*/
PUBLIC char * HTPrompt ARGS2(CONST char *, Msg, CONST char *, deflt)
{
    char * rep = 0;
    StrAllocCopy(rep, deflt);
    return rep;
}

PUBLIC char * HTPromptPassword ARGS1(CONST char *, Msg)
{
    char * rep = NULL;
    StrAllocCopy(rep, "");
    return rep;
}

PUBLIC void HTPromptUsernameAndPassword ARGS3(CONST char *,	Msg,
					      char **,		username,
					      char **,		password)
{
    *username = HTPrompt("Username: ", *username);
    *password = HTPromptPassword("Password: ");
}


/*	Write Error Message
**	-------------------
**
**
*/
PUBLIC int HTLoadError ARGS3(HTRequest *,	req,
			     int,		num,
			     CONST char *,	msg)
{
    CTRACE(stderr, "Returning... ERROR %d:\n ** %s\n", num,
	   msg ? msg : "-no-message-");

    if (num != 401 && num != 404) {	/* To make error logs smaller */
	char * buf = (char*)malloc((HTReqLine ? strlen(HTReqLine) : 0) + 20);
	if (!buf) outofmem(__FILE__, "HTLoadError");

	if (HTReqLine)
	    sprintf(buf, "(%d) \"%s\"", num, HTReqLine);
	else
	    sprintf(buf, "(%d)", num);
	HTLog_error2(msg, buf);
	free(buf);
    }
    else HTLog_error2(HTReqArg,NULL); /* Other info goes there automatically */

    if (req && req->output_stream) {
        HTStream * sink = req->output_stream;
	char * buffer = (char*)malloc(strlen(msg?msg:"") + 100);

	if (!buffer) outofmem(__FILE__, "HTLoadError");

	req->content_type = HTAtom_for("text/html");
	req->content_length = msg ? strlen(msg) : 0;
	HTStatusCode = num;
	HTReasonLine = (char*)msg;
	req->content_encoding = NULL;
	req->content_language = NULL;

	if (HTClientProtocol) {
	    char * headers;

	    if (num == 401)
		HTWWWAuthenticate = HTAA_composeAuthHeaders(req);

	    headers = HTReplyHeaders(req);
	    if (headers) {
		(*sink->isa->put_string)(sink, headers);
		free(headers);
	    }
	}
	sprintf(buffer,
		"<BODY><H1>Error %d</H1>\r\n\r\n%s</BODY>\r\n",
		num, (msg ? msg : ""));
	(*sink->isa->put_string)(sink, buffer);
	(*sink->isa->free)(sink);
	free(buffer);
    }
    return HT_LOADED;
}


PUBLIC int HTLoadStrToStream ARGS2(HTStream *,	  sink,
				   CONST char *,  str)
{
    if (!sink || !sink->isa || !str) return -1;
    (*sink->isa->put_string)(sink, str);
    return 0;
}


PUBLIC int HTCloseStream ARGS1(HTStream *, sink)
{
    if (sink && sink->isa) {
	(*sink->isa->free)(sink);
	return 0;
    }
    return -1;
}



/*____________________________________________________________________
**
**			Networking code
*/

/*		Bind to a TCP port
**		------------------
**
** On entry,
**	tsap	is a string explaining where to take data from.
**		"" 	means data is taken from stdin.
**		"*:1729" means "listen to anyone on port 1729"
**
** On exit,
**	returns		Negative value if error.
*/
int do_bind ARGS1(CONST char *, tsap)
{
#ifdef SELECT
    FD_ZERO(&open_sockets);	/* Clear our record of open sockets */
    num_sockets = 0;
#endif

/*  Deal with PASSIVE socket:
**
**	A passive TSAP is one which has been created by the inet daemon.
**	It is indicated by a void TSAP name.  In this case, the inet
**	daemon has started this process and given it, as stdin, the connection
**	which it is to use.
*/
    if (*tsap == 0) {			/* void tsap => passive */

	dynamic_allocation = FALSE;		/* not dynamically allocated */
	role = passive;	/* Passive: started by daemon */

#ifdef VMS

	{   unsigned short channel;	    /* VMS I/O channel */
	    struct string_descriptor {	    /* This is NOT a proper descriptor*/
		    int size;		    /*  but it will work.	      */
		    char *ptr;		    /* Should be word,byte,byte,long  */
	    } sys_input = {10, "SYS$INPUT:"};
	    int	status;		    /* Returned status of assign */
	    extern int sys$assign();

	    status = sys$assign(&sys_input, &channel, 0, 0);
	    com_soc = channel;	/* The channel is stdin */
	    CTRACE(tfp, "IP.......... Opened PASSIVE socket %d\n", channel);
	    return 1 - (status&1);
	}	
#else
	com_soc = 0;	    /* The channel is stdin */
	CTRACE(tfp, "IP.......... PASSIVE socket 0 assumed from inet daemon\n");
	return 0;		/* Good */
#endif

/*  Parse the name (if not PASSIVE)
*/
    } else {				/* Non-void TSAP */
	char *p;		/* pointer to string */
	char *q;
	struct hostent  *phost;	    /* Pointer to host - See netdb.h */
	char buffer[256];		/* One we can play with */
	register SockA * sin = &soc_address;

	strcpy(buffer, tsap);
	p = buffer;

/*  Set up defaults:
*/
#ifdef DECNET
	sin->sdn_family = AF_DECnet;	    /* Family = DECnet, host order  */
	sin->sdn_objnum = 0;                /* Default: new object number, */
#else  /* Internet */
	sin->sin_family = AF_INET;	    /* Family = internet, host order  */
	sin->sin_port = 0;		    /* Default: new port,    */
#endif
	dynamic_allocation = TRUE;	    /*  dynamically allocated */
	role = passive; 		    /*  by default */

/*  Check for special characters:
*/
	if (*p == WILDCARD) {		/* Any node */
	    role = master;
	    p++;
	}

/*  Strip off trailing port number if any:
*/
	for(q=p; *q; q++)
	    if (*q==':') {
	        int status = 0;
		*q++ = 0;		/* Terminate node string */
#ifdef DECNET
		sin->sdn_objnum = (unsigned char) HTCardinal(
					    &status, &q, (unsigned int)65535);
#else
		sin->sin_port = htons((unsigned short)HTCardinal(
					    &status, &q, (unsigned int)65535));
		if (status<0) return status;
#endif
		if (*q) return -2;  /* Junk follows port number */
		dynamic_allocation = FALSE;
		break;	    /* Exit from loop before we skip the zero */
	    } /*if*/

/* Get node name:
*/
#ifdef DECNET  /* Empty address (don't care about the command) */
	sin->sdn_add.a_addr[0] = 0;
	sin->sdn_add.a_addr[1] = 0;
	CTRACE(tfp, 
	    "Daemon...... Parsed address as port %d, DECnet %d.%d\n",
		    (int) sin->sdn_objnum,
		    (int) sin->sdn_add.a_addr[0],
		    (int) sin->sdn_add.a_addr[1] ) ;
#else
	if (*p == 0) {
	    sin->sin_addr.s_addr = INADDR_ANY; /* Default: any address */

	} else if (*p>='0' && *p<='9') {   /* Numeric node address: */
	    sin->sin_addr.s_addr = inet_addr(p); /* See arpa/inet.h */

	} else {		    /* Alphanumeric node name: */
	    phost=gethostbyname(p);	/* See netdb.h */
	    if (!phost) {
		CTRACE(tfp, "IP.......... Can't find internet node name \"%s\"\n",p);
		return HTInetStatus("gethostbyname");  /* Fail? */
	    }
	    memcpy(&sin->sin_addr, phost->h_addr, phost->h_length);
	}
	CTRACE(tfp, 
	    "Daemon...... Parsed address as port %d, inet %d.%d.%d.%d\n",
		    (int)ntohs(sin->sin_port),
		    (int)*((unsigned char *)(&sin->sin_addr)+0),
		    (int)*((unsigned char *)(&sin->sin_addr)+1),
		    (int)*((unsigned char *)(&sin->sin_addr)+2),
		    (int)*((unsigned char *)(&sin->sin_addr)+3));
	HTServerPort = (int)ntohs(sin->sin_port);
#endif
    } /* scope of p */


/*  Master socket for server:
*/
    if (role == master) {
	int one=1;
	struct linger Linger = {0,0};

/*  Create internet socket
*/
#ifdef DECNET
	master_soc = socket(AF_DECnet, SOCK_STREAM, 0);
#else
	master_soc = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
#endif
	if (master_soc<0) {
	    fprintf(stderr, "Couldn't create socket\n");
	    HTLog_error("Couldn't create socket");
	    return HTInetStatus("socket");
	}
	CTRACE(tfp, "IP.......... Opened socket number %d\n", master_soc);


/* #ifndef __svr4__ */
	if ((setsockopt(master_soc,SOL_SOCKET,SO_REUSEADDR,
			(char*)&one,sizeof(one)))
	    == -1) {
	    fprintf(stderr, "Couldn't set socket option\n");
	    HTLog_error("Couldn't set socket option");
	}
/* #endif */

/* #ifdef VMS */
	if ((setsockopt(master_soc,SOL_SOCKET,SO_KEEPALIVE,
			(char*)&one,sizeof(one)))
	    == -1) {
	    fprintf(stderr, "Couldn't set socket option\n");
	    HTLog_error("Couldn't set socket option");
	}
	if ((setsockopt(master_soc,SOL_SOCKET,SO_LINGER,
			(char*)&Linger,sizeof(Linger)))
	    == -1) {
	    fprintf(stderr, "Couldn't set socket option\n");
	    HTLog_error("Couldn't set socket option");
	}
/* #endif VMS */


/*  If the port number was not specified, then we search for a free one.
*/
#ifndef DECNET  /* irrelevant: no inetd */
	if (dynamic_allocation) {
	    unsigned short try;
	    for (try=FIRST_TCP_PORT; try<=LAST_TCP_PORT; try++) { 
		soc_address.sin_port = htons(try);
		if (bind(master_soc,
			(struct sockaddr*)&soc_address,
				/* Cast to generic sockaddr */
			sizeof(soc_address)) == 0)
		    break;
		if (try == LAST_TCP_PORT)
		    return HTInetStatus("bind");
	    }
	    CTRACE(tfp, "IP..........  Bound to port %d.\n",
		    ntohs(soc_address.sin_port));
	} else
#endif
	  {					/* Port was specified */
	    if (bind(master_soc,
		     (struct sockaddr*)&soc_address,	/* Cast to generic address */
		     sizeof(soc_address))<0)
		return HTInetStatus("bind");
	}
	if (listen(master_soc, LISTEN_BACKLOG)<0)
	    return HTInetStatus("listen");

	CTRACE(tfp, "Daemon...... Master socket(), bind() and listen() all OK\n");
#ifdef SELECT
	FD_SET(master_soc, &open_sockets);
	if ((master_soc+1) > num_sockets) num_sockets=master_soc+1;
#endif
	return master_soc;
    } /* if master */
    
    return -1;		/* unimplemented role */

} /* do_bind */



PRIVATE void input_timed_out NOARGS
{
    HTLog_error("Timed out when reading request");
    exit(0);
}

PRIVATE void output_timed_out NOARGS
{
    HTLog_error("Timed out while sending response");
    exit(0);
}

PRIVATE void script_timed_out NOARGS
{
    HTLog_error("Script timed out");
    exit(0);
}

PRIVATE void input_timeout_on NOARGS
{
#ifdef VMS
    HTVMS_start_timer(0, input_timed_out, HTInputTimeOut);
#else /* not VMS */
    signal(SIGALRM, (void (*)())input_timed_out);
    alarm(HTInputTimeOut);
#endif /* not VMS */
}

PUBLIC void output_timeout_on NOARGS
{
#ifdef VMS
    HTVMS_start_timer(0, output_timed_out, HTOutputTimeOut);
#else /* not VMS */
    signal(SIGALRM, (void (*)())output_timed_out);
    alarm(HTOutputTimeOut);
#endif /* not VMS */
}

PUBLIC void script_timeout_on NOARGS
{
#ifdef VMS
    HTVMS_start_timer(0, script_timed_out, HTScriptTimeOut);
#else /* not VMS */
    signal(SIGALRM, (void (*)())script_timed_out);
    alarm(HTScriptTimeOut);
#endif /* not VMS */
}

PUBLIC void timeout_off NOARGS
{
#ifdef VMS
    HTVMS_cancel_timer(0);
#else /* not VMS */
    alarm(0);
    signal(SIGALRM, SIG_IGN);
#endif /* not VMS */
}



/*	Handle one message
**	------------------
**
** On entry,
**	soc		A file descriptor for input and output.
** On exit,
**	returns		>0	Channel is still open.
**			0	End of file was found, please close file
**			<0	Error found, please close file.
*/
PUBLIC int HTHandle ARGS1(int, soc)
{
    HTRequest *req;
    int status;
    int redir_hops = 0;
    HTInputSocket * isoc;
    BOOL already_logged = NO;

    reset_server_env();

    isoc = HTInputSocket_new(soc);

    input_timeout_on();
    req = HTParseRequest(isoc);
    timeout_off();

    if (!req) {
	HTLog_error("Request parsing failed fatally");
	goto clean_up;
    }

    HTSoc = soc;
    req->isoc = isoc;
    req->output_stream = HTWriter_new(soc);

    if (req->method == METHOD_INVALID) {
	if (HTReqLine) {
	    char * msg = (char*)malloc(strlen(HTReqLine) + 60);
	    sprintf(msg, "Invalid request \"%s\" (unknown method)", HTReqLine);
	    HTStatusCode = 400;
	    HTReasonLine = "Invalid method";
	    HTLoadError(req, 400, msg);
	    free(msg);
	}
	else HTLoadError(req, 400,
			 "Invalid request -- completely unable to parse it");
	goto clean_up;
    }
    if (HTDisabled[(int)req->method]) {
	char msg[80];
	sprintf(msg, "Method %s is disabled on this server",
		HTMethod_name(req->method));
	HTStatusCode = 403;
	HTReasonLine = "Disabled method";
	HTLoadError(req, 403, msg);
	goto clean_up;
    }
    if (!HTReqArg) {
	char * msg = (char*)malloc((HTReqLine ?strlen(HTReqLine) :0)+30);
	if (!msg) outofmem(__FILE__, "HTHandle");
	sprintf(msg, "Invalid request \"%s\"",
		(HTReqLine ? HTReqLine : "-null-"));
	HTStatusCode = 400;
	HTReasonLine = "Invalid request";
	HTLoadError(req, 400, msg);
	free(msg);
	goto clean_up;
    }


do_redirection_on_fly:

/*
**	Check Access Authorization (and translate by rules at the same time).
*/
#ifdef ACCESS_AUTH

#ifdef VMS
    /* enable sysprv to get to security files */
    HTVMS_enableSysPrv();
#endif /* VMS */

    HTStatusCode = HTAA_checkAuthorization(req);
    CTRACE(stderr, "AA.......... check returned %d\n", HTStatusCode);
    CTRACE(stderr, "Translated.. \"%s\"\n",
	   HTReqTranslated ? HTReqTranslated : "-null-");

#ifdef VMS
    HTVMS_disableSysPrv();
#endif /* VMS */

    if (HTStatusCode != 200  &&  HTStatusCode != 302) {
	HTLoadError(req, HTStatusCode, HTAA_statusMessage(req));
	goto clean_up;
    }
    /* otherwise access is authorized, continue */

#ifdef VMS
  if (HTVMS_authSysPrv() == YES) {
      char *filename = HTReqTranslated;
      char *uid_name = NULL;

      if (!filename)
         filename = HTReqScript;

      /* if SETUID is specified we will try to access the file for that UID */
      if (HTDoSetUid  &&  HTUser  &&  HTUser->username)
         uid_name = HTUser->username;

      /* if UID still not specified we take the one from the protect rule */
      if (!uid_name)
         uid_name = HTAA_getUidName();

      /* check if we can access the file from UID; GID is ignored */
      if ((filename) && (0 != strcmp(uid_name,""))) {
      	 CTRACE(stderr, "HTHandle.... checking access to file '%s' for user '%s'\n",
	       	     "server", filename, uid_name);

         /* enable SYSPRV if indeed we have access; actual access is done via system field in the protection */
         HTVMS_enableSysPrv();

	CTRACE(stderr,
	       "WARNING..... serving document when running with SYSPRV\n");

         if (HTVMS_checkAccess(filename,uid_name,req->method) != YES) {
            HTVMS_disableSysPrv();
            HTLoadError(req, 403, "Forbidden -- by rule");
            goto clean_up;
         }

         /* access under VMS authorized, go on with SYSPRV enabled... */
      } /* No Protection defined, or no uidname given */
  }
  else CTRACE(stderr, "HTHandle.... Not running with SYSPRV\n");

#else /* not VMS */
    if (getuid() == 0) {
	struct passwd *pw = NULL;
	int uid = 0;
	int gid = 0;

	if (HTDoSetUid  &&  HTUser  &&  HTUser->username) {
	    pw = getpwnam(HTUser->username);
	    if (pw) {
		uid = pw->pw_uid;
		gid = pw->pw_gid;
		CTRACE(stderr, "SETUID...... %s [uid %d] [gid %d]\n",
		       HTUser->username, uid, gid);
	    }
	    else {
		HTLog_error2("SETUID ERROR no such user:", HTUser->username);
	    }
	}
	if (!pw) { 
	    uid = HTAA_getUid();
	    gid = HTAA_getGid();
	}

	if (!pw && !(pw = getpwuid(uid)))
	    HTLog_errorN("Failed to get passwd entry for uid",uid);
	else if (initgroups(pw->pw_name, gid) == -1)
	    HTLog_error2("Failed to init groups for user",pw->pw_name);

	CTRACE(stderr, "Doing... setgid(%d) and setuid(%d)\n", gid, uid);
	if (setgid(gid) == -1)
	    HTLog_errorN("Failed to set group id to", gid);
	if (setuid(uid) == -1) {
	    HTLog_errorN("Failed to set user id to", uid);
	    CTRACE(stderr, "WARNING..... serving doc when running as root\n");
	}
    }
    else CTRACE(stderr, "HTHandle.... Not running as root (uid=%d)\n",
		getuid());
#endif /* not VMS */
#endif	/* ACCESS_AUTH */



/*
**	Handle command
*/
    if (HTReason == HTAA_OK_GATEWAY) {  /* NOT a local file -- gatewaying */

	output_timeout_on();
	HTImProxy = YES;

	if (HTProxyCacheRoot  &&  req->method==METHOD_GET &&
	    !req->authorization  &&  !HTReqArgKeywords) {
	    char * cfn = NULL;
	    FILE * cf = NULL;

	    if (cache_lookup(HTReqTranslated, &cfn, &cf)) {
		/* Cache file exists */
		CTRACE(stderr,"Cache hit... Reading from \"%s\"\n", cfn);
		cf = fopen(cfn, "r");
		if (cf) {
		    if (HTClientProtocol)
			HTFileCopy(cf, req->output_stream);
		    else
			HTFileCopyNoHeaders(cf, req->output_stream);
		    fclose(cf);
		    (*req->output_stream->isa->free)(req->output_stream);
		    status = HT_LOADED;
		    HTLog_cacheAccess(cfn,
				      NULL,
				      HTUser ? HTUser->username : NULL,
				      req);
		    already_logged = YES;
		    free(cfn);
		}
		else {
		    HTLog_error2("Couldn't open cache file for reading:",cfn);
		    free(cfn);
		    /* Load from remote host */
		    status = HTLoadToStream(HTReqTranslated, NO, req);
		}
	    }
	    else {
		if (cfn && cf) {
		    /* Write to cache file */
		    req->output_stream =
			HTTee(req->output_stream,
			      HTProxyCache(cf, cfn, HTReqTranslated));
		    CTRACE(stderr,
			   "Cache....... Writing to lockfile \"%s%s\"\n",
			   cfn, LOCK_SUFFIX);
		    free(cfn);
		}
		else CTRACE(stderr, "Cache....... Not caching\n");

		status = HTLoadToStream(HTReqTranslated, NO, req);
	    }
	}
	else {
	    /* Normal retrieve with no caching */
	    status = HTLoadToStream(HTReqTranslated, NO, req);
	}

	HTImProxy = NO;
	timeout_off();

    }
    else if (HTReason == HTAA_OK_REDIRECT) {	/* Do a redirection */
	output_timeout_on();
	status = HTLoadRedirection(req);
	timeout_off();
    }
    else if (HTReqScript) {	/* Call a script to generate a document */
	status = HTCallScript(req);
    }
    else if (HTReqArgKeywords) {
	if (HTSearchScript) {	/* Search enabled */
	    if (HTReqTranslated) {
		StrAllocCopy(HTScriptPathTrans, HTReqTranslated);
		StrAllocCopy(HTScriptPathInfo, HTReqArgPath);
		StrAllocCopy(HTReqScript, HTSearchScript);
		req->content_length = 0;
		status = HTCallScript(req);
	    }
	    else {
		HTLog_error("Translated NULL when should do search");
		HTLoadError(req, 403, "Sorry, forbidden by rule");
	    }
	}
	else {
	    CTRACE(stderr, "HTHandle.... can't perform search %s\n", HTReqArg);
	    status =
		HTLoadError(req, 501,
			    "Sorry, this server does not perform searches.");
	}
    } /* If keywords given */
    else {  /* Load the document normally into the client */

	CTRACE(stderr, "HTHandle.... method %s\n", HTMethod_name(req->method));

	output_timeout_on();

        if (req->method == METHOD_GET  ||
	    req->method == METHOD_CHECKOUT) {

	    if (HTIfModifiedSince  &&  HTLastModified  &&
		parse_http_time(HTLastModified) <= HTIfModifiedSince) {
		CTRACE(stderr,
		       "Not modified since %s", ctime(&HTIfModifiedSince));
		HTStatusCode = 304;
		HTReasonLine = "Use Local Copy";
		status = HTLoadHead(req);
	    }
	    else
		status = HTRetrieve(req);

	} else if (req->method == METHOD_HEAD) {

	    status = HTLoadHead(req);

	} else if (req->method == METHOD_PUT  ||
		   req->method == METHOD_CHECKIN) {

	    BOOL OkToPut = NO;

	    if (OkToPut) {
		HTPut(req);
	    }
	    else if (HTPutScript) {
		CTRACE(stderr,
		       "HTHandle.... handling PUT with script \"%s\"\n",
		       HTPutScript);

		StrAllocCopy(HTReqScript, HTPutScript);
		StrAllocCopy(HTScriptPathInfo, HTReqArgPath);
		HTScriptPathTrans = HTParse(HTReqTranslated, "",
					    PARSE_PATH | PARSE_PUNCTUATION);
		status = HTCallScript(req);
	    }
	    else {
		HTLoadError(req, 500,
			    "This server is not configured to handle PUT");
	    }

	} else if (req->method == METHOD_POST) {

	    BOOL OkToAnnotate = NO;

	    if (OkToAnnotate) {
		HTAnnotate(req);
	    }
	    else if (HTPostScript) {
		CTRACE(stderr,
		       "HTHandle.... handling POST with script \"%s\"\n",
		       HTPostScript);

		StrAllocCopy(HTReqScript, HTPostScript);
		StrAllocCopy(HTScriptPathInfo, HTReqArgPath);
		HTScriptPathTrans=HTParse(HTReqTranslated, "",
					  PARSE_PATH | PARSE_PUNCTUATION);
		status = HTCallScript(req);
	    }
	    else {
		HTLoadError(req, 500,
			    "This server is not configured to handle POST");
	    }

	} else {
	    char *msg;

	    msg = (char*)malloc(strlen(HTMethod_name(req->method))+50);
	    if (!msg) outofmem(__FILE__, "HTHandle");

	    sprintf(msg, "Method '%s' invalid or not implemented",
		    HTMethod_name(req->method));
	    HTLoadError(req, 501, msg);
	    CTRACE(stderr, "Daemon...... %s\n", msg);
	    free(msg);
	}

	timeout_off();

    } /* Handle command */

    if (status == HT_REDIRECTION_ON_FLY) {
	CTRACE(stderr, "Redirection. on the fly for \"%s\"\n", HTLocation);
	if (++redir_hops > 10)
	    HTLoadError(req, 501,
	    "Too many redirection-on-the-fly hops, max 10 (probably looping)");
	else {
	    if (req->method != METHOD_GET  &&  req->method != METHOD_HEAD)
		req->method = METHOD_GET;
	    FREE(HTReqArg);	/* Leak fixed AL 6 Feb 1994 */
	    HTReqArg = HTLocation;
	    HTLocation = NULL;
	    compute_server_env();

	    goto do_redirection_on_fly;
	}
    }

#ifdef VMS
#ifdef ACCESS_AUTH
    /* disable SYSPRV again */
    HTVMS_disableSysPrv();
#endif /* ACCESS_AUTH */
#endif /* VMS */


 clean_up:

    /*
    **	Log the call
    */
    if (!already_logged)
	HTLog_access(NULL,  HTUser ? HTUser->username : NULL,  req);
    if (req) {
	if (req->isoc) HTInputSocket_free(req->isoc);
	HTRequest_delete(req);
    }

    return 0;		/* End of file - please close socket */

} /* HTHandle */



/*      Handle incoming messages				server_loop()
**	-------------------------
**
** On entry:
**
**      timeout         -1 for infinite, 0 for poll, else in units of 10ms
**
** On exit,
**	returns		The status of the operation, <0 if failed.
**			0	means end of file
**
*/
PRIVATE int server_loop NOARGS
{
    int tcp_status;		/* <0 if error, in general */
    int timeout = -1;		/* No timeout required but code exists */
    for(;;) {

/*  If it's a master socket, then find a slave:
*/
    	if (role == master) {
#ifdef SELECT
	    fd_set		read_chans;
	    fd_set		write_chans;
	    fd_set		except_chans;
	    int			nfound;	    /* Number of ready channels */
	    struct timeval	max_wait;   /* timeout in form for select() */
    
	    FD_ZERO(&write_chans);	    /* Clear the write mask */
	    FD_ZERO(&except_chans);	    /* Clear the exception mask */

/*  If timeout is required, the timeout structure is set up. Otherwise
**  (timeout<0) a zero is passed instead of a pointer to the struct timeval.
*/
	    if (timeout>=0) {
		max_wait.tv_sec = timeout/100;
		max_wait.tv_usec = (timeout%100)*10000;
	    }
    
	    for (com_soc=(-1); com_soc<0;) {	/* Loop while connections keep coming */
    
		
/*  The read mask expresses interest in the master channel for incoming
**  connections) or any slave channel (for incoming messages).
*/

/*  Wait for incoming connection or message
*/
	        read_chans = open_sockets;	 /* Read on all active channels */
		CTRACE(stderr,
"Daemon...... Waiting for connection or message. (Mask=%x hex, max=%x hex).\n", 
		       *(unsigned int *)(&read_chans),
		       (unsigned int)num_sockets);
		nfound=select(num_sockets, &read_chans,
			      &write_chans, &except_chans,
			      timeout >= 0 ? &max_wait : 0);
	
		if (nfound<0) return HTInetStatus("select");
		if (nfound==0) return 0;	/* Timeout */

/*	We give priority to existing connected customers. When there are
**	no outstanding commands from them, we look for new customers.
*/
/*  	If a message has arrived on one of the channels, take that channel:
*/
		{
		    int i;
		    for(i=0; i<num_sockets; i++)
			if (i != master_soc)
			    if (FD_ISSET(i, &read_chans)) {
				CTRACE(stderr,
				       "Info........ Message waiting on socket %d\n",
				       i);
				com_soc = i;		/* Got one! */
				break;
			    }
		    if (com_soc>=0) break; /* Found input socket */
		    
		} /* block */
		
/*  If an incoming connection has arrived, accept the new socket:
*/
		if (FD_ISSET(master_soc, &read_chans)) {
    			soc_addrlen = sizeof(soc_address); /* JS 930107 */
			CTRACE(tfp, "Daemon...... New incoming connection:\n");
			tcp_status = accept(master_soc,
					(struct sockaddr *)&soc_address,
					&soc_addrlen);
			if (tcp_status<0)
			    return HTInetStatus("accept");
			CTRACE(tfp, "Accepted.... new socket %d\n",
			    tcp_status);
			FD_SET(tcp_status, &open_sockets);
			if ((tcp_status+1) > num_sockets)
				num_sockets=tcp_status+1;
    
		} /* end if new connection */
    
    
	    } /* loop on event */
	
#else	/* SELECT not supported */
    
/*	    if (com_soc<0)   No slaves: must accept */
	      SockA peer_soc;
	      int peer_len = sizeof (peer_soc);
		    CTRACE(tfp, 
		    "Daemon...... Waiting for incoming connection...\n");
#ifdef DECNET
		    tcp_status = accept(master_soc, &peer_soc, &peer_len);
#else  /* For which machine is this ??? rsoc is undeclared, what's mdp ? */
		    tcp_status = accept(master_soc,
				    &rsoc->mdp.soc_tcp.soc_address,
				    &rsoc->mdp.soc_tcp.soc_addrlen);
#endif
		    if (tcp_status<0)
			return HTInetStatus("accept");
		    com_soc = tcp_status;	/* socket number */
		    CTRACE(tfp, "Accepted.... socket %d\n", tcp_status);
/*	    }  end if no slaves */