
/* MODULE							HTLog.c
**		LOGGING FOR HTTPD
**
** AUTHORS:
**	AL	Ari Luotonen	luotonen@dxcern.cern.ch
**
** HISTORY:
**	13 Feb 94  AL	Separated from HTDaemon.c
**
** BUGS:
**
**
*/

#include <stdio.h>
#include "tcp.h"
#include "HTLog.h"
#include "HTList.h"
#include "HTGroup.h"	/* HTIpMaskMatch() */
#include "HTDaemon.h"
#include "HTConfig.h"
#include "HTLog.h"
#include "HTAuth.h"

#define n_pick(x,y)	((x) ? (x) : ((y) ? (y) : "-"))
#define n_noth(x)	((x) ? (x) : "")
#define n_hyph(x)	(((x) && (*x)) ? (x) : "-")

#define HTABS(x)	(((x) < 0) ? (-(x)) : (x))
#ifdef VMS
#define gmtime localtime
#else
#define HTFlush fflush
#endif

PRIVATE FILE *	access_log		= NULL;
PRIVATE char *	access_log_name		= NULL;
PRIVATE FILE *	error_log		= NULL;
PRIVATE char *	error_log_name		= NULL;
PRIVATE FILE *	cache_access_log	= NULL;


extern BOOL HTUseGMT;
extern BOOL HTNewLogFileFormat;
extern HTList * HTNoLog;
extern char * HTClientHost;
extern char * HTClientHostName;
extern long HTProxyBytes;	/* Bytes transferred via proxy */


#define ERR_QUEUE_SIZE 20
PRIVATE char * err_queue[ERR_QUEUE_SIZE];
PRIVATE int err_index = 0;


#ifdef VMS
#define TIMES_TO_REOPEN	20
PRIVATE int HTFlush ARGS1(FILE *, fptr)
{
static access_log_times = 0;
static error_log_times = 0;

   /* flush C rtl buffers */
   fflush(fptr);
  
   /* flush RMS buffers */
   fsync(fileno(fptr));

   /* reopen file if some lines have been written, so we can read it under VMS */
   if (fptr == access_log)
   {
      access_log_times++;
      if (access_log_times > TIMES_TO_REOPEN)
      {
         access_log_times = 0;
         access_log = freopen(access_log_name, "a+", access_log, "shr=get",
                               "shr=put", "shr=upd");
      }
   }
   else
   if (fptr == error_log)
   {
      error_log_times++;
      if (error_log_times > TIMES_TO_REOPEN)
      {
         error_log_times = 0;
         error_log = freopen(error_log_name, "a+", error_log, "shr=get",
                               "shr=put", "shr=upd");
      }
   }

   return(1);
}
#endif /* VMS */


PRIVATE char * time_escapes ARGS1(CONST char *, filename)
{
    time_t t;
    struct tm * gmt;
    char * result;

    time(&t);
    gmt = gmtime(&t);

    result = malloc(strlen(filename) + 5 + 1);
    sprintf(result, filename, (gmt->tm_year) % 100, gmt->tm_mon + 1);
    return result;
}


PUBLIC BOOL HTLog_accessOpen ARGS1(CONST char *, filename)
{
    if (!filename) return NO;

    access_log_name = time_escapes(filename);
#ifdef VMS
    access_log = fopen(access_log_name, "a+", "shr=get", "shr=put", "shr=upd");
#else
    access_log = fopen(access_log_name, "a");
#endif /* VMS */

    if (access_log) {
	CTRACE(stderr, "Log......... \"%s\"\n", access_log_name);
	return YES;
    }
    else {
	CTRACE(stderr,
	       "ERROR....... Can't open log file \"%s\"\n",
	       access_log_name);
	return NO;
    }

    if (HTCacheAccessLog  &&  HTCacheAccessLog[0]=='/') {
	CTRACE(stderr, "Opening..... also cache access log \"%s\"\n",
	       HTCacheAccessLog);
#ifdef VMS
	cache_access_log = fopen(HTCacheAccessLog,
				 "a+", "shr=get", "shr=put", "shr=upd");
#else
	cache_access_log = fopen(HTCacheAccessLog, "a");
#endif /* VMS */

	if (!cache_access_log)
	    HTLog_error2("Can't open cache access log file:",
			 HTCacheAccessLog);
    }
}


PRIVATE void flush_queue NOARGS
{
    int i;

    for(i=0; i<err_index && i<ERR_QUEUE_SIZE; i++) {
	if (err_queue[i]) {
	    fprintf(error_log, "%s\n", err_queue[i]);
	    free(err_queue[i]);
	    err_queue[i] = NULL;
	}
    }
    if (err_index > ERR_QUEUE_SIZE)
	HTLog_errorN("There were more errors, queued only", ERR_QUEUE_SIZE);
    HTFlush(error_log);
    err_index = 0;
}


PUBLIC BOOL HTLog_errorOpen ARGS1(CONST char *, filename)
{
    if (error_log) {
	if (filename)
	    HTLog_error("Error log already open");
	return NO;
    }
    if (filename)
	error_log_name = time_escapes(filename);
    else if (access_log_name) {
	char * dot = strrchr(access_log_name, '.');
	if (dot) *dot = 0;

	error_log_name = (char*)malloc(strlen(access_log_name) + 7);
	sprintf(error_log_name, "%s.error", access_log_name);

	if (dot) *dot = '.';
    }
    else return NO;

#ifdef VMS
    error_log = fopen(error_log_name, "a+", "shr=get", "shr=put", "shr=upd");
#else
    error_log = fopen(error_log_name, "a");
#endif /* VMS */

    if (error_log) {
	CTRACE(stderr, "Error log... \"%s\"\n", error_log_name);
	flush_queue();
	return YES;
    }
    else {
	HTLog_error2("Can't open error log file:",error_log_name);
	return NO;
    }
}



PRIVATE char * log_time NOARGS
{
    struct tm * gorl;		/* GMT or localtime */
    static char ret[30];
    int z = 0;			/* Timezone diff in minutes */

    if (HTUseGMT)
	gorl = gmtime(&cur_time);
    else {
	gorl = localtime(&cur_time);
#if defined(SIGTSTP) && !defined(__sgi)	/* BSD and have tm_gmtoff */
	z = gorl->tm_gmtoff/60;
#else					/* SysV */
	z = timezone;
#endif
    }

    sprintf(ret, "%02d/%s/%04d:%02d:%02d:%02d %c%02d%02d",
	    gorl->tm_mday,
	    month_names[gorl->tm_mon],
	    1900 + gorl->tm_year,
	    gorl->tm_hour,
	    gorl->tm_min,
	    gorl->tm_sec,
	    z < 0 ? '-' : '+',
	    HTABS(z) / 60,
	    HTABS(z) % 60);

    return ret;
}



PUBLIC void HTLog_access ARGS3(CONST char *,	r_ident,
			       CONST char *,	authuser,
			       HTRequest *,	req)
{
    time_t t;
    struct tm * gorl;

    if (!access_log) return;

    if (HTNoLog) {
	HTList * cur = HTNoLog;
	char * templ;

	while ((templ = (char*)HTList_nextObject(cur))) {
	    if (HTIpMaskMatch(templ, HTClientHost, HTClientHostName)) {
		CTRACE(stderr,
		       "Log......... NOT making a log entry for %s [%s]\n",
		       HTClientHost, n_noth(HTClientHostName));
		return;
	    }
	}
    }

    if (HTNewLogFileFormat) {
	char buf[30];

	if (HTStatusCode > 0) {
	    if (HTReason == HTAA_OK_GATEWAY && 
		HTProxyBytes > 0 && HTStatusCode==200)
		sprintf(buf, "%d %ld", HTStatusCode, HTProxyBytes);
	    else if (req->content_length > 0)
		sprintf(buf, "%d %d", HTStatusCode, req->content_length);
	    else
		sprintf(buf, "%d -", HTStatusCode);
	}
	else strcpy(buf, " - -");

	fprintf(access_log,
		"%s %s %s [%s] \"%s\" %s\n",
		n_pick(HTClientHostName,HTClientHost),
		(r_ident && strcmp(r_ident,"unknown")) ? r_ident : "-",
		n_hyph(authuser),
		log_time(),
		n_noth(HTReqLine),
		buf);
    }
    else {	/* Still old format */
	time(&t);
	/*
	** Time fixes by STREETS@fneqs1.fnal.gov Nov 3 93.
	** Added by AL Dec 6 93.
	*/
	if (HTUseGMT)
	    gorl = gmtime(&t);
	else
	    gorl = localtime(&t);

	fprintf(access_log, "%24.24s %s %s\n",
		asctime(gorl), HTClientHost, HTReqLine);
    }

    HTFlush(access_log);
}


PUBLIC void HTLog_cacheAccess ARGS4(char *,		cache_file,
				    CONST char *,	rfc931,
				    CONST char *,	authuser,
				    HTRequest *,	req)
{
    if (!cache_file || !HTCacheAccessLog ||
	(int)strlen(cache_file) < (int)strlen(HTProxyCacheRoot) + 7) {
	HTLog_access(rfc931, authuser, req);
    }
    else {
	BOOL do_close = NO;
	if (!cache_access_log) {	/* Not already opened at startup */
	    char * logname = NULL;
	    struct stat stat_info;

	    if (stat(cache_file,&stat_info) != -1) {
		req->content_length = stat_info.st_size;
	    }

	    if (strchr(HTCacheAccessLog,'/')) {/*Absolute path (a single log)*/
		StrAllocCopy(logname, HTCacheAccessLog);
	    }
	    else { /* Log file individually for each http/ftp/gopher host */
		char * end = cache_file + strlen(HTProxyCacheRoot);
		end = strchr(end+1,'/');	/* Skip http,ftp,gopher */
		if (end) {
		    end = strchr(end+1,'/');	/* Skip machine name	*/
		    if (end) *end=0;
		}
		logname = (char*)malloc(strlen(cache_file) + 2 +
					strlen(HTCacheAccessLog));
		if (!logname) outofmem(__FILE__, "HTLog_cacheAccess");
		sprintf(logname, "%s/%s", cache_file, HTCacheAccessLog);
		if (end) *end='/';
		do_close = YES;
	    }

#ifdef VMS
	    cache_access_log=fopen(logname,"a+","shr=get","shr=put","shr=upd");
#else
	    cache_access_log = fopen(logname, "a");
#endif /* VMS */

	    if (!cache_access_log) {
		HTLog_error2("Can't open cache access log file:", logname);
		HTLog_access(rfc931, authuser, req);
	    }
	    FREE(logname);
	}

	if (cache_access_log) {	/* Ok, at least now it's opened */
	    FILE * tmp = access_log; /* This is ugly, I just change a global */
	    access_log = cache_access_log;
	    HTLog_access(rfc931, authuser, req);
	    access_log = tmp;
	    if (do_close) {
		fclose(cache_access_log);
		cache_access_log = NULL;
	    }
	    else {
		fflush(cache_access_log);
	    }
	}
    }
}


PRIVATE char * status_name NOARGS
{
    switch (HTReason) {

    /* Success */
      case HTAA_OK:			return "OK";
      case HTAA_OK_GATEWAY:		return "OK-GATEWAY";
      case HTAA_OK_REDIRECT:		return "OK-REDIRECT";

    /* 401 cases */
      case HTAA_NO_AUTH:		return "NOT AUTHENTICATED";
      case HTAA_NOT_MEMBER:		return "NOT AUTHORIZED";

    /* 403 cases */
      case HTAA_IP_MASK:		return "FORBIDDEN BY IP";
      case HTAA_IP_MASK_PROXY:		return "FORBIDDEN BY IP FOR PROXY";
      case HTAA_BY_RULE:		return "FORBIDDEN BY RULE";
      case HTAA_NO_ACL:			return "NO ACL FILE";
      case HTAA_NO_ENTRY:		return "NO ACL ENTRY";
      case HTAA_SETUP_ERROR:		return " * SETUP ERROR * ";
      case HTAA_DOTDOT:			return "DOT-DOT IN URL";
      case HTAA_HTBIN:			return "HTBIN OFF";
      case HTAA_INVALID_REDIRECT:	return "INVALID REDIRECT";
      case HTAA_NOT_ALLOWED:		return "PUT NOT ALLOWED";

    /* 404 cases */
      case HTAA_NOT_FOUND:		return "NOT FOUND";
      case HTAA_MULTI_FAILED:		return "MULTI FAILED";

    /* Others */
      default:	return "?";		/* A new status maybe?? */
    } /* switch */
}


PUBLIC void HTLog_error2 ARGS2(CONST char *, msg,
			       CONST char *, param)
{
    char * t = log_time();
    char * extras = NULL;

    CTRACE(stderr, "ERROR....... [%s] %s %s\n", t, n_hyph(msg), n_noth(param));

    if (HTClientHost) {
	char * host = n_pick(HTClientHostName,HTClientHost);
	char * user = (HTUser && HTUser->username) ? HTUser->username : "";
	char * ref  = n_noth(HTReferer);

	extras = (char*)malloc(strlen(host) + strlen(user) + strlen(ref) +100);
	if (!extras) outofmem(__FILE__, "HTLog_error2");

	sprintf(extras, " [%s] [host: %s", status_name(), host);
	if (*user) {
	    strcat(extras, " user: ");
	    strcat(extras, user);
	}
	if (*ref) {
	    strcat(extras, " referer: ");
	    strcat(extras, ref);
	}
	strcat(extras, "]");
    }

    if (!error_log) {
	char * q = (char*)malloc(strlen(n_noth(msg)) + strlen(n_noth(param)) +
				 strlen(n_noth(extras)) + 100);
	if (!q) outofmem(__FILE__, "HTLog_error2");

	sprintf(q, "[%s]%s %s %s",
		t, n_noth(extras), n_noth(msg), n_noth(param));
	err_queue[err_index++] = q;
    }
    else {
	fprintf(error_log, "[%s]%s %s %s\n",
		t, n_noth(extras), n_noth(msg), n_noth(param));
	HTFlush(error_log);
    }

    if (extras) free(extras);
}


PUBLIC void HTLog_error ARGS1(CONST char *, msg)
{
    HTLog_error2(msg, NULL);
}


PUBLIC void HTLog_errorN ARGS2(CONST char *,	msg,
			       int,		num)
{
    char buf[20];
    sprintf(buf, "%d", num);
    HTLog_error2(msg, buf);
}

